
#include "impulse.h"
#include "m_array.h"

/*
  Assumes
   Impulse : (horizon x svar->n_var) matrix or null pointer
   A       : (svar->n_var x svar->n_var) matrix
   F       : (svar->n_var x svar->n_pre) matrix or null pointer
   svar    : valid structural vector autoregression structure.
   horizon : number of impulse response horizons

  Results
   Given the system defined by svar (see initial comments in var.c), and the particular parameters
   A and F, the ith row and jth column of the kth matrix in the returned matrix array is the 
   response of the ith variable to a one standard deviation shock in the jth equation at horizon
   k.  If the passed matrix Impulse is null, then it is created.      

  Notes
   No error checking is performed. 
*/
TMatrix* ImpulseResponseArray(TMatrix *Impulse, TMatrix A, TMatrix F, int n_lag, int horizon)
{
 TMatrix X;
 int i, j, k;

 if (!Impulse) 
  {
   Impulse=CreateMatrixArray(horizon);
   for (i=0; i < horizon; i++)
    Impulse[i]=CreateMatrix(M_ROW(A),M_ROW(A));
  }

 X=CreateMatrix(horizon,M_ROW(A));
 for (j=0; j < M_ROW(A); j++)
  {
   ImpulseResponse(X,A,F,n_lag,horizon,j);
   for (i=0; i < M_ROW(A); i++)
     for (k=0; k < horizon; k++)
       Impulse[k][i][j]=X[k][i];
  }
 FreeMatrix(X);

 return Impulse;
}

/*
  Assumes
   Impulse : (horizon x svar->n_var) matrix or null pointer
   A       : (svar->n_var x svar->n_var) matrix
   F       : (svar->n_var x svar->n_pre) matrix or null pointer
   svar    : valid structural vector autoregression structure.
   horizon : number of impulse response horizons
   shock   : zero-based index of variable to shock

  Results
   Given the system defined by svar (see initial comments in var.c), and the particular parameters
   A and F, the ith row of the returned matrix gives the impulse response at horizon i to a one
   standard deviation shock to the variable indexed by shock.  If the passed matrix Impulse is
   null, then it is created.      

  Notes
   No error checking is performed. 
*/
TMatrix ImpulseResponse(TMatrix Impulse, TMatrix A, TMatrix F, int n_lag, int horizon, int shock)
{
 TMatrix A_Inverse;
 TVector y;
 int i, j, k, l, m, r;

 if (!Impulse) Impulse=CreateMatrix(horizon,M_ROW(A));

 y=InitializeVector(CreateVector(M_ROW(A)),0.0);
 y[shock]=1.0;
 A_Inverse=Inverse_LU((TMatrix)NULL,A);
 for (i=0; i < horizon; i++)
  {
   for (l=1, r=0; (l <= i) && (l <= n_lag); l++)
    for (k=0; k < M_ROW(A); k++, r++)
     for (j=0; j < M_ROW(A); j++)
      y[j]+=Impulse[i-l][k]*F[r][j];
   ProductVM(y,y,A_Inverse);
   memcpy(Impulse[i],y,sizeof(PRECISION)*M_ROW(A));
   InitializeVector(y,0.0);
  }

 FreeMatrix(A_Inverse);
 FreeVector(y);

 return Impulse;
}

/*
  Assumes
   svar        : valid structural vector autoregression structure.
   horizon     : number of impulse response horizons.
   shock       : zero-based index of variable to shock.
   sample_size : sample size used to compute histogram.

  Returns
   Creates and returns a matrix histogram.  The ith row and jth column is the histogram of the 
   response of variable j at horizon i to a a one standard deviation shock in the variable indexed
   by shock.   
*/
TMatrixHistogram* ImpulseResponseHistogram(T_SVAR *svar, int horizon, int shock, int sample_size)
{
 TMatrixHistogram *Histogram;
 TMatrix Impulse=(TMatrix)NULL, A, F;
 int i;

 A=CreateMatrix(svar->n_var,svar->n_var);
 F=CreateMatrix(svar->n_pre,svar->n_var);
 Histogram=CreateMatrixHistogram(horizon,svar->n_var,1000,HISTOGRAM_VARIABLE);
 RandomInitialValue(A,100,svar);
 for (i=0; i < sample_size; i++)
  {
   Simulate_Full(A,F,svar);
   Impulse=ImpulseResponse(Impulse,A,F,svar->n_lag,horizon,shock);
   AddMatrixObservation(Impulse,Histogram);   
  }

 FreeMatrix(Impulse);
 FreeMatrix(A);
 FreeMatrix(F);

 return Histogram;
}
