/*

A TStringVector is an array of pointers to characters.  It is assumed that
the character pointer is either the null pointer or a pointer to a valid null 
terminated string.  The length of the array is encoded in the sizeof(int)
bytes preceding the beginning of the array.  Allocation and deallocation of
TStringVector pointers must be done with CreateStringVector() and 
FreeStringVector().  Any null terminated strings are owned by the structure
and are freed by the call to FreeStringVector().

A TStringMatrix is an array of arrays of pointers to characters.  It is 
assumed that the character pointer is either the null pointer or a pointer 
to a valid null terminated string.  The number of rows and columns in the
matrix is encoded in the 2*sizeof(int) bytes preceding the beginning of the
array.  Allocation and deallocation of TStringVector pointers must be done 
with CreateStringVector() and FreeStringVector().  Any null terminated 
strings are owned by the structure and are freed by the call to 
FreeStringVector().

*/

#include "StringMatrix.h"
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <ctype.h>

static char* GetNextField(char *buffer, char delimiter, int flag);

/******************************************************************************/
/****************************** Error Functions *******************************/
/******************************************************************************/
static int STRING_ERROR=0;
static int STRING_ERROR_TERMINATE=ALL_STRING_ERR ^ EOF_STRING_ERR;
static int STRING_ERROR_VERBOSE=ALL_STRING_ERR ^ EOF_STRING_ERR;

int GetStringError(void)
{
 return STRING_ERROR;
}

void ClearStringError(void)
{
 STRING_ERROR=0;
}

int SetStringErrorTerminate(int err)
{
 int terminate=STRING_ERROR_TERMINATE;
 STRING_ERROR_TERMINATE=err & ALL_STRING_ERR;
 return terminate;
}

int SetStringErrorVerbose(int err)
{
 int verbose=STRING_ERROR_VERBOSE;
 STRING_ERROR_VERBOSE=err & ALL_STRING_ERR;
 return verbose;
}

void StringError(int err)
{
 STRING_ERROR=err & ALL_STRING_ERR;
 switch (STRING_ERROR_VERBOSE & STRING_ERROR)
  {
   case 0:
    break;
   case MEM_STRING_ERR:
    printf("\nOut of memory.\n");
    break;
   case SIZE_STRING_ERR:
    printf("\nMatrices/vectors not conformable.\n");
    break;
   case NULL_STRING_ERR:
     printf("\nUnexpected null string vector/matrix encountered.\n");
    break;
   case EOF_STRING_ERR:
     printf("\nEOF encountered.\n");
     break;
   default:
    printf("\nUnknown error.\n");
    break;
  }
 if (STRING_ERROR_TERMINATE & STRING_ERROR) exit(err);
}
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

TStringVector CreateStringVector(int m)
{
 TStringVector x;
 if (m < 0)
  {
   StringError(SIZE_STRING_ERR);
   return (TStringVector)NULL;
  }
 if (!(x=(TStringVector)((int*)malloc(m*sizeof(char*)+sizeof(int))+1)))
  {
   StringError(MEM_STRING_ERR);
   return (TStringVector)NULL;
  }
 V_DIM(x)=m;
 while (--m >= 0) x[m]=(char*)NULL;
 return x;
}

TStringMatrix CreateStringMatrix(int m, int n)
{
 TStringMatrix x;
 int i;
 if ((m < 0) || (n < 0))
  {
   StringError(SIZE_STRING_ERR);
   return (TStringMatrix)NULL;
  }
 if (!(x=(TStringMatrix)((int*)malloc(m*(n*sizeof(char*)+sizeof(char**))+2*sizeof(int))+2)))
  {
   StringError(MEM_STRING_ERR);
   return (TStringMatrix)NULL;
  }
 M_ROW(x)=m;
 M_COL(x)=n;
 x[0]=(char**)(x+m);
 for (i=1; i < m; i++) x[i]=x[i-1]+n;
 for (i=m*n-1; i >= 0; i--) x[0][i]=(char*)NULL;
 return x;
}

void FreeStringVector(TStringVector x)
{
 int i;
 if (x)
  {
   for (i=V_DIM(x)-1; i >= 0; i--)
    if (x[i]) free(x[i]);
   free((int*)x-1);
  }
}

void FreeStringMatrix(TStringMatrix x)
{
 int i;
 if (x)
  {
   for (i=M_ROW(x)*M_COL(x)-1; i >= 0; i--)
    if (x[0][i]) free(x[0][i]);
   free((int*)x-2);
  }
}

/*
   Assumes
     A : m-string vector or null pointer
     B : m-string vector

   Results
     A = B.  If A is null pointer, A is created.

   Returns
     Returns A upon success and null on failure.  Call GetStringError()
     to determine the cause of failure.
*/
TStringVector EquateStringVector(TStringVector A, TStringVector B)
{
 int i;
 TStringVector C=A;
 if (!B)
  {
   StringError(NULL_STRING_ERR);
   return (TStringVector)NULL;
  }
 if (!A)
   {
    if (!(A=CreateStringVector(V_DIM(B))))
     return (TStringVector)NULL;
   }
  else
   if (V_DIM(A) != V_DIM(B))
    {
     StringError(SIZE_STRING_ERR);
     return (TStringVector)NULL;
    }
 STRING_ERROR=0;
 for (i=V_DIM(A)-1; i >= 0; i--)
  {
   if (A[i]) free(A[i]);
   A[i]=DuplicateString(B[i]);
  }
 if (STRING_ERROR)
  {
   if (!C) FreeStringVector(A);
   return (TStringVector)NULL;
  }
 return A;
}

/*
   Assumes
     A : mxn string matrix or null pointer
     B : mxn string matrix

   Results
     A = B.  If A is null pointer, A is created.

   Returns
     Returns A upon success and null on failure.  Call GetStringError()
     to determine the cause of failure.
*/
TStringMatrix EquateStringMatrix(TStringMatrix A, TStringMatrix B)
{
 int i;
 TStringMatrix C=A;
 if (!B)
  {
   StringError(NULL_STRING_ERR);
   return (TStringMatrix)NULL;
  }
 if (!A)
   {
    if (!(A=CreateStringMatrix(M_ROW(B),M_COL(B))))
     return (TStringMatrix)NULL;
   }
  else
   if ((M_ROW(A) != M_ROW(B)) || (M_COL(A) != M_COL(B)))
    {
     StringError(SIZE_STRING_ERR);
     return (TStringMatrix)NULL;
    }
 STRING_ERROR=0;
 for (i=M_ROW(A)*M_COL(A)-1; i >= 0; i--)
  {
   if (A[0][i]) free(A[0][i]);
   A[0][i]=DuplicateString(B[0][i]);
  }
 if (STRING_ERROR)
  {
   if (!C) FreeStringMatrix(A);
   return (TStringMatrix)NULL;
  }
 return A;
}

/*
   Searches the null terminated string buffer for first occurance of delimiter.
   If delimiter does not appear, the a null pointer is returned.  The null
   terminated string buffer is not changed.

   Flag value:
     STRIP_WHITESPACE
       Returns pointer to the first non-whitespace character after the first
       occurance of delimiter, the second occurance of delimiter, or the
       terminating null character.

     Other values
       Returns a pointer to the character after the first occurance of
       delimiter.
*/
static char* GetNextField(char *buffer, char delimiter, int flag)
{
 while (*buffer)
  if (*buffer == delimiter)
    {
     buffer++;
     if (flag & STRIP_WHITESPACE)
      while (*buffer && (*buffer != delimiter) && isspace(*buffer)) buffer++;
     return buffer;
    }
   else
    buffer++;
 return (char*)NULL;
}

/*
   Creates a string vector from a null terminated delimited string buffer. The
   null terminated string buffer is not changed.

   If flag is
     STRIP_WHITESPACE
       Leading and trailing spaces are not stored.

     REMOVE_EMPTY_FIELDS
       Empty fields (after leading and trailing spaces are removed if
       STRIP_WHITESPACE is set) are not stored.

   A null return indicates a possible error.  Call GetStringError() to determine
   the error.  Upon return, GetStringError() is gaurenteed to return one of the
   following values:
     NULL_STRING_ERROR - A null string was passed.
     MEM_STRING_ERROR - Out of memory 
     0 - REMOVE_EMPTY_FIELDS was set and there were no non-empty fields.
*/
TStringVector CreateStringVectorFromDelimited(char *buffer, char delimiter, int flag)
{
 int k, n;
 TStringVector A;
 char *ptr;
 if (!buffer)
  {
   StringError(NULL_STRING_ERR);
   return (TStringVector)NULL;
  }
 if (flag & STRIP_WHITESPACE)
   for (ptr=buffer; *ptr && (*ptr != delimiter) && isspace(*ptr); ptr++);
  else
   ptr=buffer;
 for (n=0; ptr; ptr=GetNextField(ptr,delimiter,flag))
  if (!(flag & REMOVE_EMPTY_FIELDS) || (*ptr && (*ptr != delimiter)))
   n++;
 if (n > 0)
  {
   if (!(A=CreateStringVector(n)))
    {
     FreeStringVector(A);
     StringError(MEM_STRING_ERR);
     return (TStringVector)NULL;
    }
  if (flag & STRIP_WHITESPACE)
    for (ptr=buffer; *ptr && (*ptr != delimiter) && isspace(*ptr); ptr++);
   else
    ptr=buffer;
  for (n=0; ptr; ptr=GetNextField(ptr,delimiter,flag))
   if (!(flag & REMOVE_EMPTY_FIELDS) || (*ptr && (*ptr != delimiter)))
    {
     for (k=0; ptr[k] && (ptr[k] != delimiter); k++);
     if (flag & STRIP_WHITESPACE)
      while ((k > 0) && isspace(ptr[k-1])) k--;
     if (!(A[n]=(char*)malloc(k+1)))
      {
       FreeStringVector(A);
       StringError(MEM_STRING_ERR);
       return (TStringVector)NULL;
      }
     memcpy(A[n],ptr,k);
     A[n][k]='\0';
     n++;
    }
   STRING_ERROR=0;
   return A;
  }
 STRING_ERROR=0;
 return (TStringVector)NULL;
}

/*
   Creates string vector from the delimited file f.  Starting from the 
   current file position, reads until a newline character or the end of 
   the file has been reached.  If a newline is encountered, the newline
   or newline/carriage return pair is not retained.

   Sets STRING_ERROR to:
     EOF_STRING_ERR if end of file condition is true
     MEM_STRING_ERR if unable to allocate necessary memory
     0              if no error

   Returns NULL upon error.  Call GetStringError() to determine if error
   has occured or the line was empty.
*/
TStringVector ReadDelimitedStringVector(FILE *f, char delimiter, int flag)
{
 TStringVector A;
 char *buffer, *ptr;
 int n=0, m=1024;
 if (!(buffer=(char*)malloc(m)))
  {
   StringError(MEM_STRING_ERR);
   return (TStringVector)NULL;
  }
 if (!fgets(buffer,m,f))
  {
   free(buffer);
   StringError(EOF_STRING_ERR);
   return (TStringVector)NULL;
  }
 n=strlen(buffer);
 while ((n == m-1) && (buffer[n-1] != '\n'))
  if (!(buffer=(char*)realloc(ptr=buffer,m+=1024)))
    {
     free(ptr);
     StringError(MEM_STRING_ERR);
     return (TStringVector)NULL;
    }
   else
    if (!fgets(buffer+n,m-n,f))
      buffer[n]='\0';
     else
      n+=strlen(buffer+n);
 if ((n > 0) && (buffer[n-1] == '\n'))
  buffer[((n > 1) && (buffer[n-2] == '\r')) ? n-2 : n-1]='\0';
 A=CreateStringVectorFromDelimited(buffer,delimiter,flag);
 free(buffer);
 return A;
}

/*
   Creates string matrix from delimited file.
*/
TStringMatrix ReadDelimitedStringMatrix(FILE *f, char delimiter, int flag)
{
 TStringMatrix M;
 TStringVector A;
 struct TVectorStack
  {
   struct TVectorStack *next;
   TStringVector A;
  } *S, *T;
 int i, j, m=0, n=0;
 S=(struct TVectorStack*)NULL;
 do
  {
   if (A=ReadDelimitedStringVector(f,delimiter,flag))
    {
     if (!(T=(struct TVectorStack*)malloc(sizeof(struct TVectorStack))))
      {
       FreeStringVector(A);
       while (S)
        {
         FreeStringVector(S->A);
         T=S;
         S=S->next;
         free(T);
        }
       StringError(MEM_STRING_ERR);
       return (TStringMatrix)NULL;
      }
     m++;
     T->A=A;
     T->next=S;
     S=T;
     if (V_DIM(A) > n) n=V_DIM(A);
    }
  }
 while (!STRING_ERROR);
 if (M=CreateStringMatrix(m,n))
  for (i=m-1; i >= 0; i--)
   {
    for (j=V_DIM(S->A)-1; j >= 0; j--)
     {
      M[i][j]=S->A[j];
      S->A[j]=(char*)NULL;
     }
    FreeStringVector(S->A);
    T=S;
    S=S->next;
    free(T);
   }
 return M;
}

void PrintDelimitedStringVector(FILE *f, TStringVector x, char delimiter)
{
 int i;
 if (x)
  {
   if (x[0]) fprintf(f,"%s",x[0]);
   for (i=1; i < V_DIM(x); i++)
    if (x[i])
      fprintf(f,"%c%s",delimiter,x[i]);
     else
      fprintf(f,"%c",delimiter);
   fprintf(f,"\n");
  }
}

void PrintDelimitedStringMatrix(FILE *f, TStringMatrix x, char delimiter)
{
 int i, j;
 if (x)
  {
   for (i=0; i < M_ROW(x); i++)
    {
     if (x[i][0]) fprintf(f,"%s",x[i][0]);
     for (j=1; j < M_COL(x); j++)
      if (x[i][j])
        fprintf(f,"%c%s",delimiter,x[i][j]);
       else
        fprintf(f,"%c",delimiter);
     fprintf(f,"\n");
    }
  }
}

/******************************************************************************/
/************************ String Manipulation Routines ************************/
/******************************************************************************/
char* DuplicateString(char *buffer)
{
 char *dup;
 if (!buffer) return (char*)NULL;
 if (!(dup=(char*)malloc(strlen(buffer)+1)))
  {
   StringError(MEM_STRING_ERR);
   return (char*)NULL;
  }
 strcpy(dup,buffer);
 return dup;
}

void StripLeadingSpaces(char *buffer)
{
 int i;
 if (buffer)
  {
   for (i=0; isspace(buffer[i]); i++);
   if (i > 0) memmove(buffer,buffer+i,strlen(buffer+i));
  }
}

void StripTrailingSpaces(char *buffer)
{
 int i;
 if (buffer)
  {
   for (i=strlen(buffer)-1; (i >= 0) && isspace(buffer[i]); i--);
   buffer[i+1]='\0';
  }
}

void StripQuotations(char *buffer)
{
  int i;
  if (buffer)
    {
      if (buffer[i=strlen(buffer)-1] == '"') buffer[i]='\0';
      if (buffer[0] == '"') memmove(buffer,buffer+1,i+1);
    }
}

void StripSpaces(char *buffer)
{
 int i;
 if (buffer)
  {
   for (i=strlen(buffer)-1; (i >= 0) && isspace(buffer[i]); i--);
   buffer[i+1]='\0';
   for (i=0; isspace(buffer[i]); i++);
   if (i > 0) memmove(buffer,buffer+i,strlen(buffer+i));
  }
}

/*
  Floating point number is of the form

  [white space][+/-]digits[.[digits]][E/e[+/-]digits]white space/null character

  or

  [white space][+/-].digits[E/e[+/-]digits]white space/null character

  where characters in square brackets are optional.

  Returns one if valid floating point number and zero otherwise.
*/
int IsFloat(char *buffer)
{
 int i=0;

 if (!buffer) return 0;

 /* Strip leading white space */
 while (isspace(buffer[i])) i++;

 /* Mantissa OK? */
 if ((buffer[i] == '+') || (buffer[i] == '-')) i++;
 if (isdigit(buffer[i]))
   {
    while (isdigit(buffer[++i]));
    if ((buffer[i] == '.'))
     while (isdigit(buffer[++i]));
   }
  else
   if ((buffer[i] == '.'))
     if (isdigit(buffer[++i]))
       while (isdigit(buffer[++i]));
      else
       return 0;
    else
     return 0;

 /* Is exponent OK? */
 if ((buffer[i] == 'e') || (buffer[i] == 'E'))
  {
   if ((buffer[++i] == '+') || (buffer[i] == '-')) i++;
   if (isdigit(buffer[i]))
     while (isdigit(buffer[++i]));
    else
     return 0;
  }

 /* Is end of string or trailing white space */
 if (buffer[i] && !isspace(buffer[i])) return 0;

 return 1;
}

/*
  Integers are of the form

   [white space][+/-]digits[.]white space/null character

  where characters in square brackets are optional.

  Returns one if valid integer and zero otherwise.
*/
int IsInteger(char *buffer)
{
 int i=0;

 if (!buffer) return 0;

 /* Strip leading white space */
 while (isspace(buffer[i])) i++;

 /* Mantissa OK? */
 if ((buffer[i] == '+') || (buffer[i] == '-')) i++;
 if (isdigit(buffer[i]))
   {
    while (isdigit(buffer[++i]));
    if ((buffer[i] == '.')) i++;
   }
  else
   return 0;

 /* Is end of string or trailing white space */
 if (buffer[i] && !isspace(buffer[i])) return 0;

 return 1;
}


