  
#include "bmatrix.h"

#include <math.h>
#include <stdlib.h>
#include <memory.h>

/********************
#include "blas_lapack.h"
#define USE_BLAS_LAPACK
/********************/

/********************
#include "mkl.h"
#include "blas_lapack.h"
#define USE_BLAS_LAPACK
/********************/

/********************
#define USE_INLINE
/********************/

/******************************************************************************/
/***************************** Uniary Operations ******************************/
/******************************************************************************/
/*
  Assumes:
   x : n-vector
   y : n-vector
   n : positive

  Results:
   x[i] = -y[i] for 0 <= i < n

  Returns:
   0 upon success

  Notes:
   x and y do not have to be distinct
*/
int bNegative(PRECISION *x, PRECISION *y, int n)
{
 while (--n >= 0) x[n]=-y[n];
 return 0;
}

/*
  Assumes:
   x : n-vector
   y : n-vector
   n : positive

  Results:
   x[i] = fabs(y[i]) for 0 <= i < n

  Returns:
   0 upon success

  Notes:
   x and y do not have to be distinct
*/
int bAbs(PRECISION *x, PRECISION *y, int n)
{
#if defined USE_INLINE
  __asm {
        fninit
        mov  edi,x
        mov  esi,y
        mov  eax,n
        dec  eax       
a1:     fld  PRECISION_WORD ptr [esi+PRECISION_SIZE*eax]
        fabs
        fstp PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]                     
        dec  eax
        jge  a1 
       }
 return 0;
#else
 while (--n >= 0) x[n]=fabs(y[n]);
 return 0;
#endif
}

/*
  Assumes:
   x : m x n matrix in row major format
   y : n x m matrix in row major format

  Results:
   x = y'
   x[i*n+j]=y[j*m+i] for 0 <= i < m and 0 <= j < n
*/
int bTranspose(PRECISION *x, PRECISION *y, int m, int n)
{
 int i, j, k;
 for (i=k=m*n-1; k >= 0; i--)
  for (j=i; j >= 0; j-=m)
   x[k--]=y[j];
 return 0;
}

/*
  Assumes:
   x : m x m matrix in row major format

  Results:
   x = x'
   x[i*m+j]=x[j*m+i] for 0 <= i < m and 0 <= j < m
*/
int bTransposeInPlace(PRECISION *x, int m)
{
 PRECISION tmp;
 int i, j, k;
 for (j=m-1; j >= 0; j--)
  for (i=(j-1)*m+j, k=j*m+j-1; i >= 0; k--, i-=m)
   {
    tmp=x[i];
    x[i]=x[k];
    x[k]=tmp;
   }
 return 0;
}
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/


/******************************************************************************/
/***************************** Addition Routines ******************************/
/******************************************************************************/
/*
  Assumes:
   x : n-vector
   y : n-vector
   z : n-vector
   n : positive

  Results:
   x[i] = y[i] + z[i]  for 0 <= i < n

  Returns:
   0 upon success

  Notes:
   x, y and z do not have to be distinct
*/
int bAdd(PRECISION *x, PRECISION *y, PRECISION *z, int n)
{
#if defined USE_INLINE
 int rtrn;
 __asm {
        fninit
        mov     edi,x
        mov     esi,y
        mov     ebx,z
        mov     eax,n
        jmp     a2   

a1:     fld     PRECISION_WORD ptr [esi+PRECISION_SIZE*eax]
        fadd    PRECISION_WORD ptr [ebx+PRECISION_SIZE*eax]
        fstp    PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]   
        
a2:     dec     eax
        jge     a1
       
        /* check coprocessor for errors */
	fnstsw  ax                                               
	and     ax,0x000D
	je      a3
	mov     rtrn,0
        jmp     a4
a3:     mov     rtrn,FLOAT_ERR
a4:
       }      
 return rtrn;
#else
 while (--n >= 0) x[n]=y[n]+z[n];
 return 0;
#endif

}

/*
  Assumes:
   x : n-vector
   y : n-vector
   z : n-vector
   n : positive

  Results
   x[i] = y[i] - z[i]  for 0 <= i < n

  Returns:
   0 upon success

  Notes:
   x, y and z do not have to be distinct
*/
int bSubtract(PRECISION *x, PRECISION *y, PRECISION *z, int n)
{
#if defined USE_INLINE
 int rtrn;
 __asm {
        fninit
        mov     edi,x
        mov     esi,y
        mov     ebx,z
        mov     eax,n
        jmp     a2   

a1:     fld     PRECISION_WORD ptr [esi+PRECISION_SIZE*eax]
        fsub    PRECISION_WORD ptr [ebx+PRECISION_SIZE*eax]
        fstp    PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]   
        
a2:     dec     eax
        jge     a1
       
        /* check coprocessor for errors */
	fnstsw  ax                                               
	and     ax,0x000D
	je      a3
	mov     rtrn,0
        jmp     a4
a3:     mov     rtrn,FLOAT_ERR
a4:
       }      
 return rtrn;
#else
 while (--n >= 0) x[n]=y[n]-z[n];
 return 0;
#endif
}
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/


/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/*
  Assumes:
   x : n-vector
   y : n-vector
   n : positive

  Results:
   x[i] = s * y[i] for 0 <= i < n

  Returns:
   0 upon success

  Notes:
   x and y do not have to be distinct
*/
int bMultiply(PRECISION *x, PRECISION *y, PRECISION s, int n)
{
#ifndef USE_INLINE
 while (--n >= 0) x[n]=s*y[n];
#else
 __asm {
        fninit
        mov  edi,x
        mov  esi,y
        mov  eax,n
        dec  eax
        fld  s
a1:     fld  PRECISION_WORD ptr [esi+PRECISION_SIZE*eax]
        fmul st,st(1)
        fstp PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]                     
        dec  eax
        jge  a1
       }
#endif
 return 0;
}


/*
  Assumes:
   x : array of length mn
   y : array of length mp
   z : array of length pn
   m, n and p are positive

  Results:
                          x          y          z
       xt   yt   zt   row major  row major  row major    results
       ----------------------------------------------------------
        0    0    0     m x n      m x p      p x n     x = y * z
        1    0    0     n x m      m x p      p x n     x'= y * z
        0    1    0     m x n      p x m      p x n     x = y'* z
        1    1    0     n x m      p x m      p x n     x'= y'* z
        0    0    1     m x n      m x p      n x p     x = y * z'
        1    0    1     n x m      m x p      n x p     x'= y * z'
        0    1    1     m x n      p x m      n x p     x = y'* z'
        1    1    1     n x m      p x m      n x p     x'= y'* z'

   or

                          x          y          z
       xt   yt   zt   col major  col major  col major    results
       ----------------------------------------------------------
        0    0    0     n x m      p x m      n x p     x'= y'* z'
        1    0    0     m x n      p x m      n x p     x = y'* z'
        0    1    0     n x m      m x p      n x p     x'= y * z'
        1    1    0     m x n      m x p      n x p     x = y * z'
        0    0    1     n x m      p x m      p x n     x'= y'* z
        1    0    1     m x n      p x m      p x n     x = y'* z
        0    1    1     n x m      m x p      p x n     x'= y * z
        1    1    1     m x n      m x p      p x n     x = y * z

   or
                          x          y          z
       xt   yt   zt    (m x n)    (m x p)    (p x n)     results
       -----------------------------------------------------------
        0    0    0   row major  row major  row major   x = y * z
        1    0    0   col major  row major  row major   x = y * z
        0    1    0   row major  col major  row major   x = y * z
        1    1    0   col major  col major  row major   x = y * z
        0    0    1   row major  row major  col major   x = y * z
        1    0    1   col major  row major  col major   x = y * z
        0    1    1   row major  col major  col major   x = y * z
        1    1    1   col major  col major  col major   x = y * z

  Returns:
   0 upon success

  Notes:
   An (n x m) matrix x is in row major format if x[i][j]=x[i*n+j] and is in
   column major format if x[i][j]=x[i+j*m]. 
    
*/

int bMatrixMultiply(PRECISION *x, PRECISION *y, PRECISION *z, int m, int n, int p, int xt, int yt, int zt)
{
#if defined USE_INLINE
  /*
     ebx = uj
     ecx = vi
     edx = x
     esi = pu
     edi = pv
  */
 int i, j, ui, vj, puj, pvi;
 __asm {
         fninit
         cmp  yt,0
         je   short dest1                                   // if (yt)

         mov  ui,PRECISION_SIZE                             //     ui=PRECISION_SIZE
                                              
         mov  ebx,m         
         shl  ebx,PRECISION_SHIFT                           //     uj=m*PRECISION_SIZE 

         jmp  short dest2                                   //   else 

dest1:   mov  eax,p                                         
         shl  eax,PRECISION_SHIFT
         mov  ui,eax                                        //     ui=p*PRECISION_SIZE 

         mov  ebx,PRECISION_SIZE                            //     uj=PRECISION_SIZE   

dest2:   cmp  zt,0
         je   short dest3                                   // if (zt) 

         mov  ecx,1                                         //     vi=1
                                           
         mov  eax,p  
         shl  eax,PRECISION_SHIFT                                
         mov  vj,eax                                        //     vj=p*PRECISION_SIZE

         jmp  short dest4                                   //   else

dest3:   mov  ecx,n                                         //     vi=n
                                     
         mov  vj,PRECISION_SIZE                             //     vj=PRECISION_SIZE
        
dest4:   mov  eax,p
         dec  eax
         imul eax,ecx
         mov  pvi,eax                                       // pvi=(p-1)*vi

         mov  edx,m
         imul edx,n
         dec  edx
         shl  edx,PRECISION_SHIFT
         add  edx,x                                         // x+=(m*n-1) 
        
         cmp  xt,0
         je   short dest5                                   // if (xt)

         mov  eax,p
         dec  eax
         imul eax,ebx
         add  y,eax                                         //     y+=(p-1)*uj
         
         mov  eax,n
         dec  eax
         imul eax,vj                                        //     j=(n-1)*vj
outer_1: mov  j,eax                                     

         mov  edi,z
         add  edi,eax                                       //     pv=z+j
       
         mov  eax,m
         dec  eax
         imul eax,ui                                        //     i=(m-1)*ui
inner_1: mov  i,eax                                         

         mov  esi,y                   
         add  esi,eax                                       //     pu=y+i

         mov  eax,pvi                                       //     k=pvi
          
         fld  PRECISION_WORD ptr [esi]
         fmul PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]   //     st=(*pu)*pv[k];  

         sub  eax,ecx                                       
         jl   short dest2_1                                 //     (k-=vi) >= 0

dest1_1: sub  esi,ebx                                       //     pu-=uj

         fld  PRECISION_WORD ptr [esi]
         fmul PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]
         fadd                                               //     st+=(*pu)*pv[k]

         sub  eax,ecx                                       
         jge  short dest1_1                                 //     (k-=vi) >= 0

dest2_1: fstp PRECISION_WORD ptr [edx]                      //     *x=st

         sub  edx,PRECISION_SIZE                            //     x--
         
         mov  eax,i
         sub  eax,ui
         jge  short inner_1                                 //     (i=i-ui) >= 0

         mov  eax,j
         sub  eax,vj
         jge  short outer_1                                 //     (j=j-vj) >= 0
        
         jmp  short dest6                                   //   else

dest5:   mov  eax,p
         dec  eax
         imul eax,ebx
         mov  puj,eax                                       //     puj=(p-1)*uj;

         mov  eax,m
         dec  eax
         imul eax,ui                                        //     i=(m-1)*ui
outer_0: mov  i,eax

         mov  esi,y
         add  esi,eax                                       //     pu=y+i

         mov  eax,n
         dec  eax
         imul eax,vj                                        //     j=(n-1)*vj
inner_0: mov  j,eax

         mov  edi,z
         add  edi,eax                                       //     pv=z+j

         add  esi,puj
         mov  eax,pvi                                       //     pu+=puj
            
         fld  PRECISION_WORD ptr [esi]
         fmul PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]   //     st=(*pu)*pv[k] 

         sub  eax,ecx                                       
         jl   short dest2_0                                 //     (k-=vi) >= 0

dest1_0: sub  esi,ebx                                       //     pu-=uj

         fld  PRECISION_WORD ptr [esi]
         fmul PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]
         fadd                                               //     st+=(*pu)*pv[k]

         sub  eax,ecx
         jge  short dest1_0                                 //     (k-=vi) >= 0

dest2_0: fstp PRECISION_WORD ptr [edx]                      //     *x=st

         sub  edx,PRECISION_SIZE                            //     x--

         mov  eax,j
         sub  eax,vj
         jge  short inner_0                                 //     (j-=vj) >= 0

         mov  eax,i
         sub  eax,ui
         jge  short outer_0                                 //     (i-=ui) >= 0
dest6:
       }
 return 0;
#elif defined USE_BLAS_LAPACK
 int transy, transz, dy, dz;
 PRECISION beta=0.0, alpha=1.0;
#if PRECISION_SIZE == 4
 if (xt) 
   {
    if (yt) {transy='N'; dy=m;} else {transy='T'; dy=p;}   
    if (zt) {transz='N'; dz=p;} else {transz='T'; dz=n;}   
    sgemm(&transy,&transz,&m,&n,&p,&alpha,y,&dy,z,&dz,&beta,x,&m); 
   }
  else
   {
    if (yt) {transy='T'; dy=m;} else {transy='N'; dy=p;}   
    if (zt) {transz='T'; dz=p;} else {transz='N'; dz=n;}   
    sgemm(&transz,&transy,&n,&m,&p,&alpha,z,&dz,y,&dy,&beta,x,&n);  
   }
#else
 if (xt) 
   {
    if (yt) {transy='N'; dy=m;} else {transy='T'; dy=p;}   
    if (zt) {transz='N'; dz=p;} else {transz='T'; dz=n;}   
    dgemm(&transy,&transz,&m,&n,&p,&alpha,y,&dy,z,&dz,&beta,x,&m); 
   }
  else
   {
    if (yt) {transy='T'; dy=m;} else {transy='N'; dy=p;}   
    if (zt) {transz='T'; dz=p;} else {transz='N'; dz=n;}   
    dgemm(&transz,&transy,&n,&m,&p,&alpha,z,&dz,y,&dy,&beta,x,&n);  
   }
#endif
 return 0;
#elif defined USE_IMSL
 #if PRECISION_SIZE == 4
	 if (xt==0 && yt==0 && zt==0)
		imsl_f_mat_mul_rect("A*B",IMSL_A_MATRIX,m,p,y,IMSL_B_MATRIX,p,n,z,IMSL_RETURN_USER,x,0);
	 else if (xt==1 && yt==0 && zt==0)
		imsl_f_mat_mul_rect("trans(B)*trans(A)",IMSL_A_MATRIX,m,p,y,IMSL_B_MATRIX,p,n,z,IMSL_RETURN_USER,x,0);
	 else if (xt==0 && yt==1 && zt==0)
		imsl_f_mat_mul_rect("trans(A)*B",IMSL_A_MATRIX,p,m,y,IMSL_B_MATRIX,p,n,z,IMSL_RETURN_USER,x,0);
	 else if (xt==1 && yt==1 && zt==0)
		imsl_f_mat_mul_rect("trans(B)*A",IMSL_A_MATRIX,p,m,y,IMSL_B_MATRIX,p,n,z,IMSL_RETURN_USER,x,0);
	 else if (xt==0 && yt==0 && zt==1)
		imsl_f_mat_mul_rect("A*trans(B)",IMSL_A_MATRIX,m,p,y,IMSL_B_MATRIX,n,p,z,IMSL_RETURN_USER,x,0);
	 else if (xt==1 && yt==0 && zt==1)
		imsl_f_mat_mul_rect("B*trans(A)",IMSL_A_MATRIX,m,p,y,IMSL_B_MATRIX,n,p,z,IMSL_RETURN_USER,x,0);
	 else if (xt==0 && yt==1 && zt==1)
		imsl_f_mat_mul_rect("trans(A)*trans(B)",IMSL_A_MATRIX,p,m,y,IMSL_B_MATRIX,n,p,z,IMSL_RETURN_USER,x,0);
	 else if (xt==1 && yt==1 && zt==1)
		imsl_f_mat_mul_rect("B*A",IMSL_A_MATRIX,p,m,y,IMSL_B_MATRIX,n,p,z,IMSL_RETURN_USER,x,0);
	 else
		imsl_f_mat_mul_rect("A*B",IMSL_A_MATRIX,m,p,y,IMSL_B_MATRIX,p,n,z,IMSL_RETURN_USER,x,0);
 #else
	 if (xt==0 && yt==0 && zt==0)
		imsl_d_mat_mul_rect("A*B",IMSL_A_MATRIX,m,p,y,IMSL_B_MATRIX,p,n,z,IMSL_RETURN_USER,x,0);
	 else if (xt==1 && yt==0 && zt==0)
		imsl_d_mat_mul_rect("trans(B)*trans(A)",IMSL_A_MATRIX,m,p,y,IMSL_B_MATRIX,p,n,z,IMSL_RETURN_USER,x,0);
	 else if (xt==0 && yt==1 && zt==0)
		imsl_d_mat_mul_rect("trans(A)*B",IMSL_A_MATRIX,p,m,y,IMSL_B_MATRIX,p,n,z,IMSL_RETURN_USER,x,0);
	 else if (xt==1 && yt==1 && zt==0)
		imsl_d_mat_mul_rect("trans(B)*A",IMSL_A_MATRIX,p,m,y,IMSL_B_MATRIX,p,n,z,IMSL_RETURN_USER,x,0);
	 else if (xt==0 && yt==0 && zt==1)
		imsl_d_mat_mul_rect("A*trans(B)",IMSL_A_MATRIX,m,p,y,IMSL_B_MATRIX,n,p,z,IMSL_RETURN_USER,x,0);
	 else if (xt==1 && yt==0 && zt==1)
		imsl_d_mat_mul_rect("B*trans(A)",IMSL_A_MATRIX,m,p,y,IMSL_B_MATRIX,n,p,z,IMSL_RETURN_USER,x,0);
	 else if (xt==0 && yt==1 && zt==1)
		imsl_d_mat_mul_rect("trans(A)*trans(B)",IMSL_A_MATRIX,p,m,y,IMSL_B_MATRIX,n,p,z,IMSL_RETURN_USER,x,0);
	 else if (xt==1 && yt==1 && zt==1)
		imsl_d_mat_mul_rect("B*A",IMSL_A_MATRIX,p,m,y,IMSL_B_MATRIX,n,p,z,IMSL_RETURN_USER,x,0);
	 else
		imsl_d_mat_mul_rect("A*B",IMSL_A_MATRIX,m,p,y,IMSL_B_MATRIX,p,n,z,IMSL_RETURN_USER,x,0);
 #endif
 return 0;
#else
 int i, j, k, ui, uj, vi, vj, puj, pvi;
 PRECISION *pu, *pv;
 if (yt) { ui=1; uj=m; } else { ui=p; uj=1; }
 if (zt) { vi=1; vj=p; } else { vi=n; vj=1; }
 pvi=(p-1)*vi;
 x+=(m*n-1);
 if (xt)
   {
    y+=(p-1)*uj;
    for (j=(n-1)*vj; j >= 0; j-=vj)
     for (pv=z+j, i=(m-1)*ui; i >= 0; i-=ui)
      {
       k=pvi;
       pu=y+i;
       *x=(*pu)*pv[k];
       while ((k-=vi) >= 0) *x+=(*(pu-=uj))*pv[k];
       x--;
      }
   }
  else
   {
    puj=(p-1)*uj;
    for (i=(m-1)*ui; i >= 0; i-=ui)
     for (pu=y+i, j=(n-1)*vj; j >= 0; j-=vj)
      {
       k=pvi;
       pu+=puj;
       pv=z+j;
       *x=(*pu)*pv[k];
       while ((k-=vi) >= 0) *x+=(*(pu-=uj))*pv[k];
       x--;
      }
   }
 return 0;
#endif
}
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

/******************************************************************************/
/***************************** LU Decompositions ******************************/
/******************************************************************************/
/*
   Assumes
     p   : integer array of length q=min(m,n)
     x   : m x n matrix in row major format

   Results
     Computes the LU decomposition of A with partial pivoting. The LU 
     decomposition of a matrix A is

                                 A = P * L * U

     where P is a (m x m) permutation matrix, L is a (m x q) lower triangular 
     matrix with ones on the diagonal, and U is a (q x n) upper triangular
     matrix.  These matrices are stored as follows.

      U is stored in the upper part of x, including the diagonal.

      L is stored in the lower part of x.  The diagonal of L is not stored.

      The matrix P is defined by

                   P = P(0,p[0])*P(1,p[1])*...*P(q-1,p[q-1])

      where P(r,s) is the (m x m) matrix obtained by permuting the rth and sth 
      rows of the (m x m) identity matrix.  It is assumed that i <= p[i] < m.

   Returns
     0        - success
     SING_ERR - x was singular to machine precision.  LU decomposition is
                still computed and returned.
     
   Notes
     Uses partial pivoting.  Does not scale.  An (n x m) matrix x is in row major 
     format if x[i][j]=x[i*n+j] and is in column major format if x[i][j]=x[i+j*m]. 
*/
int bLU(int *p, PRECISION *x, int m, int n)
{
#if defined USE_INLINE
  int a, d, i, j, q, imax, rtrn=0;
 
__asm {
         fninit                                                              // initialize coprocessor              eax:?       ebx:?           ecx:?               edx:?    esi:?  edi:?
         mov     edx,n                                                       // edx=n                               eax:?       ebx:?           ecx:?               edx=n    esi:?  edi:?
         mov     eax,m                                                       // eax=m                               eax=m       ebx:?           ecx:?               edx=n    esi:?  edi:?
         cmp     eax,edx                                                     // compare                             eax=m       ebx:?           ecx:?               edx=n    esi:?  edi:?
         jg      short a1                                                    // jump if m-n > 0                     eax=m       ebx:?           ecx:?               edx=n    esi:?  edi:?

         mov     q,eax                                                       // q=m                                 eax=m       ebx:?           ecx:?               edx=n    esi:?  edi:?
         jmp     short a2                                                    // jump                                eax=m       ebx:?           ecx:?               edx=n    esi:?  edi:?

a1:                                                                          // assumes                             eax=m       ebx:?           ecx:?               edx=n    esi:?  edi:?
         mov     q,edx                                                       // q=n                                 eax=m       ebx:?           ecx:?               edx=n    esi:?  edi:?

a2:                                                                          // assumes                             eax=m       ebx:?           ecx:?               edx=n    esi:?  edi:?
         mov     esi,x                                                       // esi=x                               eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?
         mov     j,0                                                         // j=0                                 eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?

         // Find pivot (j == 0)
         mov     ebx,m                                                       // ebx=m                               eax:?       ebx:i=m         ecx:?               edx=n    esi=x  edi:?
         dec     ebx                                                         // ebx=m-1                             eax:?       ebx:i=m-1       ecx:?               edx=n    esi=x  edi:?
         mov     imax,ebx                                                    // imax=i                              eax:?       ebx:i=m-1       ecx:?               edx=n    esi=x  edi:?

         cmp     ebx,0                                                       // compare                             eax:?       ebx:i           ecx:?               edx=n    esi=x  edi:?
         //jle     short e1                                                    // jump if ebx <= 0                    eax:?       ebx:i           ecx:?               edx=n    esi=x  edi:?
         jle     e1

         mov     ecx,ebx                                                     // ecx=i                               eax:?       ebx:i           ecx=i               edx=n    esi=x  edi:?
         imul    ecx,edx                                                     // ecx=i*n                             eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?
         fld     PRECISION_WORD ptr[esi+PRECISION_SIZE*ecx]                  // st=x[imax*n]                        eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st= x[imax*n]
         fabs                                                                // st=|st|                             eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st=|x[imax*n]|
         dec     ebx                                                         // i--                                 eax:?       ebx:i           ecx=(i+1)*n         edx=n    esi=x  edi:?  st=|x[imax*n]|

b1:                                                                          // assumes                             eax:?       ebx:i           ecx=(i+1)*n         edx=n    esi=x  edi:?  st=|x[imax*n]|
         sub     ecx,edx                                                     // ecx-=n                              eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st=|x[imax*n]|
         fld     PRECISION_WORD ptr[esi+PRECISION_SIZE*ecx]                  // st=x[i*n]                           eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st= x[i*n]              st(1)=|x[imax*n]|
         fabs                                                                // st=|st|                             eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st=|x[i*n]|             st(1)=|x[imax*n]|

         fcom                                                                // compare                             eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st=|x[i*n]|             st(1)=|x[imax*n]|
         fnstsw  ax                                                          // move comparison result              eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st=|x[i*n]|             st(1)=|x[imax*n]|
         sahf                                                                // move comparison result              eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st=|x[i*n]|             st(1)=|x[imax*n]|
         jbe     short b2                                                    // jump if st <= st(1)                 eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st=|x[i*n]|             st(1)=|x[imax*n]|

         ffree   st(1)                                                       // clean up stack                      eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st=|x[i*n]|
         mov     imax,ebx                                                    // imax=i                              eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st=|x[imax*n]|
         jmp     short b3                                                    // jump                                eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st=|x[imax*n]|

b2:                                                                          // assumes                             eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st=|x[i*n]|
         ffree   st(0)                                                       // clean up stack                      eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st=?
         fincstp                                                             // clean up stack                      eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st=|x[imax*n]|

b3:                                                                          // assumes                             eax:?       ebx:i           ecx=i*n             edx=n    esi=x  edi:?  st=|x[imax*n]|
         dec     ebx                                                         // i--                                 eax:?       ebx:i           ecx=(i+1)*n         edx=n    esi=x  edi:?  st=|x[imax*n]|
         jge     short b1                                                    // jump if i >= 0                      eax:?       ebx:i           ecx=(i+1)*n         edx=n    esi=x  edi:?  st=|x[imax*n]|

         ffree   st(0)                                                       // clean up stack                      eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?

         //jmp     short e1                                                    // jump                                eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?
         jmp     e1

         /* Perform stored row operations */
c1:                                                                          // assumes                             eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?
         mov  d,edx                                                          // d=n                                 eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?

         mov  ebx,j                                                          // ebx=j                               eax:?       ebx=j           ecx:?               edx=n    esi=x  edi:?
         add  ebx,edx                                                        // ebx+=n                              eax:?       ebx=j+m         ecx:?               edx=n    esi=x  edi:?
         mov  a,ebx                                                          // a=j+n                               eax:?       ebx=a           ecx:?               edx=n    esi=x  edi:?

         mov  i,1                                                            // i=1                                 eax:?       ebx=a           ecx:?               edx=n    esi=x  edi:?
         mov  ecx,j                                                          // ecx=j                               eax:?       ebx=a           ecx:c=j             edx=n    esi=x  edi:?
         mov  eax,1                                                          // k=1                                 eax:k=i     ebx=a           ecx:c=j             edx=n    esi=x  edi:?

c2:                                                                          // assumes                             eax:k=i     ebx=a           ecx:c=j             edx=n    esi=x  edi:?
         fld  PRECISION_WORD ptr [esi+PRECISION_SIZE*ebx]                    // st=x[a]                             eax:k=i     ebx=a           ecx:c=j             edx=n    esi=x  edi:?  st:x[a]

         mov  ebx,d                                                          // ebx=d                               eax:k=i     ebx:b=d         ecx:c=j             edx=n    esi=x  edi:?  st:x[a]

c3:                                                                          // assumes                             eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[a]
         fld  PRECISION_WORD ptr [esi+PRECISION_SIZE*ebx]                    // st=x[b]                             eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[b]                 st(1):x[a]
         fmul PRECISION_WORD ptr [esi+PRECISION_SIZE*ecx]                    // st=x[c]*x[b]                        eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[c]*x[b]            st(1):x[a]
         fsub                                                                // st=x[a]-x[c]*x[b]                   eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[a]
         inc  ebx                                                            // b++                                 eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[a]
         add  ecx,edx                                                        // c+=n                                eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[a]
         dec  eax                                                            // k--                                 eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[a]
         jg   short c3                                                       // jump if k > 0                       eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[a]

         add  d,edx                                                          // d+=n                                eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[a]
         mov  ecx,j                                                          // ecx=j                               eax:k       ebx:b           ecx:c=j             edx=n    esi=x  edi:?  st:x[a]
         mov  eax,i                                                          // eax=i                               eax:i       ebx:b           ecx:c=j             edx=n    esi=x  edi:?  st;x[a]
         inc  eax                                                            // i++                                 eax:i       ebx:b           ecx:c=j             edx=n    esi=x  edi:?  st;x[a]
         mov  i,eax                                                          // i=eax                               eax:k=i     ebx:b           ecx:c=j             edx=n    esi=x  edi:?  st:x[a]
         cmp  ecx,eax                                                        // compare                             eax:k=i     ebx:b           ecx:c=j             edx=n    esi=x  edi:?  st:x[a]
         jl   short c4                                                       // jump if j-i < 0                     eax:k=i     ebx:b           ecx:c=j             edx=n    esi=x  edi:?  st:x[a]

         mov  ebx,a                                                          // ebx=a                               eax:k=i     ebx:a           ecx:c=j             edx=n    esi=x  edi:?  st:x[a]
         fstp PRECISION_WORD ptr [esi+PRECISION_SIZE*ebx]                    // x[a]=st                             eax:k=i     ebx:a           ecx:c=j             edx=n    esi=x  edi:?
         add  ebx,edx                                                        // a+=n                                eax:k=i     ebx:a           ecx:c=j             edx=n    esi=x  edi:?
         mov  a,ebx                                                          // a=ebx                               eax:k=i     ebx=a           ecx:c=j             edx=n    esi=x  edi:?
         jmp  short c2                                                       // jump                                eax:k=i     ebx=a           ecx:c=j             edx=n    esi=x  edi:?

c4:                                                                          // assumes (a=j*m+j)                   eax:?       ebx:?           ecx:c=j             edx=n    esi=x  edi:?  st:x[a]
         mov  ebx,a                                                          // ebx=a                               eax:?       ebx:a           ecx:c=j             edx=n    esi=x  edi:?  st:x[a]
         fst  PRECISION_WORD ptr [esi+PRECISION_SIZE*ebx]                    // x[a]=st                             eax:?       ebx:a           ecx:c=j             edx=n    esi=x  edi:?  st=x[a]

         // Perform stored row operations and find pivot
         mov  imax,ecx                                                       // imax=j                              eax:?       ebx:a           ecx:c=j             edx=n    esi=x  edi:?  st= x[imax*n+j]

         cmp  m,eax                                                          // compare                             eax:?       ebx:a           ecx:c=j             edx=n    esi=x  edi:?  st= x[imax*n+j]
         jle  short d5                                                       // jump if m-i <= 0                    eax:?       ebx:a           ecx:c=j             edx=n    esi=x  edi:?  st= x[imax*n+j]

         fabs                                                                // st=|st|                             eax:?       ebx:a           ecx:c=j             edx=n    esi=x  edi:?  st=|x[imax*n+j]|

d1:                                                                          // assumes                             eax:?       ebx:a           ecx:c=j             edx=n    esi=x  edi:?  st=|x[imax*n+j]|
         add  ebx,edx                                                        // a+=n                                eax:?       ebx:a           ecx:c=j             edx=n    esi=x  edi:?  st=|x[imax*n+j]|
         mov  a,ebx                                                          // eax=a                               eax:?       ebx=a           ecx:c=j             edx=n    esi=x  edi:?  st=|x[imax*n+j]|
         fld  PRECISION_WORD ptr [esi+PRECISION_SIZE*ebx]                    // st=x[a]                             eax:?       ebx=a           ecx:c=j             edx=n    esi=x  edi:?  st: x[a]                st(1)=|x[imax*n+j]|
         mov  eax,j                                                          // eax=j                               eax:k=j     ebx:b=d         ecx:c=j             edx=n    esi=x  edi:?  st: x[a]                st(1)=|x[imax*n+j]|
         mov  ebx,d                                                          // ebx=d                               eax:k=j     ebx:b=d         ecx:c=j             edx=n    esi=x  edi:?  st: x[a]                st(1)=|x[imax*n+j]|
         mov  ecx,eax                                                        // ecx=j                               eax:k=j     ebx:b=d         ecx:c=j             edx=n    esi=x  edi:?  st: x[a]                st(1)=|x[imax*n+j]|

d2:                                                                          // assumes                             eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st: x[a]                st(1)=|x[imax*n+j]|
         fld  PRECISION_WORD ptr [esi+PRECISION_SIZE*ebx]                    // st=x[b]                             eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st= x[b]                st(1): x[a
         fmul PRECISION_WORD ptr [esi+PRECISION_SIZE*ecx]                    // st=x[c]*x[b]                        eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st= x[c]*x[b]           st(1): x[a]
         fsub                                                                // st=x[a]-x[b]*x[c]                   eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st: x[a]                st(1)=|x[imax*n+j]|
         inc  ebx                                                            // b++                                 eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st: x[a]                st(1)=|x[imax*n+j]|
         add  ecx,edx                                                        // c+=n                                eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st: x[a]                st(1)=|x[imax*n+j]|
         dec  eax                                                            // k--                                 eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st: x[a]                st(1)=|x[imax*n+j]|
         //jg   short d2                                                       // jump if k > 0                       eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st: x[a]                st(1)=|x[imax*n+j]|
         jg   d2

         mov  ebx,a                                                          // ebx=a                               eax:k       ebx=a           ecx:c               edx=n    esi=x  edi:?  st: x[a]                st(1)=|x[imax*n+j]|
         fst  PRECISION_WORD ptr [esi+PRECISION_SIZE*ebx]                    // x[a]=st                             eax:k       ebx=a           ecx:c               edx=n    esi=x  edi:?  st= x[a]                st(1)=|x[imax*n+j]|
         fabs                                                                // st=|st|                             eax:k       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[a]|               st(1)=|x[imax*n+j]|
         fcom                                                                // compare                             eax:k       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[a]|               st(1)=|x[imax*n+j]|
         fnstsw  ax                                                          // move comparison result              eax:k       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[a]|               st(1)=|x[imax*n+j]|
         sahf                                                                // move comparison result              eax:k       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[a]|               st(1)=|x[imax*n+j]|
         jbe     short d3                                                    // jump if st <= st(1)                 eax:k       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[a]|               st(1)=|x[imax*n+j]|

         ffree   st(1)                                                       // clean up stack                      eax:k       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[a]|
         mov     eax,i                                                       // eax=i                               eax=i       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[a]|
         mov     imax,eax                                                    // imax=i (a=i*m+j)                    eax=i       ebx=a           ecx:c               edx=n    esi=x  edi:?  st:|x[imax*n+j]|
         jmp     short d4                                                    // jump                                eax=i       ebx=a           ecx:c               edx=n    esi=x  edi:?  st:|x[imax*n+j]|

d3:                                                                          // assumes                             eax:k       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[a]|               st(1)=|x[imax*n+j]|
         fincstp                                                             // clean up stack                      eax:k       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[imax*n+j]|        st(7)=|x[a]|
         ffree   st(7)                                                       // clean up stack                      eax:k       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[imax*n+j]|

d4:                                                                          // assumes                             eax:?       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[imax*n+j]|
         add     d,edx                                                       // d+=n                                eax:?       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[imax*n+j]|
         mov     eax,i                                                       // eax=i                               eax:i       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[imax*n+j]|
         inc     eax                                                         // i++                                 eax:i       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[imax*n+j]|
         mov     i,eax                                                       // i=eax                               eax=i       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[imax*n+j]|
         cmp     m,eax                                                       // compare                             eax=i       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[imax*n+j]|
         jg      short d1                                                    // jump if m-i > 0                     eax=i       ebx=a           ecx:c               edx=n    esi=x  edi:?  st=|x[imax*n+j]|

d5:                                                                          // assumes                             eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?  st=|x[imax*n+j]|
         ffree   st(0)                                                       // clean up stack                      eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?

         // Interchange rows if necessary
e1:                                                                          // assumes                             eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?
         mov     ebx,j                                                       // ebx=j                               eax:?       ebx=j           ecx:?               edx=n    esi=x  edi:?
         mov     ecx,imax                                                    // ecx=imax                            eax:?       ebx=j           ecx=imax            edx=n    esi=x  edi:?
         mov     edi,p                                                       // edi=p                               eax:?       ebx=j           ecx=imax            edx=n    esi=x  edi=p

         cmp     ebx,ecx                                                     // compare                             eax:?       ebx=j           ecx=imax            edx=n    esi=x  edi=p
         je      short e3                                                    // jump if j-imax = 0                  eax:?       ebx=j           ecx=imax            edx=n    esi=x  edi=p

         mov     dword ptr[edi+4*ebx],ecx                                    // p[j]=imax                           eax:?       ebx=j           ecx=imax            edx=n    esi=x  edi=p

         mov     eax,edx                                                     // k=n                                 eax:k       ebx=j           ecx=imax            edx=n    esi=x  edi=p
         dec     eax                                                         // k--                                 eax:k       ebx=j           ecx=imax            edx=n    esi=x  edi=p
         imul    ebx,edx                                                     // ebx=j*n                             eax:k       ebx=j*n         ecx=imax            edx=n    esi=x  edi=p
         add     ebx,eax                                                     // ebx=j*n+k                           eax:k       ebx=j*n+k       ecx=imax            edx=n    esi=x  edi=p
         imul    ecx,edx                                                     // ecx=imax*n                          eax:k       ebx=j*n+k       ecx=imax*m          edx=n    esi=x  edi=p
         add     ecx,eax                                                     // ecx=imax*n+k                        eax:k       ebx=j*n+k       ecx=imax*m+k        edx=n    esi=x  edi=p

e2:                                                                          // assumes                             eax:k       ebx=j*n+k       ecx=imax*m+k        edx:?    esi=x  edi=?
         mov     edi,dword ptr[esi+PRECISION_SIZE*ecx]
         mov     edx,dword ptr[esi+PRECISION_SIZE*ebx]
         mov     dword ptr[esi+PRECISION_SIZE*ecx],edx
         mov     dword ptr[esi+PRECISION_SIZE*ebx],edi                       // interchange x[j*m+k] and x[imax*m+k]

#if      (PRECISION_SIZE == 8)
         mov     edi,dword ptr[esi+PRECISION_SIZE*ecx+4]
         mov     edx,dword ptr[esi+PRECISION_SIZE*ebx+4]
         mov     dword ptr[esi+PRECISION_SIZE*ecx+4],edx
         mov     dword ptr[esi+PRECISION_SIZE*ebx+4],edi                     // interchange x[j*m+k] and x[imax*m+k]
#endif

         dec     ebx                                                         // ebx--                               eax:k       ebx=j*n+k-1     ecx=imax*m          edx:?    esi=x  edi:?
         dec     ecx                                                         // ecx--                               eax:k       ebx=j*n+k-1     ecx=imax*m+k-1      edx:?    esi=x  edi:?
         dec     eax                                                         // k--                                 eax:k       ebx=j*n+k       ecx=imax*m+k        edx:?    esi=x  edi:?
         jge     short e2                                                    // jump if k >= 0                      eax:k       ebx=j*n+k       ecx=imax*m+k        edx:?    esi=x  edi:?

         mov     edx,n                                                       // edx=n                               eax:k       ebx=j*n+k       ecx=imax*m+k        edx=n    esi=x  edi:?
         jmp     short f1                                                    // jump                                eax:k       ebx=j*n+k       ecx=imax*m+k        edx=n    esi=x  edi:?
e3:                                                                          // assumes                             eax:?       ebx=j           ecx=imax            edx=n    esi=x  edi=p
         mov     dword ptr[edi+4*ebx],ebx                                    // p[j]=j                              eax:?       ebx=j           ecx=imax            edx=n    esi=x  edi=p

         // Is pivot zero?
f1:                                                                          // assumes                             eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?
         mov     ecx,edx                                                     // ecx=n                               eax:?       ebx:?           ecx:a=n             edx=n    esi=x  edi:?
         inc     ecx                                                         // ecx=n+1                             eax:?       ebx:?           ecx:a=n+1           edx=n    esi=x  edi:?
         imul    ecx,j                                                       // ecx=j*n+j                           eax:?       ebx:?           ecx:a=j*n+j         edx=n    esi=x  edi:?

         fld     PRECISION_WORD ptr [esi+PRECISION_SIZE*ecx]                 // st=x[j*m+j]                         eax:?       ebx:?           ecx:a               edx=n    esi=x  edi:?  st=x[j*m+j]
         ftst                                                                // compare                             eax:?       ebx:?           ecx:a               edx=n    esi=x  edi:?  st=x[j*m+j]
         fnstsw  ax                                                          // move comparison result              eax:?       ebx:?           ecx:a               edx=n    esi=x  edi:?  st=x[j*m+j]
         sahf                                                                // move comparison result              eax:?       ebx:?           ecx:a               edx=n    esi=x  edi:?  st=x[j*m+j]
         je      short f3                                                    // jump if st != 0                     eax:?       ebx:?           ecx:a               edx=n    esi=x  edi:?  st=x[j*m+j]

         mov     ebx,m                                                       // ebx=m                               eax:?       ebx=m           ecx:a               edx=n    esi=x  edi:?  st=x[j*m+j]
         imul    ebx,edx                                                     // ebx=m*n                             eax:?       ebx=m*n         ecx:a               edx=n    esi=x  edi:?  st=x[j*m+j]
         add     ecx,edx                                                     // ecx+n                               eax:?       ebx=m*n         ecx:a               edx=n    esi=x  edi:?  st=x[j*m+j]
         cmp     ebx,ecx                                                     // compare                             eax:?       ebx=m*n         ecx:a               edx=n    esi=x  edi:?  st=x[j*m+j]
         jle     short f4                                                    // jump if m*n - a <= 0                eax:?       ebx=m*n         ecx:a               edx=n    esi=x  edi:?  st=x[j*m+j]

         fld1                                                                // st=1.0                              eax:?       ebx=m*n         ecx:a               edx=n    esi=x  edi:?  st=1.0                  st(1)=x[j*m+j]
         fdivr                                                               // st=1.0/x[j*m+j]                     eax:?       ebx=m*n         ecx:a               edx=n    esi=x  edi:?  st=1.0/x[j*m+j]

f2:                                                                          // assumes                             eax:?       ebx=m*n         ecx:a               edx=n    esi=x  edi:?  st=1.0/x[j*m+j]
         fld     st(0)                                                       // st=1.0/x[j*m+j]                     eax:?       ebx=m*n         ecx:a               edx=n    esi=x  edi:?  st=1.0/x[j*m+j]         st(1)=1.0/x[j*m+j]
         fmul    PRECISION_WORD ptr [esi+PRECISION_SIZE*ecx]                 // st=x[a]/x[j*m+j]                    eax:?       ebx=m*n         ecx:a               edx=n    esi=x  edi:?  st=x[a]/x[j*m+j]        st(1)=1.0/x[j*m+j]
         fstp    PRECISION_WORD ptr [esi+PRECISION_SIZE*ecx]                 // x[a]=st                             eax:?       ebx=m*n         ecx:a               edx=n    esi=x  edi:?  st=1.0/x[j*m+j]

         add     ecx,edx                                                     // ecx+=n                              eax:?       ebx=m*n         ecx:a               edx=n    esi=x  edi:?  st=1.0/x[j*m+j]
         cmp     ebx,ecx                                                     // compare                             eax:?       ebx=m*n         ecx:a               edx=n    esi=x  edi:?  st=1.0/x[j*m+j]
         jg      short f2                                                    // jump if m*m - a > 0                 eax:?       ebx=m*n         ecx:a               edx=n    esi=x  edi:?  st=1.0/x[j*m+j]

         jmp     short f4                                                    // jump                                eax:?       ebx=m*n         ecx:a               edx=n    esi=x  edi:?  st=1.0/x[j*m+j]

f3:                                                                          // assumes                             eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?  st=*
         mov     rtrn,SING_ERR                                               // rtrn=SING_ERR                       eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?  st=*

f4:                                                                          // assumes                             eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?  st=*
         ffree   st(0)                                                       // clean up stack                      eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?

         inc     j                                                           // j++                                 eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?
         mov     eax,q                                                       // eax=q                               eax=q       ebx:?           ecx:?               edx=n    esi=x  edi:?
         cmp     eax,j                                                       // compare                             eax=q       ebx:?           ecx:?               edx=n    esi=x  edi:?
         jg      c1                                                          // jump if q-j > 0                     eax=q       ebx:?           ecx:?               edx=n    esi=x  edi:?

         // Perform stored row operations
g1:                                                                          // assumes                             eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?
         cmp  edx,j                                                          // compare                             eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?
         jle  short g5                                                       // jump if n-j <= 0                    eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?

         mov  d,edx                                                          // d=n                                 eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?

         mov  ebx,j                                                          // ebx=j                               eax:?       ebx=j           ecx:?               edx=n    esi=x  edi:?
         add  ebx,edx                                                        // ebx+=n                              eax:?       ebx=j+n         ecx:?               edx=n    esi=x  edi:?
         mov  a,ebx                                                          // a=j+n                               eax:?       ebx=a           ecx:?               edx=n    esi=x  edi:?

         mov  i,1                                                            // i=1                                 eax:?       ebx=a           ecx:?               edx=n    esi=x  edi:?
         mov  eax,1                                                          // k=1                                 eax:k=i     ebx=a           ecx:?               edx=n    esi=x  edi:?

g2:                                                                          // assumes                             eax:k=i     ebx=a           ecx:?               edx=n    esi=x  edi:?
         cmp  m,eax                                                          // compare                             eax:k=i     ebx=a           ecx:?               edx=n    esi=x  edi:?
         jle  short g4                                                       // jump if m-i <= 0                    eax:k=i     ebx=a           ecx:?               edx=n    esi=x  edi:?

         fld  PRECISION_WORD ptr [esi+PRECISION_SIZE*ebx]                    // st=x[a]                             eax:k=i     ebx=a           ecx:?               edx=n    esi=x  edi:?  st:x[a]

         mov  ecx,j                                                          // ecx=j                               eax:k=i     ebx=a           ecx:c=j             edx=n    esi=x  edi:?
         mov  ebx,d                                                          // ebx=d                               eax:k=i     ebx:b=d         ecx:c=j             edx=n    esi=x  edi:?  st:x[a]

g3:                                                                          // assumes                             eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[a]
         fld  PRECISION_WORD ptr [esi+PRECISION_SIZE*ebx]                    // st=x[b]                             eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[b]                 st(1):x[a]
         fmul PRECISION_WORD ptr [esi+PRECISION_SIZE*ecx]                    // st=x[c]*x[b]                        eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[c]*x[b]            st(1):x[a]
         fsub                                                                // st=x[a]-x[c]*x[b]                   eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[a]
         inc  ebx                                                            // b++                                 eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[a]
         add  ecx,edx                                                        // c+=n                                eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[a]
         dec  eax                                                            // k--                                 eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[a]
         jg   short g3                                                       // jump if k > 0                       eax:k       ebx:b           ecx:c               edx=n    esi=x  edi:?  st:x[a]

         mov  ebx,a                                                          // ebx=a                               eax:?       ebx:a           ecx:?               edx=n    esi=x  edi:?  st:x[a]
         fstp PRECISION_WORD ptr [esi+PRECISION_SIZE*ebx]                    // x[a]=st                             eax:?       ebx:a           ecx:?               edx=n    esi=x  edi:?

         add  ebx,edx                                                        // a+=n                                eax:?       ebx:a           ecx:?               edx=n    esi=x  edi:?
         mov  a,ebx                                                          // a=ebx                               eax:?       ebx=a           ecx:?               edx=n    esi=x  edi:?
         add  d,edx                                                          // d+=n                                eax:?       ebx=a           ecx:?               edx=n    esi=x  edi:?
         mov  eax,i                                                          // eax=i                               eax:i       ebx=a           ecx:?               edx=n    esi=x  edi:?
         inc  eax                                                            // i++                                 eax:i       ebx=a           ecx:?               edx=n    esi=x  edi:?
         mov  i,eax                                                          // i=eax                               eax:k=i     ebx=a           ecx:?               edx=n    esi=x  edi:?
         jmp  short g2                                                       // jump                                eax:k=i     ebx=a           ecx:?               edx=n    esi=x  edi:?

g4:                                                                          // assumes                             eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?
         inc  j                                                              // j++                                 eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?
         jmp  short g1                                                       // jump                                eax:?       ebx:?           ecx:?               edx=n    esi=x  edi:?

g5:                                                                          // assumes                             eax:?       ebx:?           ecx:?               edx:?    esi:?  edi:?
      }
 return rtrn;
#elif defined USE_BLAS_LAPACK
 PRECISION *y;
 int i, info;
 if (!( y=(PRECISION*)malloc(m*n*sizeof(PRECISION))))
  return MEM_ERR;
 bTranspose(y,x,n,m); 

#if PRECISION_SIZE == 4
 sgetrf(&m,&n,y,&m,p,&info);
#else
 dgetrf(&m,&n,y,&m,p,&info);
#endif

 bTranspose(x,y,m,n);
 free(y);
 for (i=(m < n) ? m-1 : n-1; i >= 0; i--) p[i]--;
 return (info < 0) ? SING_ERR : 0;

#else
 int a, b, c, d, i, j, k, q=(m < n) ? m : n, imax, rtrn=0;
 PRECISION big, tmp;
 
 for (j=0; j < q; j++)
  {
   if (j == 0)
     {
      /* Find pivot */
      for (big=fabs(x[imax=0]), a=n, i=1; i < m; a+=n, i++)
       if (fabs(x[a]) > big)
        {
         big=fabs(x[a]);
         imax=i;
        }
     }
    else
     {
      /* Perform stored row operations */
      for (d=n, a=j+n, i=1; i <= j; d+=n, a+=n, i++)
       {
        tmp=x[a];
        for (b=d, c=j, k=i; k > 0; b++, c+=n, k--) tmp-=x[b]*x[c];
        x[a]=tmp;
       }

      /* Perform stored row operations and find pivot */
      for (big=fabs(tmp), imax=j; i < m; d+=n, a+=n, i++)
       {
        tmp=x[a];
        for (b=d, c=j, k=j; k > 0; b++, c+=n, k--) tmp-=x[b]*x[c];
        x[a]=tmp;
        if (fabs(tmp) > big)
         {
          big=fabs(tmp);
          imax=i;
         }
       }
     }

   /* Interchange rows if necessary */
   if (j != imax)
     {
      p[j]=imax;

      for (k=n-1, a=imax*n+k, b=j*n+k; k >= 0; a--, b--, k--)
       {
        tmp=x[a];
        x[a]=x[b];
        x[b]=tmp;
       }
     }
    else
     p[j]=j;

   /* Is pivot zero? */
   if (x[a=j*n+j] != 0.0)
     for (tmp=1.0/x[a], a+=n, i=m*n; a < i; a+=n) x[a]*=tmp;
    else
     rtrn=SING_ERR;
  }

 /* Perform stored row operations */
 for ( ; j < n; j++)
  for (d=n, a=j+n, i=1; i < m; d+=n, a+=n, i++)
   {
    tmp=x[a];
    for (b=d, c=j, k=i; k > 0; b++, c+=n, k--) tmp-=x[b]*x[c];
    x[a]=tmp;
   }
 
 return rtrn;
 #endif
}

/*
   Assumes
     x : array of length mm representing a triangular matrix.
     b : array of length mn

   Results
                         x
                      (m x m)      b/y
        u   xt   bt  row major  row major    solve
        ----------------------------------------------------------
        0    0    0      L        m x n    L * y = b
        1    0    0      U        m x n    U * y = b
        0    1    0      L        m x n    L'* y = b  
        1    1    0      U        m x n    U'* y = b
        0    0    1      L        n x m    L * y'= b'  (y * L'= b)
        1    0    1      U        n x m    U * y'= b'  (y * U'= b)
        0    1    1      L        n x m    L'* y'= b'  (y * L = b)
        1    1    1      U        n x m    U'* y'= b'  (y * U = b)

      or
                           x           b/y
        u   xt   bt     (m x m)      (m x n)      solve
        --------------------------------------------------
        0    0    0   L:row major   row major   L * y = b
        1    0    0   U:row major   row major   U * y = b
        0    1    0   U:col major   row major   U * y = b  
        1    1    0   L:col major   row major   L * y = b
        0    0    1   L:row major   col major   L * y = b  
        1    0    1   U:row major   col major   U * y = b  
        0    1    1   U:col major   col major   U * y = b  
        1    1    1   L:col major   col major   L * y = b  
     
     The solution y is stored in b.

   Returns
     0 (NO_ERR) - success
     SING_ERR   - x is singular

   Notes
     Because this routines tests using the xor operator, the values of u and xt
     must be either 0 or 1.  The matrix x is assumed to be triangular.  Care 
     must be taken that the matrix is either upper or lower triangular in the
     correct format.
      
     An (n x m) matrix x is in row major format if x[i][j]=x[i*n+j] and is in
     column major format if x[i][j]=x[i+j*m]. 
*/
int bSolveTriangular(PRECISION *x, PRECISION *b, int m, int n, int u, int xt, int bt)
{
#ifndef USE_INLINE
 int i, j, k, bi, bj, xi, xj, mbi;
 PRECISION *pxx, *px, *pb;
 for (j=m+1, i=j*(m-1); i >= 0; i-=j) if (x[i] == 0.0) return SING_ERR;
 if (xt) { xi=1; xj=m; } else { xi=m; xj=1; }
 if (bt) { bi=1; bj=m; } else { bi=n; bj=1; }
 mbi=(m-1)*bi;
 if (u ^ xt)
   for (x+=(m-1)*xj, j=(n-1)*bj, b+=j; j >= 0; b-=bj, j-=bj)
    for (pxx=x+(m-1)*xi, i=(m-1)*bi; i >= 0; pxx-=xi, i-=bi)  
     {
      px=pxx;
      pb=b+i;
      for (k=mbi-i; k > 0; px-=xj, k-=bi) (*pb)-=(*px)*pb[k];
      *pb/=(*px);
     }
  else
   {
    for (j=(n-1)*bj, b+=j; j >= 0; b-=bj, j-=bj)
     for (pxx=x, i=0; i <= mbi; pxx+=xi, i+=bi)  
      {
       px=pxx;
       pb=b+i;
       for (k=-i; k < 0; px+=xj, k+=bi) (*pb)-=(*px)*pb[k];
       *pb/=(*px);
      }
   }
 return 0;
#else
 int i, j, bi, bj, xi, xj, mbi;
 PRECISION *pxx;
 for (j=m+1, i=m*m-1; i >= 0; i-=j) if (x[i] == 0.0) return SING_ERR;
 if (xt) { xi=1; xj=m; } else { xi=m; xj=1; }
 if (bt) { bi=1; bj=m; } else { bi=n; bj=1; }
 mbi=(m-1)*bi;
 if (u ^ xt)
   for (x+=(m-1)*xj, j=(n-1)*bj, b+=j; j >= 0; b-=bj, j-=bj)
    for (pxx=x+(m-1)*xi, i=(m-1)*bi; i >= 0; pxx-=xi, i-=bi) 
     __asm {
            fninit
            mov  esi,pxx
            mov  eax,i
            mov  edi,eax
            shl  edi,PRECISION_SHIFT
            add  edi,b

            fld  PRECISION_WORD ptr [edi]                      //     st=*pb

            sub  eax,mbi
            neg  eax                                       
            je   short dest2_1                                 //     (k=mbi-i) > 0

            mov  ebx,xj
            shl  ebx,PRECISION_SHIFT
            mov  ecx,bi

dest1_1:    fld  PRECISION_WORD ptr [esi]
            fmul PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]
            fsub                                               //     st-=(*px)*pb[k]
         
            sub  esi,ebx                                       //     px-=bi
         
            sub  eax,ecx                                       
            jg   short dest1_1                                 //     (k-=bi) > 0  

dest2_1:    fld  PRECISION_WORD ptr [esi]
            fdiv 
            fstp PRECISION_WORD ptr [edi]                      //     *pb=st/(*px)  
           } 
  else
   for (j=(n-1)*bj, b+=j; j >= 0; b-=bj, j-=bj)
    for (pxx=x, i=0; i <= mbi; pxx+=xi, i+=bi)  
     __asm {
            fninit
            mov  esi,pxx
            mov  eax,i
            mov  edi,eax
            shl  edi,PRECISION_SHIFT
            add  edi,b 

            fld  PRECISION_WORD ptr [edi]                      //     st=*pb
                                
            neg  eax
            je   short dest2_0                                 //     (k=-i) < 0

            mov  ebx,xj
            shl  ebx,PRECISION_SHIFT
            mov  ecx,bi

dest1_0:    fld  PRECISION_WORD ptr [esi]
            fmul PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]
            fsub                                               //     st-=(*px)*pb[k]
         
            add  esi,ebx                                       //     px+=bi
         
            add  eax,ecx                                       
            jl   short dest1_0                                 //     (k+=bi) < 0  

dest2_0:    fld  PRECISION_WORD ptr [esi]
            fdiv 
            fstp PRECISION_WORD ptr [edi]                      //     *pb=st/(*px)  
           }
  return 0;
#endif
}

/*
   Assumes
     x : array of length mm representing a triangular matrix with unit diagonal
     b : array of length mn

   Results
                         x
                      (m x m)      b/y
        u   xt   bt  row major  row major    solve
        ----------------------------------------------------------
        0    0    0      L        m x n    L * y = b
        1    0    0      U        m x n    U * y = b
        0    1    0      L        m x n    L'* y = b  
        1    1    0      U        m x n    U'* y = b
        0    0    1      L        n x m    L * y'= b'  (y * L'= b)
        1    0    1      U        n x m    U * y'= b'  (y * U'= b)
        0    1    1      L        n x m    L'* y'= b'  (y * L = b)
        1    1    1      U        n x m    U'* y'= b'  (y * U = b)

      or
                           x           b/y
        u   xt   bt     (m x m)      (m x n)      solve
        --------------------------------------------------
        0    0    0   L:row major   row major   L * y = b
        1    0    0   U:row major   row major   U * y = b
        0    1    0   U:col major   row major   U * y = b  
        1    1    0   L:col major   row major   L * y = b
        0    0    1   L:row major   col major   L * y = b  
        1    0    1   U:row major   col major   U * y = b  
        0    1    1   U:col major   col major   U * y = b  
        1    1    1   L:col major   col major   L * y = b  
     
     The solution y is stored in b.

   Returns
     0 (NO_ERR) - success

   Notes
     Because this routines tests using the xor operator, the values of u and xt
     must be either 0 or 1.  The matrix x is assumed to be triangular with unit
     diagonal.  Care  must be taken that the matrix is either upper or lower 
     triangular in the correct format.
      
     An (n x m) matrix x is in row major format if x[i][j]=x[i*n+j] and is in
     column major format if x[i][j]=x[i+j*m]. 
*/
int bSolveUnitTriangular(PRECISION *x, PRECISION *b, int m, int n, int u, int xt, int bt)
{
 #ifndef USE_INLINE
 int i, j, k, bi, bj, xi, xj, mbi;
 PRECISION *pxx, *px, *pb;
 if (xt) { xi=1; xj=m; } else { xi=m; xj=1; }
 if (bt) { bi=1; bj=m; } else { bi=n; bj=1; }
 mbi=(m-1)*bi;
 if (u ^ xt)
   for (x+=(m-1)*xj, j=(n-1)*bj, b+=j; j >= 0; b-=bj, j-=bj)
    for (pxx=x+(m-1)*xi, i=(m-1)*bi; i >= 0; pxx-=xi, i-=bi)  
     {
      px=pxx;
      pb=b+i;
      for (k=mbi-i; k > 0; px-=xj, k-=bi) (*pb)-=(*px)*pb[k];
     }
  else
   {
    for (j=(n-1)*bj, b+=j; j >= 0; b-=bj, j-=bj)
     for (pxx=x, i=0; i <= mbi; pxx+=xi, i+=bi)  
      {
       px=pxx;
       pb=b+i;
       for (k=-i; k < 0; px+=xj, k+=bi) (*pb)-=(*px)*pb[k];
      }
   }
#else
 int i, j, bi, bj, xi, xj, mbi;
 PRECISION *pxx;
 if (xt) { xi=1; xj=m; } else { xi=m; xj=1; }
 if (bt) { bi=1; bj=m; } else { bi=n; bj=1; }
 mbi=(m-1)*bi;
 if (u ^ xt)
   for (x+=(m-1)*xj, j=(n-1)*bj, b+=j; j >= 0; b-=bj, j-=bj)
    for (pxx=x+(m-1)*xi, i=(m-1)*bi; i >= 0; pxx-=xi, i-=bi) 
     __asm {                                                   //                       eax    ebx    ecx    edx    esi    edi     st(0)     st(1) 
            mov  eax,i                                         // k=i                    k
            mov  edi,eax                                       // pb=i                   k                                  pb                                              
            sub  eax,mbi                                       // k=i-mbi                k                                  pb                                                  
            neg  eax                                           // k=mbi-i                k                                  pb                                               
            je   short a2                                      // jump if k > 0          k                                  pb

            shl  edi,PRECISION_SHIFT                           // pb=i*PRECISION_SIZE    k                                  pb                                                                                                
            add  edi,b                                         // pb=b+i*PRECISION_SIZE  k                                  pb                                                             
            mov  esi,pxx                                       // px=pxx                 k                           px     pb                                                                                                 
            mov  ebx,xj                                        // ebx=xj                 k      xj                   px     pb                                                                                   
            shl  ebx,PRECISION_SHIFT                           // ebx=xj*PRECISION_SIZE  k      xj                   px     pb                                               
            mov  ecx,bi                                        // ecx=bi                 k      xj     bi            px     pb                                                    

            fld  PRECISION_WORD ptr [edi]                      // st=pb[0]               k      xj     bi            px     pb     pb[0]

a1:                                                            // assumes                k      xj     bi            px     pb     pb[0]
            fld  PRECISION_WORD ptr [esi]                      // st=*px                 k      xj     bi            px     pb      *px      pb[0]                                                                          
            fmul PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]   // st=pb[k]*(*px)         k      xj     bi            px     pb   pb[k]*(*px) pb[0]                                                                                                 
            fsub                                               // st=pb[0]-(*px)*pb[k]   k      xj     bi            px     pb     pb[0]                                                        
         
            sub  esi,ebx                                       // px-=xj                 k      xj     bi            px     pb     pb[0]
         
            sub  eax,ecx                                       // k-=bi                  k      xj     bi            px     pb     pb[0]                                                                                    
            jg   short a1                                      // jump if k > 0          k      xj     bi            px     pb     pb[0]

            fstp PRECISION_WORD ptr [edi]                      // pb[0]=st               k      xj     bi            px     pb    
a2:
           } 
  else
   for (j=(n-1)*bj, b+=j; j >= 0; b-=bj, j-=bj)
    for (pxx=x, i=0; i <= mbi; pxx+=xi, i+=bi)  
     __asm {
            mov  eax,i
            mov  edi,eax
            neg  eax
            je   short dest2_0                                 // (k=-i) < 0
                               
            mov  esi,pxx                                        
            shl  edi,PRECISION_SHIFT
            add  edi,b 
            mov  ebx,xj
            shl  ebx,PRECISION_SHIFT
            mov  ecx,bi

            fld  PRECISION_WORD ptr [edi]                      // st=*pb

dest1_0:    fld  PRECISION_WORD ptr [esi]
            fmul PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]
            fsub                                               // st-=(*px)*pb[k]
         
            add  esi,ebx                                       // px+=bi
         
            add  eax,ecx                                       
            jl   short dest1_0                                 // (k+=bi) < 0  

            fstp PRECISION_WORD ptr [edi]                      // *pb=st  
dest2_0:
           }
#endif
 return 0;
}

/*
   Assumes
     p : integer array of length q  with 0 <= p[i] < m for all 0 <= i < q.
     y : array of length mn

   Results
                    x/y
        pt   yt  row major    product
        ---------------------------------------------
         0    0    m x n     x = P * y
         1    0    m x n     x = P'* y
         0    1    n x m     x'= P * y'  (x = y * P')
         1    1    n x m     x'= P'* y'  (x = y * P )

     or
                    x/y
        pt   yt   (m x n)     product
        -----------------------------------------------
         0    0  row major   x = P * y
         1    0  row major   x = P'* y
         0    1  col major   x = P * y  
         1    1  col major   x = P'* y  

     The matrix P is defined by

                    P = P(0,p[0])*P(1,p[1])*...*P(q-1,p[q])

     where P(r,s) is the (m x m) matrix obtained by permuting the rth and sth 
     rows of the (m x m) identity matrix.
     
   Notes:
     An (n x m) matrix x is in row major format if x[i][j]=x[i*n+j] and is in
     column major format if x[i][j]=x[i+j*m]. 
*/
int bPermutationMultiply(int *p, PRECISION *y, int m, int n, int q, int pt, int yt)
{
 int i, j, k, pk;
 PRECISION tmp;
 if (yt)
   if (pt)
     for (j=0; j < q; j++)
      {
       if (j != p[j])
        for (i=(n-1)*m; i >= 0; i-=m)
         {
          tmp=y[i+j];
          y[i+j]=y[i+p[j]];
          y[i+p[j]]=tmp;
         }
      }
    else
     for (j=q-1; j >= 0; j--)
      {
       if (j != p[j])
        for (i=(n-1)*m; i >= 0; i-=m)
         {
          tmp=y[i+j];
          y[i+j]=y[i+p[j]];
          y[i+p[j]]=tmp;
         }
      }
  else
   if (pt)
     for (i=0; i < q; i++)
      {
       if (i != p[i])
        {
         k=i*n;
         pk=p[i]*n;
         for (j=n-1; j >= 0; j--)
          {
           tmp=y[k+j];
           y[k+j]=y[pk+j];
           y[pk+j]=tmp;
          }
        }
      }
    else
     for (i=q-1; i >= 0; i--)
      {
       if (i != p[i])
        {
         k=i*n;
         pk=p[i]*n;
         for (j=n-1; j >= 0; j--) 
          {
           tmp=y[k+j];
           y[k+j]=y[pk+j];
           y[pk+j]=tmp;
          }
        }
      }
 return 0;
}

/*
   Assumes
     p : integer array of length q with i <= p[i] < m for all 0 <= i < m.
     x : m x m matrix in row major format

   Results
     If t is equal to zero, then x = P and x = P' otherwise.  The matrix P 
     is defined by

                    P = P(0,p[0])*P(1,p[1])*...*P(q-1,p[q-1])

     where P(r,s) is the (m x m) matrix obtained by permuting the rth and sth 
     rows of the (m x m) identity matrix.

   Notes:
     An (n x m) matrix x is in row major format if x[i][j]=x[i*n+j] and is in
     column major format if x[i][j]=x[i+j*m]. 
*/
int bPermutation(PRECISION *x, int *p, int m, int q, int t)
{
 int i, j, k;
 for (k=m*m-1; k >= 0; k--) x[k]=0.0;
 if (!t)
   for (j=m-1; j >= 0; j--)
    {
     if (j >= q)
       {
	k=q-1;
	i=j;
       }
      else
       {
	k=j-1;
	i=p[j];
       }
     for ( ; k >= 0; k--) if (i == p[k]) i=k;
     x[i*m+j]=1.0;
    }
 else
   for (j=m-1; j >= 0; j--)
    {
     if (j >= q)
       {
	k=q-1;
	i=j;
       }
      else
       {
	k=j-1;
	i=p[j];
       }
     for ( ; k >= 0; k--) if (i == p[k]) i=k;
     x[j*m+i]=1.0;
    }
 return 0;
}
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/


/******************************************************************************/
/************************ Singular Value Decomposition ************************/
/******************************************************************************/
/*
   results
     returns sqrt(a^2 + b^2)
*/
static PRECISION pythag(PRECISION a, PRECISION b)
{
 PRECISION absa=fabs(a), absb=fabs(b), quotient;

 if (absa > absb)
  {
   quotient=absb/absa;
   return absa*sqrt(1.0+quotient*quotient);
  }

 if (absb == 0) return 0.0;

 quotient=absa/absb;
 return absb*sqrt(1.0+quotient*quotient);
}

/*
   Assumes
     U : m x n matrix in row major format
     d : n-vector
     V : n x n matrix in row major format

   Returns
     0 (NO_ERR) : success
     MEM_ERR    : out of memory
     ITER_ERR   : maximum number of iterations (MAX_ITER) exceeded

   Results
     Finds U, V and d such that U = U * diag(d) * V'.  U and V are orthogonal
     matrices and the elemets of d are non-negative.  U is orthogonal in the
     sense that U'*U and U*U' are diagonal matrices with ones and zeros along
     the diagonal.

   Notes
     Based on Numerical Recipes in C routines.
*/
#define MAX_ITER 100
int bSVD(PRECISION *U, PRECISION *d, PRECISION *V, int m, int n)
{
 int flag, i, its, j, jj, k, l, nm;
 PRECISION  anorm, c, f, g, h, s, scale, x, y, z, *rv1, tmp;

 rv1=(PRECISION*)malloc(n*sizeof(PRECISION));
 if (!rv1) return MEM_ERR;
 g=scale=anorm=0.0;

 for (i=0; i < n; i++)
  {
   l=i+1;
   rv1[i]=scale*g;
   g=s=scale=0.0;

   if (i < m)
    {
     for (k=i; k < m; k++) scale+=fabs(U[k*n+i]);

     if (scale)
      {
       for (k=i; k < m; k++)
        {
         U[k*n+i]/=scale;
         s+=U[k*n+i]*U[k*n+i];
        }

       f=U[i*n+i];
       g=(f >= 0) ? -sqrt(s) : sqrt(s);
       h=f*g-s;
       U[i*n+i]=f-g;

       for (j=l; j < n; j++)
        {
         for (s=0.0, k=i; k < m; k++) s+=U[k*n+i]*U[k*n+j];
         f=s/h;
         for (k=i; k < m; k++) U[k*n+j]+=f*U[k*n+i];
        }

       for (k=i; k < m; k++) U[k*n+i]*=scale;
      }
    }

   d[i]=scale *g;
   g=s=scale=0.0;
   if (i < m && i != n-1)
    {
     for (k=l; k < n; k++) scale+=fabs(U[i*n+k]);

     if (scale)
      {
       for (k=l; k < n; k++)
        {
         U[i*n+k]/=scale;
         s+=U[i*n+k]*U[i*n+k];
        }

       f=U[i*n+l];
       g=(f >= 0) ? -sqrt(s) : sqrt(s);
       h=f*g-s;
       U[i*n+l]=f-g;

       for (k=l; k < n; k++) rv1[k]=U[i*n+k]/h;
       for (j=l; j < m; j++)
        {
         for (s=0.0, k=l; k < n; k++) s+=U[j*n+k]*U[i*n+k];
         for (k=l; k < n; k++) U[j*n+k]+=s*rv1[k];
        }

       for (k=l; k < n; k++) U[i*n+k]*=scale;
      }
    }

   if (anorm < (tmp=fabs(d[i])+fabs(rv1[i]))) anorm=tmp;
  }

 for (i=n-1; i >= 0; i--)
  {
   if (i < n-1)
    {
     if (g)
      {
       for (j=l; j < n; j++) V[j*n+i]=(U[i*n+j]/U[i*n+l])/g;
       for (j=l; j < n; j++)
        {
         for (s=0.0, k=l; k < n; k++) s+=U[i*n+k]*V[k*n+j];
         for (k=l; k < n; k++) V[k*n+j]+=s*V[k*n+i];
        }
      }

     for (j=l; j < n; j++) V[i*n+j]=V[j*n+i]=0.0;
    }

    V[i*n+i]=1.0;
   g=rv1[i];
   l=i;
  }

 for (i=(m < n) ? m-1 : n-1; i >= 0; i--)
  {
   l=i+1;
   g=d[i];

   for (j=l; j < n; j++) U[i*n+j]=0.0;

   if (g)
     {
      g=1.0/g;

      for (j=l; j < n; j++)
       {
        for (s=0.0, k=l; k < m; k++) s+=U[k*n+i]*U[k*n+j];
        f=(s/U[i*n+i])*g;
        for (k=i; k < m; k++) U[k*n+j]+=f*U[k*n+i];
       }

      for (j=i; j < m; j++) U[j*n+i]*=g;
     }
    else
     for (j=i; j < m; j++) U[j*n+i]=0.0;

   ++U[i*n+i];
  }

 for (k=n-1; k >= 0; k--)
  {
   for (its=1; its <= 30; its++)
    {
     flag=1;
     for (l=k; l >= 0; l--)
      {
       nm=l-1;

       if ((PRECISION)(fabs(rv1[l])+anorm) == anorm)
        {
         flag=0;
         break;
        }

       if ((PRECISION)(fabs(d[nm])+anorm) == anorm) break;
      }

     if (flag)
      {
       c=0.0;
       s=1.0;

       for (i=l; i <= k; i++)
        {
         f=s*rv1[i];
         rv1[i]=c*rv1[i];

         if ((PRECISION)(fabs(f)+anorm) == anorm) break;

         g=d[i];
         h=pythag(f,g);
         d[i]=h;
         h=1.0/h;
         c=g*h;
         s=-f*h;

         for (j=0; j < m; j++)
          {
           y=U[j*n+nm];
           z=U[j*n+i];
           U[j*n+nm]=y*c+z*s;
           U[j*n+i]=z*c-y*s;
          }
        }
      }

     z=d[k];

     if (l == k)
      {
       if (z < 0.0)
        {
         d[k]=-z;
         for (j=0; j < n; j++) V[j*n+k]=-V[j*n+k];
        }

       break;
      }

     if (its >= MAX_ITER)
      {
       free(rv1);
       return ITER_ERR;
      }

     x=d[l];
     nm=k-1;
     y=d[nm];
     g=rv1[nm];
     h=rv1[k];
     f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
     g=pythag(f,1.0);
     f=((x-z)*(x+z)+h*((y/(f+((f >= 0.0) ? fabs(g) : -fabs(g))))-h))/x;
     c=s=1.0;

     for (j=l; j <= nm; j++)
      {
       i=j+1;
       g=rv1[i];
       y=d[i];
       h=s*g;
       g=c*g;
       z=pythag(f,h);
       rv1[j]=z;
       c=f/z;
       s=h/z;
       f=x*c+g*s;
       g=g*c-x*s;
       h=y*s;
       y*=c;

       for (jj=0; jj < n; jj++)
        {
         x=V[jj*n+j];
         z=V[jj*n+i];
         V[jj*n+j]=x*c+z*s;
         V[jj*n+i]=z*c-x*s;
        }

       z=pythag(f,h);
       d[j]=z;

       if (z)
        {
         z=1.0/z;
         c=f*z;
         s=h*z;
        }

       f=c*g+s*y;
       x=c*y-s*g;

      for (jj=0; jj < m; jj++)
       {
        y=U[jj*n+j];
        z=U[jj*n+i];
        U[jj*n+j]=y*c+z*s;
        U[jj*n+i]=z*c-y*s;
       }
    }

   rv1[l]=0.0;
   rv1[k]=f;
   d[k]=x;
  }
 }

 free(rv1);

 return 0;
}
#undef MAX_ITER
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

/******************************************************************************/
/***************************** QR Decompositions ******************************/
/******************************************************************************/
/*
   Assumes
     Q : m x m matrix in row major format or null pointer
     R : m x n matrix in row major format

   Returns
     0 (NO_ERR) : Success
     MEM_ERR    : Out of memory

   Results
     Finds an orthogonal matrix Q and an upper triangular matrix U
     such that

                                R = Q * U
     
     The matrix U is returned in R and Q is computed only if it is
     not null.

   Notes
     The QR decomposition is formed using Householder matrices without
     pivoting.  If Q is null, then the matrix Q will have the property 
     that 

                             det(Q) = (-1)^s

     where s=min{m-1,n).
*/
int bQR(PRECISION *Q, PRECISION *R, int m, int n)
{
 int i, j, k, s=(m <= n) ? m-1 : n;
 PRECISION tmp, scale, sigma, c, sum;
 PRECISION *diag, *norm;

 if (Q)
   {
    if (!(diag=(PRECISION*)malloc(2*s*sizeof(PRECISION)))) return MEM_ERR;
    norm=diag+s;

    for (k=0; k < s; k++)
     {
      for (scale=0.0, i=k; i < m; i++)
       if ((tmp=fabs(R[i*n+k])) > scale) scale=tmp;

      if (scale == 0.0)
       {
        diag[k]=norm[k]=0.0;
        continue;
       }
   
      for (scale=1.0/scale, sigma=0.0, i=k; i < m; i++)
       {
        R[i*n+k]*=scale;
        sigma+=R[i*n+k]*R[i*n+k];
       }

      sigma=sqrt(sigma);

      if (R[k*n+k] < 0.0) sigma=-sigma;

      diag[k]=R[k*n+k]+=sigma;

      norm[k]=c=1.0/(sigma*diag[k]);

      for (j=k+1; j < n; j++)
       {
        for (sum=0.0, i=k; i < m; i++) sum+=R[i*n+k]*R[i*n+j];
        sum*=c;
        for (i=k; i < m; i++) R[i*n+j]-=R[i*n+k]*sum;
       }

      R[k*n+k]=-sigma/scale;
     }

    for (i=m*m-1; i >= 0; i--) Q[i]=0.0;
    for (i=m*m-1; i >= 0; i-=m+1) Q[i]=1.0;

    for (k=s-1; k >= 0; k--)
     {
      c=norm[k];
      for (j=k; j < m; j++)
       {
        for (sum=diag[k]*Q[k*m+j], i=k+1; i < m; i++) sum+=R[i*n+k]*Q[i*m+j];
        for (Q[k*m+j]-=diag[k]*(sum*=c), i=k+1; i < m; i++) Q[i*m+j]-=R[i*n+k]*sum;
       }
     }

    for (k=s-1; k >= 0; k--)
     for (i=k+1; i < m; i++)
      R[i*n+k]=0;

    free(diag);

    return 0;
   }
  else
   {
    for (k=0; k < s; k++)
     {
      /* Find largest element below diagonal in kth column */
      for (scale=0.0, i=m-1; i > k; i--)
       if ((tmp=fabs(R[i*n+k])) > scale) scale=tmp;

      if (scale == 0.0)
        for (j=k; j < n; j++) R[k*n+j]=-R[k*n+j];
       else
        {
         if ((tmp=fabs(R[k*n+k])) > scale) scale=tmp;

         /* Compute normalized sum of squares */
         for (scale=1.0/scale, sigma=0.0, i=k; i < m; i++)
          {
           R[i*n+k]*=scale;
           sigma+=R[i*n+k]*R[i*n+k];
          }

         sigma=sqrt(sigma);

         /* choose sign of sigma to match the diagonal elements sign */
         if (R[k*n+k] < 0.0) sigma=-sigma;

         /* c = sigma*(diagonal + sigma) */
         R[k*n+k]+=sigma;
         c=1.0/(sigma*R[k*n+k]);

         for (j=k+1; j < n; j++)
          {
           for (sum=0.0, i=k; i < m; i++) sum+=R[i*n+k]*R[i*n+j];
           sum*=c;
           for (i=k; i < m; i++) R[i*n+j]-=R[i*n+k]*sum;
          }

         R[k*n+k]=-sigma/scale;
         for (i=k+1; i < m; i++) R[i*n+k]=0;
        }
     }

    return 0;
   }
}
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

/******************************************************************************/
/************************** Cholesky Decompositions ***************************/
/******************************************************************************/
/*
   Assumes
     X : m x m symmetric matrix in either row or column major format

   Returns
     0 (NO_ERR)  : success
     POS_DEF_ERR : X not positive definite

   Results
                      T 
        u    t    row major     solves
        ----------------------------------
        0    0        L        X = L'* L  
        1    0        U        X = U'* U
        0    1        L        X = L * L'
        1    1        U        X = U * U'

      or
                           
        u    t        T         solves
        ----------------------------------
        0    0   L:row major   X = T'* T
        1    0   U:row major   X = T'* T
        0    1   U:col major   X = T'* T 
        1    1   L:col major   X = T'* T
       


     Upon successful exit T is upper triangular with positive diagonal and 
     satisfies X = T' * T.  T overwrites X.

   Notes
     Failure usually indicates X is not positive definite.  
*/
int bCholesky(PRECISION *X, int m, int u, int t)
{
#ifndef USE_INLINE
 int i, j, k, b;
 PRECISION scale, *pX, *pXi, *pXj;

 if (u)
   if (t)
     for (i=m-1, pXi=X+i*m; i >= 0; pXi-=m, i--)
      {
       for (j=m-1, pX=X+i+j*m; j > i; pX-=m, j--) *pX=0.0;

       for (k=i+1; k < m; k++) *pX-=pXi[k]*pXi[k];
  
       if (*pX <= 0.0) return POSDEF_ERR;
       scale=1.0/(*pX=sqrt(*pX));

       pXj=pXi;
       for (j--; j >= 0; j--)
        {
         pX-=m;
         pXj-=m;
         for (k=i+1; k < m; k++) *pX-=pXi[k]*pXj[k];
         *pX*=scale;
        }
      }
    else
     for (i=0, pXi=X; i < m; pXi++, i++)
      {
       for (j=0, pX=X+i*m; j < i; pX++, j++) *pX=0.0;

       for (k=(i-1)*m; k >= 0; k-=m) *pX-=pXi[k]*pXi[k];

       if (*pX <= 0.0) return POSDEF_ERR;
       scale=1.0/(*pX=sqrt(*pX));

       pXj=pXi;
       for (j++; j < m; j++)
        {
         pXj++;
         pX++;        
         for (k=(i-1)*m; k >= 0; k-=m) *pX-=pXi[k]*pXj[k]; 
         *pX*=scale;
        }
      }
  else
   if (t)
    for (i=0, pXi=X; i < m; pXi+=m, i++)
      {
       for (j=0, pX=X+i; j < i; pX+=m, j++) *pX=0.0;
       
       for (k=i-1; k >= 0; k--) *pX-=pXi[k]*pXi[k];

       if (*pX <= 0.0) return POSDEF_ERR;
       scale=1.0/(*pX=sqrt(*pX));

       pXj=pXi;
       for (j++; j < m; j++)
        {
         pX+=m;
         pXj+=m;
         for (k=i-1; k >= 0; k--) *pX-=pXi[k]*pXj[k];
         *pX*=scale;
        }
      }
    else
     for (b=m*m, i=m-1, pXi=X+i; i >= 0; pXi--, i--)
      {     
       for (j=m-1, pX=X+i*m+j; j > i; pX--, j--) *pX=0.0;

       for (k=(i+1)*m; k < b; k+=m) *pX-=pXi[k]*pXi[k];

       if (*pX <= 0.0) return POSDEF_ERR;
       scale=1.0/(*pX=sqrt(*pX));

       pXj=pXi;
       for (j--; j >= 0; j--)
        {
         pXj--;
         pX--;
         for (k=(i+1)*m; k < b; k+=m)  *pX-=pXi[k]*pXj[k];
         *pX*=scale;
        }
      }
 return 0;
#else
 int i, b, rtrn=0;
 __asm{                                                       //    eax        ebx      ecx        edx        esi        edi          st(0)          st(1)          st(2) 
        fninit
        mov    edx,m                                          //     ?          ?        ?          m          ?          ?                        

        cmp    u,0
        //je     short c0                                       //     ?          ?        ?          m          ?          ?
        je     c0

        cmp    t,0
        //je     short b0                                       //     ?          ?        ?          m          ?          ?
        je     b0
 /********************************************************************************************************************************/
                                                              //     ?         ?         ?          m          ?          ?
        mov    eax,edx
        dec    eax                                            //     i         ?         ?          m          ?          ?     

        mov    edi,edx
        imul   edi,eax
        shl    edi,PRECISION_SHIFT
        add    edi,X                                          //     i         ?         ?          m          ?        X+i*m       

a1:                                                           //     i         ?         ?          m          ?        X+i*m 
        mov    i,eax

        mov    ebx,edx
        dec    ebx                                            //     i         j         ?          m          ?        X+i*m

        mov    esi,edx
        imul   esi,ebx
        add    esi,eax
        shl    esi,PRECISION_SHIFT
        add    esi,X                                          //     i         j         ?          m       X+i+j*m     X+i*m
       
a2:                                                           //     i         j         ?          m       X+i+j*m     X+i*m
        cmp    ebx,eax
        jle    short a3

        mov    dword ptr[esi],0

#if (PRECISION_SIZE == 8)
        mov    dword ptr[esi+4],0
#endif

        mov    ecx,m                                          //     i         j         m          m       X+i+j*m     X+i*m
        shl    ecx,PRECISION_SHIFT                            //     i         j  m*PRECISION_SIZE  m       X+i+j*m     X+i*m
        sub    esi,ecx                                        //     i         j  m*PRECISION_SIZE  m       X+i+j*m     X+i*m
        dec    ebx                                            //     i         j         ?          m       X+i+j*m     X+i*m
        jmp    short a2                                       //     i         j         ?          m       X+i+j*m     X+i*m

a3:
        fld    PRECISION_WORD ptr[esi]                        //     i         j         ?          m       X+i+j*m     X+i*m       X[i+j*m]

a4:                                                           //     k         j         ?          m       X+i+j*m     X+i*m       X[i+j*m]
        inc    eax                                            //     k         j         ?          m       X+i+j*m     X+i*m       X[i+j*m]
        cmp    edx,eax
        jle    short a5

        fld    PRECISION_WORD ptr[edi+PRECISION_SIZE*eax]     //     k         j         ?          m       X+i+j*m     X+i*m        X[i*m+k]      X[i+j*m]
        fld    st(0)                                          //     k         j         ?          m       X+i+j*m     X+i*m        X[i*m+k]      X[i*m+k]       X[i+j*m]
        fmul                                                  //     k         j         ?          m       X+i+j*m     X+i*m       X[i*m+k]^2     X[i+j*m]
        fsub                                                  //     k         j         ?          m       X+i+j*m     X+i*m        X[i+j*m]

        jmp    short a4                                       //     k         j         ?          m       X+i+j*m     X+i*m        X[i+j*m]

a5:                                                           //     ?         j         ?          m       X+i+j*m     X+i*m        X[i+j*m]
        ftst
        fnstsw  ax                                               
        sahf                                                      
        //jbe     short e1                                      //     ?         j         ?          m       X+i+j*m     X+i*m        X[i+j*m]
        jbe     e1

        fsqrt                                                 //     ?         j         ?          m       X+i+j*m     X+i*m        X[i+j*m] 
        fst     PRECISION_WORD ptr[esi]                       //     ?         j         ?          m       X+i+j*m     X+i*m        X[i+j*m] 
        
        fld1                                                  //     ?         j         ?          m       X+i+j*m     X+i*m            1           
        fdivr                                                 //     ?         j         ?          m       X+i+j*m     X+i*m         scale 

        mov    ecx,edi                                        //     ?         j       X+j*m        m       X+i+j*m     X+i*m         scale

a6:                                                           //     ?         j       X+j*m        m       X+i+j*m     X+i*m         scale
        dec    ebx                                            //     ?         j       X+j*m        m       X+i+j*m     X+i*m         scale
        jl     short a9                                       //     ?         j       X+j*m        m       X+i+j*m     X+i*m         scale

        mov    eax,m                                          //     m         j       X+j*m        m       X+i+j*m     X+i*m         scale
        shl    eax,PRECISION_SHIFT                        // m*PRECISION_SIZE  j       X+j*m        m       X+i+j*m     X+i*m         scale
        sub    ecx,eax                                    // m*PRECISION_SIZE  j       X+j*m        m       X+i+j*m     X+i*m         scale
        sub    esi,eax                                        //     ?         j       X+j*m        m       X+i+j*m     X+i*m         scale

        fld    PRECISION_WORD ptr[esi]                        //     ?         j       X+j*m        m       X+i+j*m     X+i*m        X[i+j*m]        scale

        mov    eax,i                                          //     k         j       X+j*m        m       X+i+j*m     X+i*m        X[i+j*m]        scale

a7:                                                           //     k         j       X+j*m        m       X+i+j*m     X+i*m        X[i+j*m]        scale
        inc    eax                                            
        cmp    edx,eax
        jle    short a8

        fld    PRECISION_WORD ptr[edi+PRECISION_SIZE*eax]     //     k         j       X+j*m        m       X+i+j*m     X+i*m        X[i*m+k]       X[i+j*m]        scale
        fmul   PRECISION_WORD ptr[ecx+PRECISION_SIZE*eax]     //     k         j       X+j*m        m       X+i+j*m     X+i*m   X[i*m+k]*X[j*m+k]   X[i+j*m]        scale
        fsub                                                  //     k         j       X+j*m        m       X+i+j*m     X+i*m        X[i+j*m]        scale

        jmp    short a7                                       //     k         j       X+j*m        m       X+i+j*m     X+i*m        X[i+j*m]        scale

a8:
        fmul   st,st(1)                                       //     k         j       X+j*m        m       X+i+j*m     X+i*m        X[i+j*m]        scale
        fstp   PRECISION_WORD ptr[esi]                        //     k         j       X+j*m        m       X+i+j*m     X+i*m         scale
        jmp    short a6                                       //     k         j       X+j*m        m       X+i+j*m     X+i*m         scale

a9:     ffree  st(0)                                          //     ?         ?         ?          m       X+i+j*m     X+i*m

        mov    eax,edx
        shl    eax,PRECISION_SHIFT
        sub    edi,eax                                        //     ?         ?         ?          m       X+i+j*m     X+i*m

        mov    eax,i                                          //     i         ?         ?          m       X+i+j*m     X+i*m
        dec    eax                                            //     i         ?         ?          m       X+i+j*m     X+i*m
        //jge    short a1                                       //     i         ?         ?          m       X+i+j*m     X+i*m
        jge    a1

        //jmp    short e2                                       //     ?         ?         ?          ?          ?          ?
        jmp    e2
/********************************************************************************************************************************/ 
b0:                                                           //     ?          ?        ?          m          ?          ?                        
        mov    eax,0

        mov    edi,X                                          //     i                              m                    X+i

b1:                                                           //     i                              m                    X+i
        mov    i,eax

        mov    ebx,0                                          //     i         j                    m       X+i*m+j      X+i

        mov    esi,eax
        imul   esi,edx
        shl    esi,PRECISION_SHIFT
        add    esi,X                                          //     i         j                    m       X+i*m+j      X+i

b2:       
        cmp    eax,ebx
        jle    short b3

        mov    dword ptr[esi],0

#if (PRECISION_SIZE == 8)
        mov    dword ptr[esi+4],0
#endif

        add    esi,PRECISION_SIZE
        inc    ebx
        jmp    short b2

b3:
        fld    PRECISION_WORD ptr[esi]
        dec    eax                                            
        imul   eax,edx                                        //     k         j                    m       X+i+j*m     X+i

        cmp    eax,0
        jl     short b5

b4:
        fld    PRECISION_WORD ptr[edi+PRECISION_SIZE*eax]
        fld    st(0)
        fmul
        fsub

        sub    eax,edx
        jge    short b4

b5:
        ftst
        fnstsw  ax                                               
        sahf
        //jbe     short e1
        jbe     e1

        fsqrt
        fst     PRECISION_WORD ptr[esi]
        
        fld1
        fdivr  

        mov    ecx,edi                                        //               j        X+j         m       X+i*m+j     X+i

b6:                                                           //               j        X+j         m       X+i*m+j     X+i
        inc    ebx
        cmp    edx,ebx
        jle    short b9

        add    ecx,PRECISION_SIZE
        add    esi,PRECISION_SIZE

        fld    PRECISION_WORD ptr[esi]

        mov    eax,i                                          
        dec    eax
        imul   eax,edx                                        //     k         j       X+j*m        m       X+i+j*m     X+i*m

        cmp    eax,0
        jl     short b8

b7:
        fld    PRECISION_WORD ptr[edi+PRECISION_SIZE*eax]
        fmul   PRECISION_WORD ptr[ecx+PRECISION_SIZE*eax]
        fsub

        sub    eax,edx                                            
        jge    short b7

b8:
        fmul   st,st(1)
        fstp   PRECISION_WORD ptr[esi]
        jmp    short b6


b9:     ffree   st(0) 

        add     edi,PRECISION_SIZE

        mov     eax,i
        inc     eax
        cmp     edx,eax
        //jg      short b1
        jg      b1

        //jmp     short e2
        jmp     e2
/********************************************************************************************************************************/  
c0:
        mov    eax,t
        cmp    eax,0
        //je     short d0
        je     d0
/********************************************************************************************************************************/ 
        mov    eax,0                                          //     i                              m                   X+i*m

        mov    edi,X                                          //                                    m                   X+i*m

c1:                                                           //     i                              m                   X+i*m
        mov    i,eax

        mov    ebx,0                                          //     i         j                    m       X+i+j*m     X+i*m
       
        mov    esi,eax
        shl    esi,PRECISION_SHIFT
        add    esi,X                                          //     i                              m       X+i+j*m     X+i*m

c2:       
        cmp    eax,ebx
        jle    short c3

        mov    dword ptr[esi],0

#if (PRECISION_SIZE == 8)
        mov    dword ptr[esi+4],0
#endif

        mov    ecx,edx
        shl    ecx,PRECISION_SHIFT
        add    esi,ecx
        inc    ebx
        jmp    short c2

c3:
        fld    PRECISION_WORD ptr[esi]
c4:                                                           //     k         j                    m       X+i+j*m    X+i*m
        dec    eax                                            
        jl     short c5

        fld    PRECISION_WORD ptr[edi+PRECISION_SIZE*eax]
        fld    st(0)
        fmul
        fsub

        jmp    short c4

c5:
        ftst
        fnstsw  ax                                               
        sahf                                                      
        //jbe     short e1
        jbe     e1

        fsqrt
        fst     PRECISION_WORD ptr[esi]
        
        fld1
        fdivr  

        mov    ecx,edi                                        //               j        X+j         m       X+i+j*m    X+i*m

c6:                                                           //               j        X+j         m       X+i+j*m    X+i*m
        inc    ebx
        cmp    edx,ebx
        jle    short c9

        mov    eax,edx
        shl    eax,PRECISION_SHIFT
        add    ecx,eax
        add    esi,eax

        fld    PRECISION_WORD ptr[esi]

        mov    eax,i
c7:                                                          //     k         j         X+j        m       X+i+j*m    X+i*m                      
        dec    eax                                           
        jl     short c8

        fld    PRECISION_WORD ptr[edi+PRECISION_SIZE*eax]
        fmul   PRECISION_WORD ptr[ecx+PRECISION_SIZE*eax]
        fsub

        jmp    short c7

c8:
        fmul   st,st(1)
        fstp   PRECISION_WORD ptr[esi]
        jmp    short c6

c9:     ffree   st(0) 

        mov    eax,edx
        shl    eax,PRECISION_SHIFT
        add    edi,eax

        mov    eax,i
        inc    eax
        cmp    edx,eax
        //jg     short c1
        jg     c1

        //jmp    short e2
        jmp    e2
/********************************************************************************************************************************/ 
d0:                                                           //     ?          ?        ?          m          ?          ?
        mov    eax,edx
        imul   eax,edx
        mov    b,eax

        mov    eax,edx
        dec    eax                                            //     i         ?         ?          m       X+i*m+j      X+i         

        mov    edi,eax
        shl    edi,PRECISION_SHIFT
        add    edi,X                                          //     ?         ?         ?          m          ?         X+i

        mov    esi,edx
        imul   esi,edx
        shl    esi,PRECISION_SHIFT
        add    esi,X                                          //     ?         j         ?          m       X+i*m+j      X+i   
                
d1:                                                           //     i         ?         ?          m       X+i*m+j      X+i 
        mov    i,eax
        
        mov    ebx,edx                                        //     i         j         ?          m       X+i*m+j      X+i
        dec    ebx

        sub    esi,PRECISION_SIZE
               
d2:                                                           //     i         j         ?          m       X+i*m+j      X+i
        cmp    ebx,eax
        jle    short d3

        mov    dword ptr[esi],0

#if (PRECISION_SIZE == 8)
        mov    dword ptr[esi+4],0
#endif

        sub    esi,PRECISION_SIZE                             //     i         j         ?          m       X+i*m+j      X+i
        dec    ebx
        jmp    short d2                                       //     i         j         ?          m       X+i*m+j      X+i

d3:
        fld    PRECISION_WORD ptr[esi]                        //     i         j         ?          m       X+i*m+j      X+i        X[i+j*m]

        imul   eax,edx
   
d4:                                                           //     k         j         ?          m       X+i*m+j      X+i        X[i+j*m]
        add    eax,edx                                        //     k         j         ?          m       X+i*m+j      X+i        X[i+j*m]
        cmp    b,eax
        jle    short d5

        fld    PRECISION_WORD ptr[edi+PRECISION_SIZE*eax]     //     k         j         ?          m       X+i*m+j      X+i         X[i*m+k]      X[i+j*m]
        fld    st(0)                                          //     k         j         ?          m       X+i*m+j      X+i         X[i*m+k]      X[i*m+k]       X[i+j*m]
        fmul                                                  //     k         j         ?          m       X+i*m+j      X+i        X[i*m+k]^2     X[i+j*m]
        fsub                                                  //     k         j         ?          m       X+i*m+j      X+i         X[i+j*m]

        jmp    short d4                                       //     k         j         ?          m       X+i*m+j      X+i         X[i+j*m]
 
d5:                                                           //     ?         j         ?          m       X+i*m+j      X+i         X[i+j*m]
        ftst                                                    
        fnstsw  ax                                               
        sahf                                                      
        jbe     short e1                                      //     ?         j         ?          m       X+i*m+j      X+i         X[i+j*m] 
  
        fsqrt                                                 //     ?         j         ?          m       X+i*m+j      X+i         X[i+j*m] 
        fst     PRECISION_WORD ptr[esi]                       //     ?         j         ?          m       X+i*m+j      X+i         X[i+j*m] 
        
        fld1                                                  //     ?         j         ?          m       X+i*m+j      X+i             1           
        fdivr                                                 //     ?         j         ?          m       X+i*m+j      X+i          scale 

        mov    ecx,edi                                        //     ?         j        X+j         m       X+i*m+j      X+i          scale

d6:                                                           //     ?         j        X+j         m       X+i*m+j      X+i          scale
        dec    ebx                                            //     ?         j        X+j         m       X+i*m+j      X+i          scale
        jl     short d9                                       //     ?         j        X+j         m       X+i*m+j      X+i          scale

        sub    ecx,PRECISION_SIZE                             //     ?         j        X+j         m       X+i*m+j      X+i          scale
        sub    esi,PRECISION_SIZE                             //     ?         j        X+j         m       X+i*m+j      X+i          scale

        fld    PRECISION_WORD ptr[esi]                        //     ?         j        X+j         m       X+i*m+j      X+i         X[i+j*m]        scale

        mov    eax,i                                          //     k         j        X+j         m       X+i*m+j      X+i         X[i+j*m]        scale
        imul   eax,edx

d7:                                                           //     k         j        X+j         m       X+i*m+j      X+i         X[i+j*m]        scale
        add    eax,edx                                            
        cmp    b,eax
        jle    short d8

        fld    PRECISION_WORD ptr[edi+PRECISION_SIZE*eax]     //     k         j        X+j         m       X+i*m+j      X+i         X[i*m+k]       X[i+j*m]        scale
        fmul   PRECISION_WORD ptr[ecx+PRECISION_SIZE*eax]     //     k         j        X+j         m       X+i*m+j      X+i    X[i*m+k]*X[j*m+k]   X[i+j*m]        scale
        fsub                                                  //     k         j        X+j         m       X+i*m+j      X+i         X[i+j*m]        scale

        jmp    short d7                                       //     k         j        X+j         m       X+i*m+j      X+i         X[i+j*m]        scale

d8:
        fmul   st,st(1)                                       //     k         j        X+j         m       X+i*m+j      X+i         X[i+j*m]        scale
        fstp   PRECISION_WORD ptr[esi]                        //     k         j        X+j         m       X+i*m+j      X+i          scale
        jmp    short d6                                       //     k         j        X+j         m       X+i*m+j      X+i          scale

d9:     ffree  st(0)                                          //     ?         ?         ?          m       X+i*m+j      X+i

        sub    edi,PRECISION_SIZE

        mov    eax,i                                          //     i         ?         ?          m       X+i*m+j      X+i
        dec    eax                                            //     i         ?         ?          m       X+i*m+j      X+i
        jge    short d1                                       //     i         ?         ?          m       X+i*m+j      X+i

        jmp    short e2 
/********************************************************************************************************************************/ 
e1:     mov     rtrn,POSDEF_ERR                 
e2:
       }
 return rtrn;
#endif
 
}
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

/*******************************************************************************

  Accumulation:
   u is an m x p matrix
   v is a p x n matrix
   0 <= i < m
   0 <= j < n
   ui=p  uj=1  (u is in row major format)
   ui=1  uj=m  (u is in column major format)
   vi=n  vj=1  (v is in row major format)
   vi=1  vj=p  (v is in column major format)  
   
   Computes the following sum:

     u[i*ui + 0*uj]*v[0*vi + j*vj] + u[i*ui+1*uj]*v[1*vi+j*vj] + ...

                                   ... + u[i*ui + (p-1)*uj]*v[(p-1)*vi + j*vj]
      

  *** C code ***

  int k=(p-1)*vi;
  PRECISION *pu=u + i*ui + (p-1)*uj;
  PRECISION *pv=v + j*vj;

  PRECISION tmp=(*pu)*pv[k];
  while ((k-=vi) >= 0) tmp+=(*(pu-=uj))*pv[k];
  w(i,j)=tmp;

  *** assembly code *** 

  __asm {

         
         // Upon entry:  
         //   eax=(p-1)*kv                           
         //   ebx=ku*PRECISION_SIZE  
         //   ecx=kv  
         //   esi=u(i) + (p-1)*uj*PRECISION_SIZE
         //   edi=v(j)
         // 
         // Upon exit:   
         //   eax=0         
         //   ebx=ku*PRECISION_SIZE  
         //   ecx=kv  
         //   esi=u(i)
         //   edi=v(j)   
             
         fld  PRECISION_WORD ptr [esi]
         fmul PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]      // tmp=(*pu)*v(j)[k]

         sub  eax,ecx                                          // k-=kv
         jl   short dest2                                      // k >= 0

dest1  : sub  esi,ebx                                          // pu-=ku

         fld  PRECISION_WORD ptr [esi]
         fmul PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]      
         fadd                                                  // tmp+=(*pu)*v(j)[k]

         sub  eax,ecx                                          // k-=kv
         jge  short dest1_0                                    // k >= 0

dest2  : fstp PRECISION_WORD ptr [edx]                         // w(i,j)=tmp
        }

 Notes:
  The register edx is unused.  It is recommended that this register be used to store a pointer
  to the storage position for the accumlated value.  Furthermore, the direction of outer loops 
  should be such that edx-PRECISION_SIZE is the storage position for the next accumlated value.
   
  If vi is known to be one, then ecx is available.

  If uj is known to be one, then ebx is available and u+(i-1)*ui+(p-1)*uj = u+i*ui-1.  If u is
  a matrix in row major format, then uj will be one. 

*******************************************************************************/


/*******************************************************************************
  Back solving
   Given u(i), v(j), ku, kv, and p, computes

    v(j)[0]=(u(i)[(p-1)*ku]*v(j)[(p-1)*kv] + ... + u(i)[ku]*v(j)[kv])/u(i)[0]

   p is assumed to be positive and ku and kv are positive.
  
  

  *** C code ***

  int k=(p-1)*kv;
  PRECISION *pu=u(i)+(p-1)*ku;

  PRECISION tmp=(*pu)*v(j)[k];
  while (k != 0) tmp+=(*(pu-=ku))*v(j)[k-=kv];

  *** assembly code *** 

  __asm {

         
         // Upon entry:  
         //   eax=(p-1)*kv  
         //   ebx=ku*PRECISION_SIZE  
         //   ecx=kv  
         //   esi=u(i) + (p-1)*uj*PRECISION_SIZE
         //   edi=v(j)
         // 
         // Upon exit:   
         //   eax=0         
         //   ebx=ku*PRECISION_SIZE  
         //   ecx=kv  
         //   esi=u(i)
         //   edi=v(j)   
             
         fld  PRECISION_WORD ptr [esi]
         fmul PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]

         test eax,eax
         je   short dest2

dest1:   sub  eax,ecx                                       // if kv=1, replace with:  dec  eax 
         sub  esi,ebx                                       // if ku=1, replace with:  sub  esi,PRECISION_SIZE

         fld  PRECISION_WORD ptr [esi]
         fmul PRECISION_WORD ptr [edi+PRECISION_SIZE*eax]
         fadd

         test eax,eax
         jne  dest1

dest2:   fstp PRECISION_WORD ptr [edx]                      // pop accumulated value off stack and store
       
         sub  edx,PRECISION_SIZE                          

        }

 Notes:
  The register edx is unused.  It is recommended that this register be used to store a pointer
  to the storage position for the accumlated value.  Furthermore, the direction of outer loops 
  should be such that edx-PRECISION_SIZE is the storage position for the next accumlated value.
   
  If vi is known to be one, then ecx is available.

  If uj is known to be one, then ebx is available and u+(i-1)*ui+(p-1)*uj = u+i*ui-1.  If u is
  a matrix in row major format, then uj will be one. 

*******************************************************************************/






