
#include <math.h>
#include "exp.h"
#include "rand.h"

/*
   Returns a deviate drawn from the distribution with density function
   proportional to

                   f(x) = (|x|^T)*exp(-0.5*T*(x-c)^2)

   Uses rejection method with

                          w(0)                  w(1)
          g(x) = -------------------- + --------------------
                  1 + ((x-b(0))/a)^2     1 + ((x-b(1))/a)^2

   where a=sqrt(2/T), b(0)=(c+sqrt(c^2+4))/2, b(1)=(c-sqrt(c^2+4))/2,
   and w(i)=f(b(i)).  Drawing from the distribution with density proportional 
   to g can be accomplished by drawing from the distribution with density
   given by

                                     1
                  h(x) = --------------------------
                          a*pi(1 + ((x-b(i))/a)^2)

  with probability w(i)/(w(0)+w(1)).  If u is a uniformly distributed random
  variable on (0,1), then b(i)+a*tan(PI*u) has density equal to h.

  Note:  If u = sqrt(T)x, then the density of u is proportional to

               f(x) = (|u|^T)*exp(-0.5*(u-sqrt(T)c)^2)

*/
#define PI 3.14159265358979323846
PRECISION Simulate(int T, PRECISION c)
{
 PRECISION a, b0, b1, log_w0, w1_w0, b0_b1, w, x, t, s;

 a=sqrt(2.0/(double)T);

 b0_b1=sqrt(c*c+4);
 if (c < 0)
   {
    b0=0.5*(c-b0_b1);
    b1=0.5*(c+b0_b1);
    b0_b1/=-a;                                      // (b0-b1)/a
   }
  else
   {
    b0=0.5*(c+b0_b1);
    b1=0.5*(c-b0_b1);
    b0_b1/=a;                                       // (b0-b1)/a
   }

 log_w0=T*(log(fabs(b0))-0.5*b1*b1);                // log(w0)=log(f(b0))

 w1_w0=exp(T*(log(fabs(b1))-0.5*b0*b0)-log_w0);     // w1/w0=f(b1)/f(b0)

 w=1.0/(1.0+w1_w0);                                 // w0/(w0+w1)=f(b0)/(f(b0)+f(b1));

 do
  {
   s=tan(PI*uniform());
   if (uniform() < w)
     {
      x=b0+a*s;
      t=s+b0_b1;                                    // (x-b1)/a
      s=1.0/(1.0+s*s)+w1_w0/(1.0+t*t);
     }
    else
     {
      x=b1+a*s;
      t=s-b0_b1;                                    // (x-b0)/a
      s=1.0/(1.0+t*t)+w1_w0/(1.0+s*s);
     }
   t=x-c;
  }
 while (uniform()*s > exp(T*(log(fabs(x))-0.5*t*t)-log_w0));

 return x;
}
#undef PI

