
#include "matrix.h"
#include "bmatrix.h"

/******************************************************************************/
/******************************* Error Functions*******************************/
/******************************************************************************/
static int MATRIX_ERROR=0;
static int MATRIX_ERROR_TERMINATE=ALL_ERRORS;
static int MATRIX_ERROR_VERBOSE=ALL_ERRORS;

int GetMatrixError(void)
{
 return MATRIX_ERROR;
}

void ClearMatrixError(void)
{
 MATRIX_ERROR=0;
}

int SetMatrixErrorTerminate(int err)
{
 int terminate=MATRIX_ERROR_TERMINATE;
 MATRIX_ERROR_TERMINATE=err & ALL_ERRORS;
 return terminate;
}

int SetMatrixErrorVerbose(int err)
{
 int verbose=MATRIX_ERROR_VERBOSE;
 MATRIX_ERROR_VERBOSE=err & ALL_ERRORS;
 return verbose;
}

void MatrixError(int err)
{
 if (!(MATRIX_ERROR=err & KNOWN_ERRORS)) MATRIX_ERROR=UNKNOWN_ERR;
 switch (MATRIX_ERROR_VERBOSE & MATRIX_ERROR)
  {
   case MEM_ERR:
     printf("\nOut of memory.\n");
    break;
   case SIZE_ERR:
    printf("\nMatrices/vectors not conformable.\n");
    break;
   case NULL_ERR:
     printf("\nUnexpected null matrix encountered.\n");
    break;
   case SING_ERR:
    printf("\nSingular matrix.\n");
    break;
   case ITER_ERR:
    printf("\nMaximum iteration limit exceeded.\n");
    break;
   case POSDEF_ERR:
    printf("\nMatrix not positive definite.\n");
    break;
   case FLOAT_ERR:
    printf("\nFloating point error.\n");
    break;
   case UNKNOWN_ERR:
    printf("\nUnknown error.\n");
    break;
   case 0:
    break;
   default:
    printf("\nNon-supported verbose error\n");
    break;
  }
 if (MATRIX_ERROR_TERMINATE & MATRIX_ERROR) exit(err);
}
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

/******************************************************************************/
/**************************** Allocation Routines *****************************/
/******************************************************************************/
/*
   Allocates memory for a m-dimensional vector.  On success, returns a pointer
   to the vector.  On failure, calls MatrixError() and returns null.
*/
TVector CreateVector(int m)
{
 TVector x;
 if (m <= 0) 
  {
   MatrixError(SIZE_ERR); 
   return (TVector)NULL; 
  }
 if (x=(TVector)((int *)malloc(m*sizeof(PRECISION)+sizeof(int))+1))
   V_DIM(x)=m;
  else
   MatrixError(MEM_ERR);
 return x;
}

/*
   Assumes
     x : valid vector or null pointer

   Results
     Deallocates memory for vector.  If x is null pointer, calls MatrixError().

   Notes
     The vector x MUST have been previously allocated with a call to
     CreateVector() or be null.
*/
void FreeVector(TVector x)
{
 if (x) free((int *)x-1);
}

/*
   Allocates memory for (m x n) matrix.  On success, returns a pointer to the
   matrix.  On failure, calls MatrixError() and returns null.
*/
TMatrix CreateMatrix(int m, int n)
{
 TMatrix X;
 int i;
 if ((m <= 0) || (n <= 0))
  { 
   MatrixError(SIZE_ERR); 
   return (TMatrix)NULL; 
  }
 if (X=(TMatrix)((int *)malloc(m*(sizeof(PRECISION *)+n*sizeof(PRECISION))+2*sizeof(int))+2))
   {
    M_ROW(X)=m;
    M_COL(X)=n;
    X[0]=(PRECISION *)(X+m);
    for (i=1; i < m; i++) X[i]=X[i-1]+n;
   }
  else
   MatrixError(MEM_ERR);
 return X;
}

/*
  Assumes
     X : valid matrix or null pointer

   Results
     Deallocates memory for matrix.  If X is null pointer, calls MatrixError().

   Notes
     The matrix X MUST have been previously allocated with a call to
     CreateMatrix() or be null.
*/
void FreeMatrix(TMatrix X)
{
 if (X) free((int *)X-2);
}
/******************************************************************************/
/******************************************************************************/

/******************************************************************************/
/************************** Initialization Routines ***************************/
/******************************************************************************/
/*
   Assumes
     x : m vector

   Results
     x[i] = c

   Returns
     x
*/
TVector InitializeVector(TVector x, PRECISION c)
{
 int i;
 if (!x)
   MatrixError(NULL_ERR);
  else
   for (i=V_DIM(x); --i >= 0; x[i]=c);
 return x;
}

/*
   Assumes
     X : m x n matrix

   Results
     X[i][j] = c

   Returns
     X
*/
TMatrix InitializeMatrix(TMatrix X, PRECISION c)
{
 int i;
 PRECISION *pX;
 if (!X)
   MatrixError(NULL_ERR);
  else
   for (pX=X[0], i=M_ROW(X)*M_COL(X)-1; i >= 0; i--) pX[i]=c;
 return X;
}
/******************************************************************************/
/******************************************************************************/


/******************************************************************************/
/**************************** Assignment Routines *****************************/
/******************************************************************************/
/*
   Assumes
     x : m-vector or null pointer
     y : m-vector

   Results
     x = y.  If x is null pointer, x is created.

   Returns
     Returns x upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.
*/
TVector EquateVector(TVector x, TVector y)
{
 if (!y) 
  {
   MatrixError(NULL_ERR); 
   return (TVector)NULL; 
  }
 if (!x)
   { 
    if (!(x=CreateVector(V_DIM(y)))) 
     return (TVector)NULL; 
   }
  else
   if (V_DIM(x) != V_DIM(y))
    { 
     MatrixError(SIZE_ERR); 
     return (TVector)NULL; 
    }
 memcpy(x,y,V_DIM(y)*sizeof(PRECISION));
 return x;
}

/*
   Assumes
     X : m x n matrix or null pointer
     Y : m x n matrix

   Results
     X = Y.  If X is null pointer, X is created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.
*/
TMatrix EquateMatrix(TMatrix X, TMatrix Y)
{
 if (!Y) 
  {
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (!X)
   { 
    if (!(X=CreateMatrix(M_ROW(Y),M_COL(Y)))) 
     return (TMatrix)NULL; 
   }
  else
   if ((M_ROW(Y) != M_ROW(X)) || (M_COL(Y) != M_COL(X)))
    {
     MatrixError(SIZE_ERR); 
     return (TMatrix)NULL; 
    }
 memcpy(X[0],Y[0],M_ROW(Y)*M_COL(Y)*sizeof(PRECISION));
 return X;
}

/*
   Assumes
     X : m x m matrix or null pointer

   Results
     X is set to the m x m identity matrix.  If X is null pointer, X 
     is created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.
*/
TMatrix IdentityMatrix(TMatrix X, int m)
{
 int i;
 PRECISION *pX;
 if (!X)
   { 
    if (!(X=CreateMatrix(m,m))) 
    return (TMatrix)NULL; 
   }
  else
   if ((m != M_ROW(X)) || (m != M_COL(X)))
    { 
     MatrixError(SIZE_ERR); 
     return (TMatrix)NULL; 
    }
 for (pX=X[0], i=m*m-1; i >= 0; i--) pX[i]=0.0;
 for (i=m-1; i >= 0; i--) X[i][i]=1.0;
 return X;
}

/*
   Assumes
     X : m x m matrix or null pointer
     y : m-vector

   Results
     X = diag(y[0],y[1], ... , y[m-1]).  If X is null pointer, X is 
     created.

   Returns
     Returns x upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.
*/
TMatrix DiagonalMatrix(TMatrix X, TVector y)
{
 int i, j;
 if (!y) 
  { 
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (!X)
   { 
    if (!(X=CreateMatrix(V_DIM(y),V_DIM(y)))) 
     return (TMatrix)NULL; 
   }
  else
   if ((V_DIM(y) != M_ROW(X)) || (V_DIM(y) != M_COL(X)))
    { 
     MatrixError(SIZE_ERR); 
     return (TMatrix)NULL; 
    }
 for (i=V_DIM(y)-1; i >= 0; i--)
  {
   for (j=V_DIM(y)-1; j >= 0; j--) X[i][j]=0;
   X[i][i]=y[i];
  }
 return X;
}

/*
   Assumes
     X : m x 1 matrix or null pointer
     y : m-vector

   Results
     X is equal to the column vector y.  If X is null pointer, X is 
     created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.
*/
TMatrix ColumnMatrix(TMatrix X, TVector y)
{
 if (!y) 
  { 
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (!X)
   { 
    if (!(X=CreateMatrix(V_DIM(y),1))) 
     return (TMatrix)NULL; 
   }
  else
   if ((V_DIM(y) != M_ROW(X)) || (1 != M_COL(X)))
    { 
     MatrixError(SIZE_ERR); 
     return (TMatrix)NULL; 
    }
 memcpy(X[0],y,V_DIM(y)*sizeof(PRECISION));
 return X;
}

/*
   Assumes
     X : 1 x m matrix or null pointer
     y : m-vector

   Results
     X is equal to the row vector y.  If X is null pointer, X is 
     created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.
*/
TMatrix RowMatrix(TMatrix X, TVector y)
{
 if (!y) 
  { 
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (!X)
   { 
    if (!(X=CreateMatrix(1,V_DIM(y)))) 
     return (TMatrix)NULL; 
   }
  else
   if ((1 != M_ROW(X)) || (V_DIM(y) != M_COL(X)))
    { 
     MatrixError(SIZE_ERR); 
     return (TMatrix)NULL; 
    }
 memcpy(X[0],y,V_DIM(y)*sizeof(PRECISION));
 return X;
}

/*
   Assumes
     x : m vector or null pointer
     y : m vector

   Results
     x[i] = abs(y[i]).  If x is null pointer, x is created.

   Returns
     Returns x upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     x and y do not have to be distinct vectors.
*/
TVector AbsV(TVector x, TVector y)
{
 if (!y) 
  { 
   MatrixError(NULL_ERR); 
   return (TVector)NULL; 
  }
 if (!x)
   { 
    if (!(x=CreateVector(V_DIM(y)))) 
     return (TVector)NULL; 
   }
  else
   if (V_DIM(x) != V_DIM(y)) 
    { 
     MatrixError(SIZE_ERR); 
     return (TVector)NULL; 
    }
 bAbs(x,y,V_DIM(y));
 return x;
}

/*
   Assumes
     X : m x n matrix or null pointer
     Y : m x n matrix

   Results
     X[i][j] = abs(Y[i][j]).  If X is null pointer, X is created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     X and Y do not have to be distinct matrices
*/
TMatrix AbsM(TMatrix X, TMatrix Y)
{
 if (!Y) 
  {
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (!X)
   { 
    if (!(X=CreateMatrix(M_ROW(Y),M_COL(Y))))
     return (TMatrix)NULL; 
   }
  else
   if ((M_ROW(X) != M_ROW(Y)) || (M_COL(X) != M_COL(Y)))
    {
     MatrixError(SIZE_ERR); 
     return (TMatrix)NULL; 
    }
 bAbs(X[0],Y[0],M_ROW(Y)*M_COL(Y));
 return X;
}

/*
   Assumes
     x : m vector or null pointer
     y : m vector

   Results
     x = -x.  If x is null pointer, x is created.

   Returns
     Returns x upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     x and y do not have to be distinct vectors.
*/
TVector MinusV(TVector x, TVector y)
{
 if (!y) 
  {
   MatrixError(NULL_ERR); 
   return (TVector)NULL; 
  }
 if (!x)
   {
    if (!(x=CreateVector(V_DIM(y))))
     return (TVector)NULL; 
   }
  else
   if (V_DIM(x) != V_DIM(y))
    {
     MatrixError(SIZE_ERR); 
     return (TVector)NULL; 
    }
 bNegative(x,y,V_DIM(y));
 return x;
}

/*
   Assumes
     X : m x n matrix or null pointer
     Y : m x n matrix

   Results
     X = -Y.  If X is null pointer, X is created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     X and Y do not have to be distinct matrices.
*/
TMatrix MinusM(TMatrix X, TMatrix Y)
{
 if (!Y)
  {
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (!X)
   {
    if (!(X=CreateMatrix(M_ROW(Y),M_COL(Y))))
     return (TMatrix)NULL; 
   }
  else
   if ((M_ROW(X) != M_ROW(Y)) || (M_COL(X) != M_COL(Y)))
    {
     MatrixError(SIZE_ERR); 
     return (TMatrix)NULL; 
    }
 bNegative(X[0],Y[0],M_ROW(Y)*M_COL(Y));
 return X;
}

/*
   Assumes
     X : n x m matrix or null pointer
     Y : m x n matrix

   Results
     X = Y'.  If X is null pointer, X is created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     If Y is square, X and Y do not have to be distinct matrices
*/
TMatrix Transpose(TMatrix X, TMatrix Y)
{
 if (!Y)
  {
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (X == Y)
   {
    if (M_ROW(Y) != M_COL(Y))
     {
      MatrixError(SIZE_ERR);
      return (TMatrix)NULL;
     }
    bTransposeInPlace(Y[0],M_ROW(Y));
   }
  else
   {
    if (!X)
      {
       if (!(X=CreateMatrix(M_COL(Y),M_ROW(Y)))) 
        return (TMatrix)NULL;
      }
     else
      if ((M_ROW(X) != M_COL(Y)) || (M_COL(X) != M_ROW(Y)))
       {
        MatrixError(SIZE_ERR); 
        return (TMatrix)NULL; 
       }
    bTranspose(X[0],Y[0],M_COL(Y),M_ROW(Y));
   }
 return X;
}

/*
   Assumes
     X : (rows x cols) matrix or null pointer
     Y : (c x d) matrix, where c >= brow+rows and d >= bcol+cols

   Results
     X[i][j] = Y[brow+i][bcol+j], for 0 <= i < rows and 0 <= j < cols.  If
     X is null pointer, then the matrix X is created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.
*/
TMatrix SubMatrix(TMatrix X, TMatrix Y, int brow, int bcol, int rows, int cols)
{
 int i, s;
 if (!Y)
  {
   MatrixError(NULL_ERR);
   return (TMatrix)NULL;
  }
 if ((brow+rows > M_ROW(Y)) || (bcol+cols > M_COL(Y)))
  {
   MatrixError(SIZE_ERR);
   return (TMatrix)NULL;
  }
 if (!X)
   {
    if (!(X=CreateMatrix(rows,cols)))
     return (TMatrix)NULL;
   }
  else
   if ((rows != M_ROW(X)) || (cols != M_COL(X)))
    {
     MatrixError(SIZE_ERR);
     return (TMatrix)NULL;
    }
 if (cols == M_COL(Y))
   memcpy(X[0],Y[brow],rows*cols*sizeof(PRECISION));
  else
   for (s=cols*sizeof(PRECISION), i=rows-1; i >= 0; i--)
    memcpy(X[i],Y[brow+i]+bcol,s);
 return X;
}

/*
   Assumes
     x   : m-vector or null pointer
     Y   : m x n matrix
     col : 0 <= col < n

   Results
     x[i] = Y[i][col], for 0 <= i < m.  If x is null pointer, then the vector x
     is created.

   Returns
     Returns x upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.
*/
TVector ColumnVector(TVector x, TMatrix Y, int col)
{
 int i;
 if (!Y)
  {
   MatrixError(NULL_ERR);
   return (TVector)NULL; 
  }
 if ((col < 0) || (M_COL(Y) <= col))
  {
   MatrixError(SIZE_ERR);
   return (TVector)NULL;
  }
 if (!x)
   {
    if (!(x=CreateVector(M_ROW(Y))))
     return (TVector)NULL;
   }
  else
   if (V_DIM(x) != M_ROW(Y))
    {
     MatrixError(SIZE_ERR);
     return (TVector)NULL;
    }
 for (i=M_ROW(Y)-1; i >= 0; i--) x[i]=Y[i][col];
 return x;
}

/*
   Assumes
     x   : n-vector or null pointer
     Y   : m x n matrix
     row : 0 <= row < m

   Results
     x[j] = Y[col][j], for 0 <= j < n.  If x is null pointer, then the vector x
     is created.

   Returns
     Returns x upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.
*/
TVector RowVector(TVector x, TMatrix Y, int row)
{
 if (!Y)
  {
   MatrixError(NULL_ERR);
   return (TVector)NULL;
  }
 if ((row < 0) || (M_ROW(Y) <= row))
  {
   MatrixError(SIZE_ERR);
   return (TVector)NULL;
  }
 if (!x)
   {
    if (!(x=CreateVector(M_COL(Y))))
     return (TVector)NULL;
   }
  else
   if (V_DIM(x) != M_COL(Y))
    {
     MatrixError(SIZE_ERR);
     return (TVector)NULL;
    }
 memcpy(x,Y[row],M_COL(Y)*sizeof(PRECISION));
 return x;
}
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/


/******************************************************************************/
/***************************** Addition Routines ******************************/
/******************************************************************************/
/*
   Assumes
     x : m-vector or null pointer
     y : m-vector
     z : m-vector

   Results
     x = y + z.  If x is null pointer, x is created.

   Returns
     Returns x upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     x, y, and z do not have to be distinct vectors
*/
TVector AddVV(TVector x, TVector y, TVector z)
{
 if (!y || !z)
  {
   MatrixError(NULL_ERR);
   return (TVector)NULL;
  }
 if (V_DIM(y) != V_DIM(z))
  {
   MatrixError(SIZE_ERR);
   return (TVector)NULL;
  }
 if (!x)
   {
    if (!(x=CreateVector(V_DIM(y))))
     return (TVector)NULL;
   }
  else
   if (V_DIM(x) != V_DIM(y))
    {
     MatrixError(SIZE_ERR);
     return (TVector)NULL;
    }
 bAdd(x,y,z,V_DIM(y));
 return x;
}

/*
   Assumes
     x : m-vector or null pointer
     y : m-vector
     z : m-vector

   Results
     x = y - z.  If x is null pointer, x is created.

   Returns
     Returns x upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     x, y, and z do not have to be distinct vectors
*/
TVector SubtractVV(TVector x, TVector y, TVector z)
{
 if (!y || !z)
  {
   MatrixError(NULL_ERR);
   return (TVector)NULL;
  }
 if (V_DIM(y) != V_DIM(z))
  {
   MatrixError(SIZE_ERR);
   return (TVector)NULL;
  }
 if (!x)
   {
    if (!(x=CreateVector(V_DIM(y))))
     return (TVector)NULL;
   }
  else
   if (V_DIM(x) != V_DIM(y))
    {
     MatrixError(SIZE_ERR);
     return (TVector)NULL;
    }
 bSubtract(x,y,z,V_DIM(y));
 return x;
}

/*
   Assumes
     X : m x n matrix or null pointer
     Y : m x n matrix
     Z : m x n matrix

   Results
     X = Y + Z.  If X is null pointer, x is created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     X, Y, and Z do not have to be distinct matrices
*/
TMatrix AddMM(TMatrix X, TMatrix Y, TMatrix Z)
{
 if (!Y || !Z)
  {
   MatrixError(NULL_ERR);
   return (TMatrix)NULL;
  }
 if ((M_ROW(Y) != M_ROW(Z)) || (M_COL(Y) != M_COL(Z)))
  {
   MatrixError(SIZE_ERR);
   return (TMatrix)NULL;
  }
 if (!X)
   {
    if (!(X=CreateMatrix(M_ROW(Z),M_COL(Z))))
     return (TMatrix)NULL;
   }
  else
   if ((M_ROW(X) != M_ROW(Z)) || (M_COL(X) != M_COL(Z)))
    {
     MatrixError(SIZE_ERR);
     return (TMatrix)NULL;
    }
 bAdd(X[0],Y[0],Z[0],M_ROW(Z)*M_COL(Z));
 return X;
}

/*
   Assumes
     X : m x n matrix or null pointer
     Y : m x n matrix
     Z : m x n matrix

   Results
     X = Y - Z.  If X is null pointer, X is created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     X, Y, and Z do not have to be distinct matrices
*/
TMatrix SubtractMM(TMatrix X, TMatrix Y, TMatrix Z)
{
 if (!Y || !Z)
  {
   MatrixError(NULL_ERR);
   return (TMatrix)NULL;
  }
 if ((M_ROW(Y) != M_ROW(Z)) || (M_COL(Y) != M_COL(Z)))
  {
   MatrixError(SIZE_ERR);
   return (TMatrix)NULL;
  }
 if (!X)
   {
    if (!(X=CreateMatrix(M_ROW(Z),M_COL(Z))))
     return (TMatrix)NULL;
   }
  else
   if ((M_ROW(X) != M_ROW(Z)) || (M_COL(X) != M_COL(Z)))
    {
     MatrixError(SIZE_ERR);
     return (TMatrix)NULL;
    }
 bSubtract(X[0],Y[0],Z[0],M_ROW(Z)*M_COL(Z));
 return X;
}
/******************************************************************************/
/******************************************************************************/


/******************************************************************************/
/************************** Multiplication Routines ***************************/
/******************************************************************************/
/*
   Assumes
     x : m-vector or null pointer
     y : m-vector
     s : scalar

   Results
     x = s*y.  If x is null pointer, x is created.

   Returns
     Returns x upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     x and y do not have to be distinct
*/
TVector ProductSV(TVector x, PRECISION s, TVector y)
{
 if (!y)
  {
   MatrixError(NULL_ERR);
   return (TVector)NULL;
  }
 if (!x)
   {
    if (!(x=CreateVector(V_DIM(y))))
     return (TVector)NULL;
   }
  else
   if (V_DIM(x) != V_DIM(y))
    {
     MatrixError(SIZE_ERR);
     return (TVector)NULL;
    }
 bMultiply(x,y,s,V_DIM(y));
 return x;
}

/*
   Assumes
     X : m x n matrix or null pointer
     y : m x n matix
     s : scalar

   Results
     X = s*Y.  If X is null pointer, X is created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     X and Y do not have to be distinct
*/
TMatrix ProductSM(TMatrix X, PRECISION s, TMatrix Y)
{
 if (!Y)
  {
   MatrixError(NULL_ERR);
   return (TMatrix)NULL;
  }
 if (!X)
   {
    if (!(X=CreateMatrix(M_ROW(Y),M_COL(Y))))
     return (TMatrix)NULL;
   }
  else
   if ((M_ROW(X) != M_ROW(Y)) || (M_COL(X) != M_COL(Y)))
    {
     MatrixError(SIZE_ERR);
     return (TMatrix)NULL;
    }
 bMultiply(X[0],Y[0],s,M_ROW(Y)*M_COL(Y));
 return X;
}

/*
   Assumes
     x : n-vector or null pointer
     y : m-vector
     Z : m x n matrix

   Results
     x = y * Z.  If x is null pointer, x is created.

   Returns
     Returns x upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     If Z is square, x and y do not have to be distinct
*/
TVector ProductVM(TVector x, TVector y, TMatrix Z)
{
 PRECISION *ptr;
 if (!y || !Z)
  {
   MatrixError(NULL_ERR);
   return (TVector)NULL;
  }
 if (V_DIM(y) != M_ROW(Z))
  {
   MatrixError(SIZE_ERR);
   return (TVector)NULL;
  }
 if (!x)
   {
    if (!(x=CreateVector(M_COL(Z))))
     return (TVector)NULL;
    bMatrixMultiply(x,y,Z[0],1,V_DIM(x),V_DIM(y),0,0,0);
   }
  else
   {
    if (V_DIM(x) != M_COL(Z))
     {
      MatrixError(SIZE_ERR);
      return (TVector)NULL;
     }
    if (x == y)
      {
       if (!(ptr=(PRECISION*)malloc(V_DIM(x)*sizeof(PRECISION))))
        {
         MatrixError(MEM_ERR);
         return (TVector)NULL;
        }
       bMatrixMultiply(ptr,y,Z[0],1,V_DIM(x),V_DIM(y),0,0,0);
       memcpy(x,ptr,V_DIM(x)*sizeof(PRECISION));
       free(ptr);
      }
     else
      bMatrixMultiply(x,y,Z[0],1,V_DIM(x),V_DIM(y),0,0,0);
   }
 return x;
}

/*
   Assumes
     x : m-vector or null pointer
     Y : m x n matrix
     z : n-vector

   Results
     x = Y * z.  If x is null pointer, x is created.

   Returns
     Returns x upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     If Y is square, x and z do not have to be distinct
*/
TVector ProductMV(TVector x, TMatrix Y, TVector z)
{
 PRECISION *ptr;
 if (!Y || !z)
  {
   MatrixError(NULL_ERR);
   return (TVector)NULL;
  }
 if (M_COL(Y) != V_DIM(z))
  {
   MatrixError(SIZE_ERR);
   return (TVector)NULL;
  }
 if (!x)
   {
    if (!(x=CreateVector(M_ROW(Y))))
     return (TVector)NULL;
    bMatrixMultiply(x,Y[0],z,V_DIM(x),1,V_DIM(z),0,0,0);
   }
  else
   {
    if (V_DIM(x) != M_ROW(Y))
     {
      MatrixError(SIZE_ERR);
      return (TVector)NULL;
     }
    if (x == z)
      {
       if (!(ptr=(PRECISION*)malloc(V_DIM(x)*sizeof(PRECISION))))
        {
         MatrixError(MEM_ERR);
         return (TVector)NULL;
        }
       bMatrixMultiply(ptr,Y[0],z,V_DIM(x),1,V_DIM(z),0,0,0);
       memcpy(x,ptr,V_DIM(x)*sizeof(PRECISION));
       free(ptr);
      }
     else
      bMatrixMultiply(x,Y[0],z,V_DIM(x),1,V_DIM(z),0,0,0);
   }
 return x;
}


/*
   Assumes
     X : m x n matrix or null pointer
     Y : m x r matrix
     Z : r x n matrix

   Results
     X = Y * Z.  If X is null pointer, X is created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     If Z is square, X does not have to be distinct from Y
     If Y is square, X does not have to be distinct from Z
*/
TMatrix ProductMM(TMatrix X, TMatrix Y, TMatrix Z)
{
 PRECISION *ptr;
 if (!Y || !Z)
  {
   MatrixError(NULL_ERR);
   return (TMatrix)NULL;
  }
 if (M_COL(Y) != M_ROW(Z))
  {
   MatrixError(SIZE_ERR);
   return (TMatrix)NULL;
  }
 if (!X)
   {
    if (!(X=CreateMatrix(M_ROW(Y),M_COL(Z)))) 
     return (TMatrix)NULL;
    bMatrixMultiply(X[0],Y[0],Z[0],M_ROW(X),M_COL(X),M_COL(Y),0,0,0);
   }
  else
   {
    if ((M_ROW(X) != M_ROW(Y)) || (M_COL(X) != M_COL(Z)))
     {
      MatrixError(SIZE_ERR); 
      return (TMatrix)NULL; 
     }
    if ((X == Y) || (X == Z))
      {
       if (!(ptr=(PRECISION*)malloc(M_ROW(X)*M_COL(X)*sizeof(PRECISION))))
        {
         MatrixError(MEM_ERR); 
         return (TMatrix)NULL; 
        }
       bMatrixMultiply(ptr,Y[0],Z[0],M_ROW(X),M_COL(X),M_COL(Y),0,0,0);
       memcpy(X[0],ptr,M_ROW(X)*M_COL(X)*sizeof(PRECISION));
       free(ptr);
      }
     else
      bMatrixMultiply(X[0],Y[0],Z[0],M_ROW(X),M_COL(X),M_COL(Y),0,0,0);
   }
 return X;
}
/******************************************************************************/
/******************************************************************************/


/******************************************************************************/
/********************* Multiplication/Transpose Routines **********************/
/******************************************************************************/
/*
   Assumes
     X : m x n matrix or null pointer
     Y : r x m matrix
     Z : r x n matrix

   Results
     X = Y'*Z.  If X is null pointer, X is created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     Y and Z do not have to be distinct matrices.
     If Y is square, X and Z do not have to be distinct matrices.
     If Z is square, X and Y do not have to be distinct matrices.
*/
TMatrix TransposeProductMM(TMatrix X, TMatrix Y, TMatrix Z)
{
 PRECISION *ptr;
 if (!Y || !Z)
  {
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (M_ROW(Y) != M_ROW(Z))
  { 
   MatrixError(SIZE_ERR); 
   return (TMatrix)NULL; 
  }
 if (!X)
   {
    if (!(X=CreateMatrix(M_COL(Y),M_COL(Z)))) 
     return (TMatrix)NULL;
    bMatrixMultiply(X[0],Y[0],Z[0],M_ROW(X),M_COL(X),M_ROW(Y),0,1,0);
   }
  else
   {
    if ((M_ROW(X) != M_COL(Y)) || (M_COL(X) != M_COL(Z)))
     {
      MatrixError(SIZE_ERR); 
      return (TMatrix)NULL; 
     }
    if ((X == Y) || (X == Z))
      {
       if (!(ptr=(PRECISION*)malloc(M_ROW(X)*M_COL(X)*sizeof(PRECISION))))
        {
         MatrixError(MEM_ERR); 
         return (TMatrix)NULL; 
        }
       bMatrixMultiply(ptr,Y[0],Z[0],M_ROW(X),M_COL(X),M_ROW(Y),0,1,0);
       memcpy(X[0],ptr,M_ROW(X)*M_COL(X)*sizeof(PRECISION));
       free(ptr);
      }
     else
      bMatrixMultiply(X[0],Y[0],Z[0],M_ROW(X),M_COL(X),M_ROW(Y),0,1,0);
   }
 return X;
}

/*
   Assumes
     X : m x n matrix or null pointer
     Y : m x p matrix
     Z : n x p matrix

   Results
     X = Y*Z'.  If X is null pointer, X is created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.

   Notes
     Y and Z do not have to be distinct matrices.
     If Y is square, X and Z do not have to be distinct matrices.
     If Z is square, X and Y do not have to be distinct matrices.
*/
TMatrix ProductTransposeMM(TMatrix X, TMatrix Y, TMatrix Z)
{
 PRECISION *ptr;
 if (!Y || !Z)
  {
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (M_COL(Y) != M_COL(Z))
  {
   MatrixError(SIZE_ERR); 
   return (TMatrix)NULL; 
  }
 if (!X)
   {
    if (!(X=CreateMatrix(M_ROW(Y),M_ROW(Z))))
     return (TMatrix)NULL;
    bMatrixMultiply(X[0],Y[0],Z[0],M_ROW(X),M_COL(X),M_COL(Y),0,0,1);
   }
  else
   {
    if ((M_ROW(X) != M_ROW(Y)) || (M_COL(X) != M_ROW(Z)))
     {
      MatrixError(SIZE_ERR); 
      return (TMatrix)NULL; 
     }
    if ((X == Y) || (X == Z))
      {
       if (!(ptr=(PRECISION*)malloc(M_ROW(X)*M_COL(X)*sizeof(PRECISION))))
        {
         MatrixError(MEM_ERR); 
         return (TMatrix)NULL; 
        }
       bMatrixMultiply(ptr,Y[0],Z[0],M_ROW(X),M_COL(X),M_COL(Y),0,0,1);
       memcpy(X[0],ptr,M_ROW(X)*M_COL(X)*sizeof(PRECISION));
       free(ptr);
      }
     else
      bMatrixMultiply(X[0],Y[0],Z[0],M_ROW(X),M_COL(X),M_COL(Y),0,0,1);
   }
 return X;
}
/******************************************************************************/
/******************************************************************************/


/******************************************************************************/
/********************** Multiplication/Inverse Routines ***********************/
/******************************************************************************/
/*
   Assumes
     x : m vector or null
     Y : m x m matrix
     z : m vector

   Results
     x = Inverse(Y) * z.

   Returns
     Returns x upon success, null otherwise.

   Notes
     The vectors x and z do not have to be distinct. Uses Gaussian elimination 
     with partial pivoting and back substitution.  A null return indicates 
     that either Y or z was null, Y was singular, the matrices were not of the 
     required size, or the routine was unable to allocate needed memory.  Call 
     GetMatrixError() to determine which of these occured.
*/
TVector InverseProductMV(TVector x, TMatrix Y, TVector z)
{
 PRECISION *LU;
 int *p;
 TVector rtrn=(TVector)NULL;
 if (!Y || !z) 
   MatrixError(NULL_ERR); 
  else
   if ((M_ROW(Y) != M_COL(Y)) || (M_COL(Y) != V_DIM(z)))
     MatrixError(SIZE_ERR); 
    else
     if (!(p=(int*)malloc(M_ROW(Y)*sizeof(int))))
       MatrixError(MEM_ERR); 
      else
       {
        if (!(LU=(PRECISION*)malloc(M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION))))
          MatrixError(MEM_ERR);
         else
          {  
           memcpy(LU,Y[0],M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION));
           if (bLU(p,LU,M_ROW(Y),M_ROW(Y)))
             MatrixError(SING_ERR);
            else
             if (rtrn=((x == z) ? x : EquateVector(x,z)))
              {
               bPermutationMultiply(p,rtrn,V_DIM(rtrn),1,M_ROW(Y),1,0);
               bSolveUnitTriangular(LU,rtrn,V_DIM(rtrn),1,0,0,0);
               bSolveTriangular(LU,rtrn,V_DIM(rtrn),1,1,0,0);
              }
           free(LU);
          }
        free(p);
       }
 return rtrn;
}

/*
   Assumes
     x : m vector or null
     Y : m x m matrix upper triangular matrix
     z : m vector

   Results
     x = Inverse(Y) * z.

   Returns
     Returns x upon success, null otherwise.

   Notes
     The vectors x and z do not have to be distinct. Uses back substitution.  
     A null return indicates that either Y or z was null, Y was singular, or
     the matrices were not of the required size.  Call GetMatrixError() to 
     determine which of these occured.
*/
TVector InverseProductUV(TVector x, TMatrix Y, TVector z)
{
 TVector rtrn=(TVector)NULL;
 int err;
 if (!Y || !z) 
   MatrixError(NULL_ERR); 
  else
   if ((M_ROW(Y) != M_COL(Y)) || (M_COL(Y) != V_DIM(z)))
     MatrixError(SIZE_ERR); 
    else
     if (rtrn=((x == z) ? x : EquateVector(x,z)))
      if (err=bSolveTriangular(Y[0],rtrn,V_DIM(rtrn),1,1,0,0))
       {
        if (rtrn != x) FreeVector(rtrn);
        rtrn=(TVector)NULL;
        MatrixError(err);
       }   
 return rtrn;
}

/*
   Assumes
     x : m vector or null
     Y : m x m matrix lower triangular matrix
     z : m vector

   Results
     x = Inverse(Y) * z.

   Returns
     Returns x upon success, null otherwise.

   Notes
     The vectors x and z do not have to be distinct. Uses back substitution.  
     A null return indicates that either Y or z was null, Y was singular, or
     the matrices were not of the required size.  Call GetMatrixError() to 
     determine which of these occured.
*/
TVector InverseProductLV(TVector x, TMatrix Y, TVector z)
{
 TVector rtrn=(TVector)NULL;
 int err;
 if (!Y || !z) 
   MatrixError(NULL_ERR); 
  else
   if ((M_ROW(Y) != M_COL(Y)) || (M_COL(Y) != V_DIM(z)))
     MatrixError(SIZE_ERR); 
    else
     if (rtrn=((x == z) ? x : EquateVector(x,z)))
      if (err=bSolveTriangular(Y[0],rtrn,V_DIM(rtrn),1,0,0,0))
       {
        if (rtrn != x) FreeVector(rtrn);
        rtrn=(TVector)NULL;
        MatrixError(err);
       }   
 return rtrn;
}

/*
   Assumes
     X : m x n matrix or null
     Y : m x m matrix
     Z : m x n matrix

   Results
     X = Inverse(Y) * Z.

   Returns
     Returns X upon success, null otherwise.

   Notes
     Size permitting, X, Y and Z do not have to be distinct. Uses Gaussian 
     elimination with partial pivoting and back substitution.  A null return 
     indicates that either Y or Z was null, Y was singular, the matrices were
     not of the required size, or the routine was unable to allocate needed 
     memory.  Call GetMatrixError() to determine which of these occured.
*/
TMatrix InverseProductMM(TMatrix X, TMatrix Y, TMatrix Z)
{
 PRECISION *LU;
 int *p;
 TMatrix rtrn=(TMatrix)NULL;
 if (!Y || !Z) 
   MatrixError(NULL_ERR); 
  else
   if ((M_ROW(Y) != M_COL(Y)) || (M_COL(Y) != M_ROW(Z)))
     MatrixError(SIZE_ERR); 
    else
     if (!(p=(int*)malloc(M_ROW(Y)*sizeof(int))))
       MatrixError(MEM_ERR); 
      else
       {
        if (!(LU=(PRECISION*)malloc(M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION))))
          MatrixError(MEM_ERR);
         else
          {  
           memcpy(LU,Y[0],M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION));
           if (bLU(p,LU,M_ROW(Y),M_ROW(Y)))
             MatrixError(SING_ERR);
            else
             if (rtrn=((X == Z) ? X : EquateMatrix(X,Z)))
              {
               bPermutationMultiply(p,rtrn[0],M_ROW(rtrn),M_COL(rtrn),M_ROW(Y),1,0);
               bSolveUnitTriangular(LU,rtrn[0],M_ROW(rtrn),M_COL(rtrn),0,0,0);
               bSolveTriangular(LU,rtrn[0],M_ROW(rtrn),M_COL(rtrn),1,0,0);
              }
           free(LU);
          }
        free(p);
       }
 return rtrn;
}

/*
   Assumes
     X : m x n matrix or null
     Y : m x m upper triangular matrix
     Z : m x n matrix

   Results
     X = Inverse(Y) * Z.

   Returns
     Returns X upon success, null otherwise.

   Notes
     Size permitting, X, Y and Z do not have to be distinct. Uses back 
     substitution.  A null return indicates that either Y or Z was null, 
     Y was singular, or the matrices were not of the required size.  Call 
     GetMatrixError() to determine which of these occured.
*/
TMatrix InverseProductUM(TMatrix X, TMatrix Y, TMatrix Z)
{
 PRECISION *ptr;
 TMatrix rtrn=(TMatrix)NULL;
 int err;
 if (!Y || !Z) 
   MatrixError(NULL_ERR); 
  else
   if ((M_ROW(Y) != M_COL(Y)) || (M_COL(Y) != M_ROW(Z)))
     MatrixError(SIZE_ERR); 
    else
     if (X == Y)
       if (!(ptr=(PRECISION*)malloc(M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION))))
         MatrixError(MEM_ERR);
        else
         {
          memcpy(ptr,Y[0],M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION));
          if (rtrn=((X == Z) ? X : EquateMatrix(X,Z)))
           if (err=bSolveTriangular(ptr,rtrn[0],M_ROW(rtrn),M_COL(rtrn),1,0,0))
            {
             if (rtrn != X) FreeMatrix(rtrn);
             rtrn=(TMatrix)NULL;
             MatrixError(err);
            }
          free(ptr);
         }
      else
       if (rtrn=((X == Z) ? X : EquateMatrix(X,Z)))
        if (err=bSolveTriangular(Y[0],rtrn[0],M_ROW(rtrn),M_COL(rtrn),1,0,0))
         {
          if (rtrn != X) FreeMatrix(rtrn);
          rtrn=(TMatrix)NULL;
          MatrixError(err);
         } 
 return rtrn;
}

/*
   Assumes
     X : m x n matrix or null
     Y : m x m lower triangular matrix
     Z : m x n matrix

   Results
     X = Inverse(Y) * Z.

   Returns
     Returns X upon success, null otherwise.

   Notes
     Size permitting, X, Y and Z do not have to be distinct. Uses back 
     substitution.  A null return indicates that either Y or Z was null, 
     Y was singular, or the matrices were not of the required size.  Call 
     GetMatrixError() to determine which of these occured.
*/
TMatrix InverseProductLM(TMatrix X, TMatrix Y, TMatrix Z)
{
 PRECISION *ptr;
 TMatrix rtrn=(TMatrix)NULL;
 int err;
 if (!Y || !Z) 
   MatrixError(NULL_ERR); 
  else
   if ((M_ROW(Y) != M_COL(Y)) || (M_COL(Y) != M_ROW(Z)))
     MatrixError(SIZE_ERR); 
    else
     if (X == Y)
       if (!(ptr=(PRECISION*)malloc(M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION))))
         MatrixError(MEM_ERR);
        else
         {
          memcpy(ptr,Y[0],M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION));
          if (rtrn=((X == Z) ? X : EquateMatrix(X,Z)))
           if (err=bSolveTriangular(ptr,rtrn[0],M_ROW(rtrn),M_COL(rtrn),0,0,0))
            {
             if (rtrn != X) FreeMatrix(rtrn);
             rtrn=(TMatrix)NULL;
             MatrixError(err);
            }
          free(ptr);
         }
      else
       if (rtrn=((X == Z) ? X : EquateMatrix(X,Z)))
        if (err=bSolveTriangular(Y[0],rtrn[0],M_ROW(rtrn),M_COL(rtrn),0,0,0))
         {
          if (rtrn != X) FreeMatrix(rtrn);
          rtrn=(TMatrix)NULL;
          MatrixError(err);
         } 
 return rtrn;
}

/*
   Assumes
     x : n vector or null pointer
     y : n vector
     Z : n x n invertible matrix

   Results
     x = y * Inverse(Z).  
    
   Returns
     Returns x upon success, null otherwise.

   Notes
     The vectors x and y do not have to be distinct. Uses Gaussian elimination 
     with partial pivoting and back substitution.  A null return indicates 
     that either y or Z was null, Z was singular, the matrices were not of the 
     required size, or the routine was unable to allocate needed memory.  Call 
     GetMatrixError() to determine which of these occured.
*/
TVector ProductInverseVM(TVector x, TVector y, TMatrix Z)
{
 PRECISION *LU;
 int *p, err;
 TVector rtrn=(TVector)NULL;
 if (!y || !Z) 
   MatrixError(NULL_ERR); 
  else
   if ((M_ROW(Z) != M_COL(Z)) || (V_DIM(y) != M_ROW(Z)))
     MatrixError(SIZE_ERR); 
    else
     if (!(p=(int*)malloc(M_ROW(Z)*sizeof(int))))
       MatrixError(MEM_ERR); 
      else
       {
        if (!(LU=(PRECISION*)malloc(M_ROW(Z)*M_ROW(Z)*sizeof(PRECISION))))
          MatrixError(MEM_ERR);
         else
          {  
           memcpy(LU,Z[0],M_ROW(Z)*M_ROW(Z)*sizeof(PRECISION));
           if (err=bLU(p,LU,M_ROW(Z),M_ROW(Z)))
             MatrixError(err);
            else
             if (rtrn=((x == y) ? x : EquateVector(x,y)))
              {
               bSolveTriangular(LU,rtrn,V_DIM(rtrn),1,1,1,1);                
               bSolveUnitTriangular(LU,rtrn,V_DIM(rtrn),1,0,1,1);
               bPermutationMultiply(p,rtrn,V_DIM(rtrn),1,M_ROW(Z),0,1);                            
              }
           free(LU);
          }
        free(p);
       }
 return rtrn;
}

/*
   Assumes
     x : n vector or null pointer
     y : n vector
     Z : n x n upper triangular matrix

   Results
     x = y * Inverse(Z).  
    
   Returns
     Returns x upon success, null otherwise.

   Notes
     The vectors x and y do not have to be distinct. Uses back substitution.  
     A null return indicates that either y or Z was null, Z was singular, or
     the matrices were not of the  required size.  Call GetMatrixError() to 
     determine which of these occured.
*/
TVector ProductInverseVU(TVector x, TVector y, TMatrix Z)
{
 TVector rtrn=(TVector)NULL;
 int err;
 if (!y || !Z) 
   MatrixError(NULL_ERR); 
  else
   if ((M_ROW(Z) != M_COL(Z)) || (V_DIM(y) != M_ROW(Z)))
     MatrixError(SIZE_ERR); 
    else
     if (rtrn=((x == y) ? x : EquateVector(x,y)))
      if (err=bSolveTriangular(Z[0],rtrn,V_DIM(rtrn),1,1,1,1))
       {
        if (rtrn != x) FreeVector(rtrn);
        rtrn=(TVector)NULL;
        MatrixError(err);                         
       }
 return rtrn;
}

/*
   Assumes
     x : n vector or null pointer
     y : n vector
     Z : n x n lower triangular matrix

   Results
     x = y * Inverse(Z).  
    
   Returns
     Returns x upon success, null otherwise.

   Notes
     The vectors x and y do not have to be distinct. Uses back substitution.  
     A null return indicates that either y or Z was null, Z was singular, or
     the matrices were not of the  required size.  Call GetMatrixError() to 
     determine which of these occured.
*/
TVector ProductInverseVL(TVector x, TVector y, TMatrix Z)
{
 TVector rtrn=(TVector)NULL;
 int err;
 if (!y || !Z) 
   MatrixError(NULL_ERR); 
  else
   if ((M_ROW(Z) != M_COL(Z)) || (V_DIM(y) != M_ROW(Z)))
     MatrixError(SIZE_ERR); 
    else
     if (rtrn=((x == y) ? x : EquateVector(x,y)))
      if (err=bSolveTriangular(Z[0],rtrn,V_DIM(rtrn),1,0,1,1))
       {
        if (rtrn != x) FreeVector(rtrn);
        rtrn=(TVector)NULL;
        MatrixError(err);                         
       }
 return rtrn;
}

/*
   Assumes
     X : m x n matrix
     Y : m x n matrix
     Z : n x n matrix

   Results
     X = Y * Inverse(Z).  
    
   Returns
     Returns X upon success, null otherwise.

   Notes
     Size permitting, X, Y and Z do not have to be distinct. Uses Gaussian 
     elimination with partial pivoting and back substitution.  A null return 
     indicates that either Y or Z was null, Z was singular, the matrices were
     not of the required size, or the routine was unable to allocate needed 
     memory.  Call GetMatrixError() to determine which of these occured.
*/
TMatrix ProductInverseMM(TMatrix X, TMatrix Y, TMatrix Z)
{
 PRECISION *LU;
 int *p;
 TMatrix rtrn=(TMatrix)NULL;
 if (!Y || !Z) 
   MatrixError(NULL_ERR); 
  else
   if ((M_ROW(Z) != M_COL(Z)) || (M_COL(Y) != M_ROW(Z)))
     MatrixError(SIZE_ERR); 
    else
     if (!(p=(int*)malloc(M_ROW(Z)*sizeof(int))))
       MatrixError(MEM_ERR); 
      else
       {
        if (!(LU=(PRECISION*)malloc(M_ROW(Z)*M_ROW(Z)*sizeof(PRECISION))))
          MatrixError(MEM_ERR);
         else
          {  
           memcpy(LU,Z[0],M_ROW(Z)*M_ROW(Z)*sizeof(PRECISION));
           if (bLU(p,LU,M_ROW(Z),M_ROW(Z)))
             MatrixError(SING_ERR);
            else
             if (rtrn=((X == Y) ? X : EquateMatrix(X,Y)))
              {
               bSolveTriangular(LU,rtrn[0],M_COL(rtrn),M_ROW(rtrn),1,1,1);
               bSolveUnitTriangular(LU,rtrn[0],M_COL(rtrn),M_ROW(rtrn),0,1,1);
               bPermutationMultiply(p,rtrn[0],M_COL(rtrn),M_ROW(rtrn),M_ROW(Z),0,1);  
              } 
           free(LU);
          }
        free(p);
       }
 return rtrn;
}

/*
   Assumes
     X : m x n matrix
     Y : m x n matrix
     Z : n x n upper triangular matrix

   Results
     X = Y * Inverse(Z).  
    
   Returns
     Returns X upon success, null otherwise.

   Notes
     Size permitting, X, Y and Z do not have to be distinct. Uses back 
     substitution.  A null return indicates that either Y or Z was null, 
     Z was singular, or the matrices were not of the required size.  Call 
     GetMatrixError() to determine which of these occured.
*/
TMatrix ProductInverseMU(TMatrix X, TMatrix Y, TMatrix Z)
{
 PRECISION *ptr;
 TMatrix rtrn=(TMatrix)NULL;
 int err;
 if (!Y || !Z) 
   MatrixError(NULL_ERR); 
  else
   if ((M_ROW(Z) != M_COL(Z)) || (M_COL(Y) != M_ROW(Z)))
     MatrixError(SIZE_ERR); 
    else
     if (X == Z)
       if (!(ptr=(PRECISION*)malloc(M_ROW(Z)*M_ROW(Z)*sizeof(PRECISION))))
         MatrixError(MEM_ERR);
        else
         {
          memcpy(ptr,Z[0],M_ROW(Z)*M_ROW(Z)*sizeof(PRECISION));
          if (rtrn=((X == Y) ? X : EquateMatrix(X,Y)))
           if (err=bSolveTriangular(ptr,rtrn[0],M_COL(rtrn),M_ROW(rtrn),1,1,1))
            {
             if (rtrn != X) FreeMatrix(rtrn);
             rtrn=(TMatrix)NULL;
             MatrixError(err);
            }
          free(ptr);
         }
      else
       if (rtrn=((X == Y) ? X : EquateMatrix(X,Y)))
        if (err=bSolveTriangular(Z[0],rtrn[0],M_COL(rtrn),M_ROW(rtrn),1,1,1))
         {
          if (rtrn != X) FreeMatrix(rtrn);
          rtrn=(TMatrix)NULL;
          MatrixError(err);
         } 
 return rtrn;
}

/*
   Assumes
     X : m x n matrix
     Y : m x n matrix
     Z : n x n upper triangular matrix

   Results
     X = Y * Inverse(Z).  
    
   Returns
     Returns X upon success, null otherwise.

   Notes
     Size permitting, X, Y and Z do not have to be distinct. Uses back 
     substitution.  A null return indicates that either Y or Z was null, 
     Z was singular, or the matrices were not of the required size.  Call 
     GetMatrixError() to determine which of these occured.
*/
TMatrix ProductInverseML(TMatrix X, TMatrix Y, TMatrix Z)
{
 PRECISION *ptr;
 TMatrix rtrn=(TMatrix)NULL;
 int err;
 if (!Y || !Z) 
   MatrixError(NULL_ERR); 
  else
   if ((M_ROW(Z) != M_COL(Z)) || (M_COL(Y) != M_ROW(Z)))
     MatrixError(SIZE_ERR); 
    else
     if (X == Z)
       if (!(ptr=(PRECISION*)malloc(M_ROW(Z)*M_ROW(Z)*sizeof(PRECISION))))
         MatrixError(MEM_ERR);
        else
         {
          memcpy(ptr,Z[0],M_ROW(Z)*M_ROW(Z)*sizeof(PRECISION));
          if (rtrn=((X == Y) ? X : EquateMatrix(X,Y)))
           if (err=bSolveTriangular(ptr,rtrn[0],M_COL(rtrn),M_ROW(rtrn),0,1,1))
            {
             if (rtrn != X) FreeMatrix(rtrn);
             rtrn=(TMatrix)NULL;
             MatrixError(err);
            }
          free(ptr);
         }
      else
       if (rtrn=((X == Y) ? X : EquateMatrix(X,Y)))
        if (err=bSolveTriangular(Z[0],rtrn[0],M_COL(rtrn),M_ROW(rtrn),0,1,1))
         {
          if (rtrn != X) FreeMatrix(rtrn);
          rtrn=(TMatrix)NULL;
          MatrixError(err);
         } 
 return rtrn;
}
/******************************************************************************/
/******************************************************************************/

/******************************************************************************/
/************************** Matrix Inverse Routines ***************************/
/******************************************************************************/
/*
   Assumes
     X : m x m matrix or null pointer
     Y : m x m matrix

   Results
     X = inverse(Y)

   Returns
    Returns X upon success, null otherwise

   Notes
     The matrices X and Y do not have to be distinct.  Uses Gaussian
     elimination with partial pivoting and back substitution to form the
     inverse.  A null return indicates that either Y or Z was null, Y was
     singular, or unable to allocate required memory.  Call GetMatrixError()
     to determine which of these occured.
*/
TMatrix Inverse_LU(TMatrix X, TMatrix Y)
{
 PRECISION *LU;
 int *p;
 TMatrix rtrn=(TMatrix)NULL;
 if (!Y) 
   MatrixError(NULL_ERR); 
  else
   if (M_ROW(Y) != M_COL(Y))
     MatrixError(SIZE_ERR); 
    else
     if (!(p=(int*)malloc(M_ROW(Y)*sizeof(int))))
       MatrixError(MEM_ERR); 
      else
       {
        if (!(LU=(PRECISION*)malloc(M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION))))
          MatrixError(MEM_ERR);
         else
          {  
           memcpy(LU,Y[0],M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION));
           if (bLU(p,LU,M_ROW(Y),M_ROW(Y)))
             MatrixError(SING_ERR);
            else
             if (rtrn=IdentityMatrix(X,M_ROW(Y)))
              {
               bPermutationMultiply(p,rtrn[0],M_ROW(rtrn),M_COL(rtrn),M_ROW(Y),1,0);
               bSolveUnitTriangular(LU,rtrn[0],M_ROW(rtrn),M_COL(rtrn),0,0,0);
               bSolveTriangular(LU,rtrn[0],M_ROW(rtrn),M_COL(rtrn),1,0,0);                        
              }
           free(LU);
          }
        free(p);
       }
 return rtrn;
}

/*
   Assumes
     X : m x m matrix or null pointer
     Y : m x m matrix

   Results
     X = inverse(Y)

   Returns
    Returns X upon success, null otherwise

   Notes
    The matrices X and Y do not have to be distinct.  Uses singular value
    decomposition to form the inverse.  A null return indicates that either
    Y or Z was null, Y was singular, or unable to allocate required memory.
    Call GetMatrixError() to determine which of these occured.
*/
TMatrix Inverse_SVD(TMatrix X, TMatrix Y)
{
 int i, j;
 TMatrix rtrn=(TMatrix)NULL;
 PRECISION scale, tolerance, *U, *V, *d;
 if (!Y)
   MatrixError(NULL_ERR);
  else
   if (M_ROW(Y) != M_COL(Y))
     MatrixError(SIZE_ERR);
    else
     if (!(U=(PRECISION*)malloc(M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION))))
       MatrixError(MEM_ERR);
      else
       {
        if (!(V=(PRECISION*)malloc(M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION))))
          MatrixError(MEM_ERR);
         else
          {
           if (!(d=(PRECISION*)malloc(M_ROW(Y)*sizeof(PRECISION))))
             MatrixError(MEM_ERR);
            else
             {
              memcpy(U,Y[0],M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION));
              if (!bSVD(U,d,V,M_ROW(Y),M_ROW(Y))) 
               {
                for (tolerance=d[0], j=M_ROW(Y)-1; j > 0; j--) 
                 if (tolerance < d[j]) tolerance=d[j];
                tolerance*=MACHINE_EPSILON;
                for (j=M_ROW(Y)-1; j >= 0; j--)
                 {
                  if (d[j] < tolerance) 
                   { 
                    MatrixError(SING_ERR);  
                    free(U);
                    free(V);
                    free(d);
                    return (TMatrix)NULL;
                   }
                  scale=1.0/d[j];
                  for (i=(M_ROW(Y)-1)*M_ROW(Y)+j; i >= 0; i-=M_ROW(Y)) V[i]*=scale;
                 }
                if (X && ((M_ROW(X) != M_ROW(Y)) || (M_COL(X) != M_ROW(Y))))
                  MatrixError(SIZE_ERR);
                 else
                  if (!(rtrn=X ? X : CreateMatrix(M_ROW(Y),M_ROW(Y))))
                    MatrixError(MEM_ERR);
                   else
                    bMatrixMultiply(rtrn[0],V,U,M_ROW(Y),M_ROW(Y),M_ROW(Y),0,0,1);
               } 
              free(d);
             }
           free(V);
          }
        free(U);
       }
 return rtrn;
}

/*
   Assumes
     X : m x m matrix or null pointer
     Y : m x m symmetric matrix

   Results
     X = inverse(Y)

   Returns
    Returns X upon success, null otherwise

   Notes
    The matrices X and Y do not have to be distinct.  Uses Cholesky
    decomposition to form the inverse.  Only the upper half of Y is used.
    A null return indicates that either Y or Z was null, Y was singular,
    or unable to allocate required memory.  Call GetMatrixError() to
    determine which of these occured.
*/
TMatrix Inverse_Cholesky(TMatrix X, TMatrix Y)
{
 TMatrix rtrn=(TMatrix)NULL;
 PRECISION *ptr;
 if (!Y)
   MatrixError(NULL_ERR);
  else
   if (M_ROW(Y) != M_COL(Y))
     MatrixError(SIZE_ERR);
    else
     if (!(ptr=(PRECISION*)malloc(M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION))))
       MatrixError(MEM_ERR);
      else
       {
        memcpy(ptr,Y[0],M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION));
        if (bCholesky(ptr,M_ROW(Y),1,0))
          MatrixError(SING_ERR);
         else
          {
           if (rtrn=IdentityMatrix(X,M_ROW(Y)))
            {
             bSolveTriangular(ptr,rtrn[0],M_ROW(Y),M_ROW(Y),1,0,0);
             bMatrixMultiply(ptr,rtrn[0],rtrn[0],M_ROW(Y),M_ROW(Y),M_ROW(Y),0,0,1);
             memcpy(rtrn[0],ptr,M_ROW(Y)*M_ROW(Y)*sizeof(PRECISION));
            }
          }
        free(ptr);
       }
 return rtrn;
}

/*
   Assumes
     X : m x m matrix or null pointer
     T : m x m upper triangular matrix

   Results
     X = inverse(T)

   Returns
     Returns X upon success, null otherwise

   Notes
     The matrices X and T do not have to be distinct.  Uses back substitution
     to form the inverse.  A null return indicates that T was null or
     singular, or unable to allocate required memory.  Call
     GetMatrixError() to determine which of these occured.
*/
TMatrix Inverse_UT(TMatrix X, TMatrix T)
{
 PRECISION *ptr;
 int err;
 if (!T)
  {
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (M_ROW(T) != M_COL(T))
  {
   MatrixError(SIZE_ERR); 
   return (TMatrix)NULL; 
  }
 if (!X)
   {
    if (X=IdentityMatrix(X,M_ROW(T))) 
     if (err=bSolveTriangular(T[0],X[0],M_ROW(T),M_ROW(T),1,0,0))
      {
       MatrixError(err); 
       FreeMatrix(X); 
       return (TMatrix)NULL; 
      }
   }
  else
   if (X == T)
     {
      if (!(ptr=(PRECISION*)malloc(M_ROW(T)*M_ROW(T)*sizeof(PRECISION))))
       {
        MatrixError(MEM_ERR); 
        return (TMatrix)NULL; 
       }
      memcpy(ptr,T[0],M_ROW(T)*M_ROW(T)*sizeof(PRECISION));
      if (X=IdentityMatrix(X,M_ROW(T))) 
       if (err=bSolveTriangular(ptr,X[0],M_ROW(T),M_ROW(T),1,0,0))
        {
         MatrixError(err); 
         free(ptr); 
         return (TMatrix)NULL; 
        }     
      free(ptr);
     }
    else
     if (X=IdentityMatrix(X,M_ROW(T))) 
      if (err=bSolveTriangular(T[0],X[0],M_ROW(T),M_ROW(T),1,0,0))
       {
        MatrixError(err); 
        return (TMatrix)NULL; 
       }
 return X;
}

/*
   Assumes
     X : m x m matrix or null pointer
     T : m x m lower triangular matrix

   Results
     X = inverse(T)

   Returns
     Returns X upon success, null otherwise

   Notes
     The matrices X and T do not have to be distinct.  Uses back substitution
     to form the inverse.  A null return indicates that either T was null,
     Y was singular, or unable to allocate required memory.  Call
     GetMatrixError() to determine which of these occured.
*/
TMatrix Inverse_LT(TMatrix X, TMatrix T)
{
 PRECISION *ptr;
 int err;
 if (!T)
  {
   MatrixError(NULL_ERR);
   return (TMatrix)NULL; 
  }
 if (M_ROW(T) != M_COL(T))
  {
   MatrixError(SIZE_ERR); 
   return (TMatrix)NULL; 
  }
 if (!X)
   {
    if (X=IdentityMatrix(X,M_ROW(T)))
     if (err=bSolveTriangular(T[0],X[0],M_ROW(T),M_ROW(T),0,0,0))
      {
       MatrixError(err); 
       FreeMatrix(X); 
       return (TMatrix)NULL; 
      }
   }
  else
   if (X == T)
     {
      if (!(ptr=(PRECISION*)malloc(M_ROW(T)*M_ROW(T)*sizeof(PRECISION))))
       {
        MatrixError(MEM_ERR); 
        return (TMatrix)NULL; 
       }
      memcpy(ptr,T[0],M_ROW(T)*M_ROW(T)*sizeof(PRECISION));
      if (X=IdentityMatrix(X,M_ROW(T)))
       if (err=bSolveTriangular(ptr,X[0],M_ROW(T),M_ROW(T),0,0,0))
        {
         MatrixError(err); 
         free(ptr); 
         return (TMatrix)NULL; 
        }     
      free(ptr);
     }
    else
     if (X=IdentityMatrix(X,M_ROW(T)))
      if (err=bSolveTriangular(T[0],X[0],M_ROW(T),M_ROW(T),0,0,0))
       {
        MatrixError(err); 
        return (TMatrix)NULL; 
       }
 return X;
}
/******************************************************************************/
/******************************************************************************/

/******************************************************************************/
/*************************** Miscellaneous Routines ***************************/
/******************************************************************************/
/*
   Assumes
     x : m-vector

   Results
     returns norm of x.
*/
PRECISION Norm(TVector x)
{
 int i;
 PRECISION result;
 if (!x)
  {
   MatrixError(NULL_ERR);
   return 0.0;
   }
 ClearMatrixError();
 for (result=x[0]*x[0], i=V_DIM(x)-1; i > 0; i--)
  result+=x[i]*x[i];
 return sqrt(result);
}

/*
   Assumes
     x : m-vector
     y : m-vector

   Results
     returns dot product of x and y.

   Notes
     On error returns 0.0.  Call GetMatrixError() to 
     determine the type of error.
*/
PRECISION DotProduct(TVector x, TVector y)
{
 PRECISION result;
 int i;
 if (!x || !y)
  {
   MatrixError(NULL_ERR);
   return 0.0;
  }
 if (V_DIM(x) != V_DIM(y))
  {
   MatrixError(SIZE_ERR);
   return 0.0;
  }
 ClearMatrixError();
 for (result=x[0]*y[0], i=V_DIM(x)-1; i > 0; i--) result+=x[i]*y[i];
 return result;
}

/*
   Assumes
     x : m-vector
     y : m-vector
     S : m x m matrix

   Results
     returns x'*S*y

   Notes
     In order for S to be a inner product, S must be positive definite
     and symmetric.  A zero return could indicate a error condition.  
     Call GetMatrixError() to determine if an error has occured.
*/
PRECISION InnerProduct(TVector x, TVector y, TMatrix S)
{
 PRECISION result=0.0, tmp;
 int i, j;
 if (!x || !y || !S)
  {
   MatrixError(NULL_ERR);
   return 0.0;
  }
 if ((V_DIM(x) != M_ROW(S)) || (V_DIM(y) != M_COL(S)))
  {
   MatrixError(SIZE_ERR);
   return 0.0;
  }
 ClearMatrixError();
 for (i=V_DIM(x)-1; i >= 0; i--)
  {
   for (tmp=S[i][0]*y[0], j=V_DIM(y)-1; j > 0; j--) tmp+=S[i][j]*y[j];
   result+=x[i]*tmp;
  }
 return result;
}

/*
   Assumes
     X : m x n matrix
     y : m-vector
     z : n-vector

   Results
     Returns X = y' * z upon success and a null pointer on failure.
*/
TMatrix OuterProduct(TMatrix X, TVector y, TVector z)
{
  int i, j;
  if (!y || !z)
    {
      MatrixError(NULL_ERR);
      return (TMatrix)NULL;
    }
  if (!X)
    { if (!(X=CreateMatrix(V_DIM(y),V_DIM(z)))) return (TMatrix)NULL; }
  else
    if ((M_ROW(X) != V_DIM(y)) || (M_COL(X) != V_DIM(z)))
      {
	MatrixError(SIZE_ERR);
	return (TMatrix)NULL;
      }
  for (i=V_DIM(y)-1; i >= 0; i--)
    for (j=V_DIM(z)-1; j >= 0; j--)
      X[i][j]=y[i]*z[j];
  return X;
}

/*
   Assumes
     X : m x m matrix

   Results
     Returns the determinate of X upon success.  Returns
     0 upon failure.  Call GetMatrixError() to determine
     failure.

   Notes
     Uses the LU decomposition to compute the determinate.
*/
PRECISION Determinant_LU(TMatrix X)
{
 PRECISION rtrn=0.0, *LU;
 int i, sgn, *p;
 if (!X)
   MatrixError(NULL_ERR);
  else
   if (M_ROW(X) != M_COL(X))
     MatrixError(SIZE_ERR);
    else
     if (!(p=(int*)malloc(M_ROW(X)*sizeof(int))))
       MatrixError(MEM_ERR); 
      else
       {
        if (!(LU=(PRECISION*)malloc(M_ROW(X)*M_ROW(X)*sizeof(PRECISION))))
          MatrixError(MEM_ERR);
         else
          {  
           memcpy(LU,X[0],M_ROW(X)*M_ROW(X)*sizeof(PRECISION));
           if (!bLU(p,LU,M_ROW(X),M_ROW(X)))
            { 
             for (sgn=1, i=M_ROW(X)-2; i >= 0; i--) 
              if (p[i] != i) sgn=-sgn;
             for (rtrn=0.0, i=M_ROW(X)*M_ROW(X)-1; i >= 0; i-=M_ROW(X)+1)
              if (LU[i] < 0.0)
                {
                 rtrn+=log(-LU[i]);
                 sgn=-sgn;
                }
               else
                if (LU[i] > 0.0)
                  rtrn+=log(LU[i]);
                 else
                  break;
             rtrn=(i >= 0) ? 0.0 : sgn*exp(rtrn);
            }   
           MATRIX_ERROR=0;       
           free(LU);
          } 
        free(p);
       }
 return rtrn;
}

/*
   Assumes
     X : m x m matrix

   Results
     Returns the determinate of X upon success.  Returns
     0 upon failure.  Call GetMatrixError() to determine
     failure.

   Notes
     Uses the QR decomposition to compute the determinate.
*/
PRECISION Determinant_QR(TMatrix X)
{
 PRECISION rtrn=0.0, *R;
 int i, err, sgn;
 if (!X)
   MatrixError(NULL_ERR);
  else 
   if (M_ROW(X) != M_COL(X))
     MatrixError(SIZE_ERR);
    else
     if (!(R=(PRECISION*)malloc(M_ROW(X)*M_ROW(X)*sizeof(PRECISION))))
       MatrixError(MEM_ERR);
      else
       {  
        memcpy(R,X[0],M_ROW(X)*M_ROW(X)*sizeof(PRECISION));
        if (err=bQR((PRECISION*)NULL,R,M_ROW(X),M_ROW(X)))
          MatrixError(err);
         else
          { 
           MATRIX_ERROR=0;
           sgn=(M_ROW(X) % 2) ? 1 : -1;
           for (rtrn=0.0, i=M_ROW(X)*M_ROW(X)-1; i >= 0; i-=M_ROW(X)+1)
            if (R[i] < 0.0)
              {
               rtrn+=log(-R[i]);
               sgn=-sgn;
              }
             else
              if (R[i] > 0.0)
                rtrn+=log(R[i]);
               else
                {
                 free(R);
                 return 0.0;
                }
           rtrn=sgn*exp(rtrn);
          }
        free(R);
       }
 return rtrn;
}

/*
   Assumes
     X : m x m matrix

   Results
     Returns the trace of X.
*/
PRECISION Trace(TMatrix X)
{
 PRECISION trace=0.0;
 int i;
 if (!X)
   MatrixError(NULL_ERR);
  else
   if (M_ROW(X) != M_COL(X)) 
     MatrixError(SIZE_ERR);
    else
     for (trace=X[0][0], i=M_ROW(X)-1; i > 0; i--) trace+=X[i][i];
 return trace;
}

/*
   Assumes
     X : m x n matrix

   Results
     Returns the rank of X.  Return -1 upon error.  Use GetMatrixERROR()
     to determine error type.


   Notes
     Uses the singular value decomposition to compute the rank.
*/
int Rank_SVD(TMatrix X)
{
 PRECISION min;
 int rank=-1, i;
 TMatrix U, V;
 TVector d;
 if (!X)
   MatrixError(NULL_ERR);
  else
   if (!(U=CreateMatrix(M_ROW(X),M_COL(X))))
     MatrixError(MEM_ERR);
    else
     {
      if (!(V=CreateMatrix(M_COL(X),M_COL(X))))
        MatrixError(MEM_ERR);
       else
        {
         if (!(d=CreateVector(M_COL(X))))
           MatrixError(MEM_ERR);
          else
           {
            if (i=SVD(U,d,V,X))
              MatrixError(i);
             else
              {
               for (min=d[0], i=V_DIM(d)-1; i > 0; i--) 
                if (d[i] > min) 
                 min=d[i];
               /*min*=((M_ROW(X) < M_COL(X)) ? M_ROW(X) : M_COL(X))*MACHINE_EPSILON;*/
               min*=SQRT_MACHINE_EPSILON;
               for (i=(rank=V_DIM(d))-1; i >= 0; i--) 
                if (d[i] < min) rank--;
              }
            FreeVector(d);
           }
         FreeMatrix(V);
        }
      FreeMatrix(U);
     }
 return rank;
}

/*
   Assumes
     x : n-vector or null pointer
     Y : n x (n-1) matrix

   Results
     The vector x is set to the cross product of the columns of Y.

   Notes
     The cross product of the columns of Y is the vector such that

      (1) Y'x = 0
      (2) det([Y x]) >= 0
      (3) norm(x) = volume of the parallelpiped spanned by the columns of Y
     
     Uses the LU decomposition to compute the cross product.
*/

TVector CrossProduct_LU(TVector x, TMatrix Y)
{
 int i, j, sgn;
 PRECISION s;
 TMatrix X;
 TPermutation P;
 TVector z=(TVector)NULL;
 if (!Y)
  {
   MatrixError(NULL_ERR);
   return (TVector)NULL;
  }
 if (M_ROW(Y) != M_COL(Y)+1)
  { 
   MatrixError(SIZE_ERR); 
   return (TVector)NULL;
  }
 if (!x)
   {
    if (!(z=x=CreateVector(M_ROW(Y))))
     return (TVector)NULL;
   }
  else
   if (M_ROW(Y) != V_DIM(x)) 
    { 
     MatrixError(SIZE_ERR); 
     return (TVector)NULL; 
    }
 if (X=CreateMatrix(M_ROW(Y),M_COL(Y)))
  {
   if (P=CreatePermutation(M_ROW(Y)))
    {
     if (LU(P,X,Y))
       {
        x[M_ROW(Y)-1]=1.0;
        for (i=M_ROW(Y)-2; i >= 0; i--)
         {
          x[i]=0.0;
          for (j=M_ROW(Y)-1; j > i; j--)
           x[i]-=x[j]*X[j][i]; 
         }
        x=ProductTransposeVP(x,x,P);
        for (sgn=1, i=M_ROW(X)-2; i >= 0; i--) 
         if (P[i] != i) sgn=-sgn;
        for (s=0.0, i=M_ROW(X)-2; i >= 0; i--)
         if (X[i][i] < 0.0)
           {
            s+=log(-X[i][i]);
            sgn=-sgn;
           }
          else
           if (X[i][i] > 0.0)
             s+=log(X[i][i]);
            else
             break;
        s=(i >= 0) ? 0.0 : sgn*exp(s);
        ProductSV(x,s,x);
       }
      else
       {
        FreePermutation(P); 
        FreeMatrix(X);         
        if (z) FreeVector(z);
        return (TVector)NULL;
       }
     FreePermutation(P);      
    }
   FreeMatrix(X);
  }
 return x;
}

/*
   Assumes
     x : n-vector or null pointer
     Y : n x (n-1) matrix

   Results
     The vector x is set to the cross product of the columns of Y.

   Notes
     The cross product of the columns of Y is the vector such that

      (1) Y'x = 0
      (2) det([Y x]) >= 0
      (3) norm(x) = volume of the parallelpiped spanned by the columns of Y
     
     Uses the QR decomposition to compute the cross product.
*/

TVector CrossProduct_QR(TVector x, TMatrix Y)
{
 int i, sgn;
 PRECISION s;
 TMatrix Q, R;
 TVector z=(TVector)NULL;;

 if (!Y)
  {
   MatrixError(NULL_ERR); 
   return (TVector)NULL; 
  }
 if (M_ROW(Y) != M_COL(Y)+1)
  {
   MatrixError(SIZE_ERR);
   return (TVector)NULL; 
  }
 if (!x) 
   {
    if (!(z=x=CreateVector(M_ROW(Y)))) 
     return (TVector)NULL;
   }
  else
   if (M_ROW(Y) != V_DIM(x)) 
    { 
     MatrixError(SIZE_ERR); 
     return (TVector)NULL; 
    }
 if (Q=CreateMatrix(M_ROW(Y),M_ROW(Y)))
  {
   if (R=CreateMatrix(M_ROW(Y),M_COL(Y)))
    {
     if (QR(Q,R,Y))
       {
        sgn=(M_ROW(Y) % 2) ? 1 : -1;
        for (s=0.0, i=M_ROW(Y)-2; i >= 0; i--)
         if (R[i][i] < 0)
           {
            s+=log(-R[i][i]);
            sgn=-sgn;
           }
          else
           if (R[i][i] > 0)
             s+=log(R[i][i]);
            else
             break;
        s=(i >= 0) ? 0.0 : sgn*exp(s);
        for (i=M_ROW(Y)-1; i >= 0; i--)
         x[i]=s*Q[i][M_ROW(Y)-1];
       }
      else
       {
        FreeMatrix(R);
        FreeMatrix(Q); 
        if (z) FreeVector(z);
        return (TVector)NULL;
       }
     FreeMatrix(R);      
    }
   FreeMatrix(Q);
  }
 return x;
}
/******************************************************************************/
/******************************************************************************/

/******************************************************************************/
/***************************** Kronecker Routines *****************************/
/******************************************************************************/
TVector Vec(TVector x, TMatrix Y)
{
 if (!Y)
  {
   MatrixError(NULL_ERR);
   return (TVector)NULL;
  }
 if (!x)
   {
    if (!(x=CreateVector(M_ROW(Y)*M_COL(Y))))
    return (TVector)NULL;
   }
  else
   if (M_ROW(Y)*M_COL(Y) != V_DIM(x))
    {
     MatrixError(SIZE_ERR);
     return (TVector)NULL;
    }
 bTranspose(x,Y[0],M_COL(Y),M_ROW(Y));
 return x;
}

TMatrix KroneckerProduct(TMatrix X, TMatrix Y, TMatrix Z)
{
 int i, j, r, s, u, v;
 PRECISION tmp;
 if (!Y || !Z)
  {
   MatrixError(NULL_ERR);
   return (TMatrix)NULL;
  }
 if (!X)
   {
    if (!(X=CreateMatrix(M_ROW(Y)*M_ROW(Z),M_COL(Y)*M_COL(Z))))
    return (TMatrix)NULL;
   }
  else
   if ((M_ROW(Y)*M_ROW(Z) != M_ROW(X)) || (M_COL(Y)*M_COL(Z) != M_COL(X)))
    {
     MatrixError(SIZE_ERR);
     return (TMatrix)NULL;
    }
 for (i=M_ROW(Y)-1, u=i*M_ROW(Z); i >= 0; u-=M_ROW(Z), i--)
  for (j=M_COL(Y)-1, v=j*M_COL(Z); j >= 0; v-=M_COL(Z), j--)
   {
    tmp=Y[i][j];
    for (r=M_ROW(Z)-1; r >= 0; r--)
     for (s=M_COL(Z)-1; s >= 0; s--)
       X[u+r][v+s]=tmp*Z[r][s];
   }
 return X;
}
/******************************************************************************/
/******************************************************************************/

/******************************************************************************/
/****************************** Output Routines *******************************/
/******************************************************************************/
/*
   Assumes
     x : m-vector

   Results
     Prints x to f using format.

   Notes
     If format is the NULL pointer, uses default formating.  The elements
     of x are output as double independently of the setting of PRECISION.
*/
void PrintVector(FILE *f, TVector x, char *format)
{
 int i, m;
 if (!x) return;
 m=V_DIM(x);
 if (!format) format="%7.3lf\t";
 for (i=0; i < m; i++) fprintf(f,format,(double)(x[i]));
 fprintf(f,"\n");
}

/*
   Assumes
     X : m x n matrix

   Results
     Prints X to f using format.

   Notes
     If format is the NULL pointer, uses default formating.  The elements
     of X are output as double independently of the setting of PRECISION.
*/
void PrintMatrix(FILE *f, TMatrix X, char *format)
{
 int i, j, m, n;
 if (!X) return;
 m=M_ROW(X); n=M_COL(X);
 if (!format) format="%7.3lf\t";
 for (i=0; i < m; i++)
  {
   for (j=0; j < n; j++) fprintf(f,format,(double)(X[i][j]));
   fprintf(f,"\n");
  }
}

/*
   Assumes
     x : m vector
     f : an pointer to an open file

   Results
     Reads in matrix X from f.  Return 1 on success, 0 otherwise.
*/
int ReadVector(FILE *f, TVector x)
{
 int i, m;
 if (!x) return 0;
 for (i=0, m=V_DIM(x); i < m; i++)
#if (PRECISION_SIZE == 8)
 if (fscanf(f," %lf ",&(x[i])) != 1) return 0; 
#else
 if (fscanf(f," %f ",&(x[i])) != 1) return 0; 
#endif
 return 1;
}

/*
   Assumes
     X : m x n matrix
     f : an pointer to an open file

   Results
     Reads in matrix X from f.  Return 1 on success, 0 otherwise.
*/
int ReadMatrix(FILE *f, TMatrix X)
{
 int i, j, m, n;
 if (!X) return 0;
 m=M_ROW(X); n=M_COL(X);
 for (i=0; i < m; i++)
  for (j=0; j < n; j++)
#if (PRECISION_SIZE == 8)
   if (fscanf(f," %lf ",&(X[i][j])) != 1) return 0;
#else
  if (fscanf(f," %f ",&(X[i][j])) != 1) return 0; 
#endif
 return 1;
}

/*
   Assumes
     f : pointer to open binary file
     x : m matrix

   Results
     Outputs x in binary format.  

   Returns
     1 upon success, 0 upon failure

   Notes
     Format
       int
             0x1004 - vector float
             0x2004 - matrix float
             0x1008 - vector double
             0x2008 - mattrix double

       int   rows if matrix, dimension if vector
       int   columns if matrix, absent if vector
       data  m binary floating point numbers

*/
int OutVectorFloat(FILE *f, TVector x)
{
 float y;
 int i, format_code=0x1000+sizeof(float);
 if (!x) return 0;
 if (fwrite(&format_code,sizeof(int),1,f) != 1) return 0;
#if (PRECISION_SIZE == 8)
 if (fwrite(((int *)x)-1,sizeof(int),1,f) != 1) return 0;
 for (i=V_DIM(x)-1; i >= 0; i--) 
  {
   y=(float)x[i];
   if (fwrite(&y,sizeof(float),1,f) != 1) return 0;
  }
#else
 if (fwrite(((int *)x)-1,V_DIM(x)*sizeof(float)+sizeof(int),1,f) != 1) return 0; 
#endif
 return 1;
}

/*
   Assumes
     f : pointer to open binary file
     x : m matrix

   Results
     Outputs x in binary format.  Returns 1 on success and 0 otherwise.

   Notes
     Format
       int
             0x1004 - vector float
             0x2004 - matrix float
             0x1008 - vector double
             0x2008 - mattrix double

       int   rows if matrix, dimension if vector
       int   columns if matrix, absent if vector
       data  m binary floating point numbers

*/
int OutVectorDouble(FILE *f, TVector x)
{
 double y;
 int i, format_code=0x1000+sizeof(double);
 if (!x) return 0;
 if (fwrite(&format_code,sizeof(int),1,f) == 1)
#if (PRECISION_SIZE == 8)
 if (fwrite(((int *)x)-1,sizeof(int),1,f) != 1) return 0;
  for (i=V_DIM(x)-1; i >= 0; i--) 
   {
    y=(double)x[i];
    if (fwrite(&y,sizeof(double),1,f) != 1) return 0;
   }
#else
 if (fwrite(((int *)x)-1,V_DIM(x)*sizeof(double)+sizeof(int),1,f) != 1) return 0; 
#endif
 return 1;
}


/*
   Assumes
     f : pointer to open binary file
     X : m x n matrix

   Results
     Outputs X in binary format.  Returns 1 on success and 0 otherwise.

   Notes
     Format
       int
             0x1004 - vector float
             0x2004 - matrix float
             0x1008 - vector double
             0x2008 - mattrix double

       int   rows if matrix, dimension if vector
       int   columns if matrix, absent if vector
       data  m x n binary floating point numbers
*/
int OutMatrixFloat(FILE *f, TMatrix X)
{
 float y;
 int i, j, format_code=0x2000+sizeof(float);
 if (!X) return 0;
 if (fwrite(&format_code,sizeof(int),1,f) != 1) return 0;
#if (PRECISION_SIZE == 8)
 if (fwrite(((int *)X)-2,sizeof(int),2,f) != 2) return 0;
  for (i=M_ROW(X)-1; i >= 0; i--)
   for (j=M_COL(X)-1; j >= 0; j--)
    {
     y=(float)X[i][j];
     if (fwrite(&y,sizeof(float),1,f) != 1) return 0;
    } 
#else
 if (fwrite(((int *)X)-2,M_ROW(X)*M_COL(X)*sizeof(float)+2*sizeof(int),1,f) != 1) return 0;
#endif
 return 1;
}

/*
   Assumes
     f : pointer to open binary file
     X : m x n matrix

   Results
     Outputs X in binary format.  Returns 1 on success and 0 otherwise.

   Notes
     Format
       int
             0x1004 - vector float
             0x2004 - matrix float
             0x1008 - vector double
             0x2008 - mattrix double

       int   rows if matrix, dimension if vector
       int   columns if matrix, absent if vector
       data  m x n binary floating point numbers
*/
int OutMatrixDouble(FILE *f, TMatrix X)
{
 double y;
 int i, j, format_code=0x2000+sizeof(double);
 if (!X) return 0;
 if (fwrite(&format_code,sizeof(int),1,f) != 1) return 0;
#if (PRECISION_SIZE == 8)
 if (fwrite(((int *)X)-2,sizeof(int),2,f) != 2) return 0;
  for (i=M_ROW(X)-1; i >= 0; i--)
   for (j=M_COL(X)-1; j >= 0; j--)
    {
     y=X[i][j];
     if (fwrite(&y,sizeof(double),1,f) != 1) return 0;
    }
#else
 if (fwrite(((int *)X)-2,M_ROW(X)*M_COL(X)*sizeof(double)+2*sizeof(int),1,f) != 1) return 0; 
#endif
 return 1;
}


/*
   Assumes
     f : pointer to open binary file
     x : m matrix or null pointer

   Results
     Reads x from binary format.  Returns 1 on success and 0 otherwise.

   Notes
     Format
       int
             0x1004 - vector float
             0x2004 - matrix float
             0x1008 - vector double
             0x2008 - mattrix double

       int   number rows if matrix, dimension if vector
       int   number columns if matrix, absent if vector
       data  m binary floating point numbers
*/
TVector InVector(FILE *f, TVector x)
{
 int i, d[2], precision, del=0, position=ftell(f);
 void *y=(void*)NULL;

 if (fread(d,sizeof(int),2,f) != 2) goto EXIT_ERROR;

 switch(d[0])
  {
   case 0x1000+sizeof(float): precision=sizeof(float); break;
   case 0x1000+sizeof(double): precision=sizeof(double); break;
   default: goto EXIT_ERROR;
  }

 i=d[1];

 if (!x)
   {
    x=CreateVector(i);
    del=1;
   }
  else
    if (V_DIM(x) != i) goto EXIT_ERROR;

 if (precision != sizeof(PRECISION))
   {
    if (!(y=malloc(i*precision))) MatrixError(MEM_ERR);
    if (fread(y,i*precision,1,f) != 1) goto EXIT_ERROR;
    if (precision == sizeof(float))
      while (--i >= 0) x[i]=((float*)y)[i];
     else
      while (--i >= 0) x[i]=((double*)y)[i];
    free(y);
   }
  else
   if (fread(x,i*sizeof(PRECISION),1,f) != 1) goto EXIT_ERROR;

 return x;

EXIT_ERROR:
 fseek(f,position,SEEK_SET);
 if (del) FreeVector(x);
 if (y) free(y);
 return (TVector)NULL;
}

/*
   Assumes
     f : pointer to open binary file
     X : m x n matrix or null pointer

   Results
     Reads x from binary format.  Returns 1 on success and 0 otherwise.

   Notes
     Format
       int
             0x1004 - vector float
             0x2004 - matrix float
             0x1008 - vector double
             0x2008 - mattrix double

       int   number rows if matrix, dimension if vector
       int   number columns if matrix, absent if vector
       data  m x n binary floating point numbers
*/
TMatrix InMatrix(FILE *f, TMatrix X)
{
 int i, d[3], precision, del=0, position=ftell(f);
 void *Y=(void*)NULL;

 if (fread(d,sizeof(int),3,f) != 3) goto EXIT_ERROR;

 switch(d[0])
  {
   case 0x2000+sizeof(float): precision=sizeof(float); break;
   case 0x2000+sizeof(double): precision=sizeof(double); break;
   default: goto EXIT_ERROR;
  }

 if (!X)
   {
    X=CreateMatrix(d[1],d[2]);
    del=1;
   }
  else
   if ((M_ROW(X) != d[1]) || (M_COL(X) != d[2])) goto EXIT_ERROR;

 i=d[1]*d[2];

 if (precision != sizeof(PRECISION))
   {
    if (!(Y=malloc(i*precision))) MatrixError(MEM_ERR);
    if (fread(Y,i*precision,1,f) != 1) goto EXIT_ERROR;
    if (precision == sizeof(float))
      while (--i >= 0) X[0][i]=((float*)Y)[i];
     else
      while (--i >= 0) X[0][i]=((double*)Y)[i];
    free(Y);
   }
  else
   if (fread(X[0],i*sizeof(PRECISION),1,f) != 1) goto EXIT_ERROR;

 return X;

EXIT_ERROR:
 fseek(f,position,SEEK_SET);
 if (del) FreeMatrix(X);
 if (Y) free(Y);
 return (TMatrix)NULL;
}
/******************************************************************************/
/******************************************************************************/

/******************************************************************************/
/*************************** Matrix Decompositions ****************************/
/******************************************************************************/
/*
   Assumes
     U : m x n matrix
     d : n-vector
     V : n x n matrix
     A : m x n matrix

   Returns
     1 upon success, and 0 on failure.  

   Results
     Finds U, V and d such that A = U * diag(d) * V'.  U and V are orthogonal
     matrices and the elemets of d are non-negative.  U is orthogonal in the
     sense that U'*U is a diagonal matrix with ones and zeros along the
     diagonal.
*/
int SVD(TMatrix U, TVector d, TMatrix V, TMatrix A)
{
 if (!U || !d || !V || !A)
  {
   MatrixError(NULL_ERR);
   return 0;
  }
 if ((M_ROW(U) != M_ROW(A)) || (M_COL(U) != M_COL(A))
       || (M_ROW(V) != M_COL(A)) || (M_COL(V) != M_COL(A))
            || (V_DIM(d) != M_COL(A)))
  {
   MatrixError(SIZE_ERR);
   return 0;
  }
 if (U == V)
   {
     MatrixError(ARG_ERR);
     return 0;
   }
 if (U != A) memcpy(U[0],A[0],M_ROW(A)*M_COL(A)*sizeof(PRECISION));
 if (bSVD(U[0],d,V[0],M_ROW(A),M_COL(A)))
  {
   MatrixError(ITER_ERR);
   return 0; 
  } 
 return 1;
}

/*
   Assumes
     U : m x m matrix
     X : m x m symmetric positive definite matrix

   Results
     X = U' * U, where U is a upper triangular matrix with positive diagonal.

   Returns
     1 success
     0 failure, call GetMatrixError() to

   Notes
     Failure usually indicates X is not positive definite.  Only the upper
     half of X is used.  The matrices U and X need not be distinct.
*/
TMatrix CholeskyUT(TMatrix U, TMatrix X)
{
 int err;
 if (!X) 
  {
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (M_ROW(X) != M_COL(X))
  {
   MatrixError(SIZE_ERR); 
   return (TMatrix)NULL; 
  }
 if (!U)
   {
    if (!(U=EquateMatrix((TMatrix)NULL,X)))
     return (TMatrix)NULL;
    if (err=bCholesky(U[0],M_ROW(X),1,0))
     {
      FreeMatrix(U);
      MatrixError(err); 
      return (TMatrix)NULL; 
     }
   }
  else
   {
    if (U != X) 
     {
      if ((M_ROW(X) != M_ROW(U)) || (M_ROW(X) != M_COL(U)))
       {
        MatrixError(SIZE_ERR);
        return (TMatrix)NULL;
       }
      memcpy(U[0],X[0],M_ROW(X)*M_ROW(X)*sizeof(PRECISION));
     }
    if (err=bCholesky(U[0],M_ROW(X),1,0))
     {
      MatrixError(err); 
      return (TMatrix)NULL; 
     }
   }
 return U; 
}

/*
   Assumes
     L : m x m matrix
     X : m x m symmetric positive definite matrix

   Results
     X = L' * L, where L is a lower triangular matrix with positive diagonal.

   Returns
     1 success
     0 failure, call GetMatrixError() to

   Notes
     Failure usually indicates X is not positive definite.  Only the upper
     half of X is used.  The matrices L and X need not be distinct.
*/
TMatrix CholeskyLT(TMatrix L, TMatrix X)
{
 int err;
 if (!X) 
  {
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL;
  }
 if (M_ROW(X) != M_COL(X))
  {
   MatrixError(SIZE_ERR); 
   return (TMatrix)NULL; 
  }
 if (!L)
   {
    if (!(L=EquateMatrix((TMatrix)NULL,X)))
     return (TMatrix)NULL;
    if (err=bCholesky(L[0],M_ROW(X),0,0))
     {
      FreeMatrix(L);
      MatrixError(err); 
      return (TMatrix)NULL; 
     }
   }
  else
   {
    if (L != X) 
     {
      if ((M_ROW(X) != M_ROW(L)) || (M_ROW(X) != M_COL(L)))
       {
        MatrixError(SIZE_ERR);
        return (TMatrix)NULL;
       }
      memcpy(L[0],X[0],M_ROW(X)*M_ROW(X)*sizeof(PRECISION));
     }
    if (err=bCholesky(L[0],M_ROW(X),0,0))
     {
      MatrixError(err); 
      return (TMatrix)NULL; 
     }
   }
 return L; 
}

/*
   Assumes
     Q : n x n matrix or null pointer
     R : n x m matrix
     X : n x m matrix

   Results
     Finds an orthogonal matrix Q and an upper triangular matrix R such that

                                   X = Q * R

     When using bmatrix.c, det(Q) = (-1)^s, where s=min(m-1,n).

   Returns
     1 - Success
     0 - Error, call GetMatrixError() to determine the type of error made.

   Notes
     The matrices R and X do not have to be distinct.  The QR decomposition is
     formed using Householder matrices without pivoting.
*/
int QR(TMatrix Q, TMatrix R, TMatrix X)
{
 int err;
 if (!X || !R) 
   err=NULL_ERR; 
  else
   if ((M_ROW(X) != M_ROW(R)) || (M_COL(X) != M_COL(R)))
     err=SIZE_ERR;
    else
     if (Q == R)
       err=ARG_ERR;
      else
       if (Q && ((M_ROW(Q) != M_ROW(X)) || (M_COL(Q) != M_ROW(X))))
	 err=SIZE_ERR;
        else 
         {
          if (R != X) memcpy(R[0],X[0],M_ROW(X)*M_COL(X)*sizeof(PRECISION));
          err=bQR(Q ? Q[0] : (PRECISION*)NULL,R[0],M_ROW(X),M_COL(X));
         }
 if (!err) return 1;
 MatrixError(err);
 return 0;
}

/*
   Assumes
     P : m x m permutation matrix
     X : m x n matrix
     A : m x n matrix invertible matrix

   Results
     Computes the LU decomposition of A with partial pivoting.  The 
     decomposition is

                                A = P * L * U

     where P is a permutation matrix, L is lower triangular with ones along
     the diagonal, and U is upper triangular.  These matrices are stored as
     follows.

        L is m x k, where k is the smaller of n and m, and is stored in the
        lower half of LU.  The diagonal of L is not stored.

        U is k x n, where k is the smaller of n and m, and is stored in the
        upper half of X, including the diagonal.   

        P is the integer representation of a permutation matrix.  See the
        header file matrix.h for a description of its internal 
        reqresentation.


   Returns
     1 : success
     0 : failure, call GetMatrixError() to determine the cause.

   Notes
     The matrices X and A do not have to be distinct.  Uses partial pivoting.
*/
int LU(TPermutation P, TMatrix X, TMatrix A)
{
 if (!P || !X || !A) 
  {
   MatrixError(NULL_ERR); 
   return 0; 
  }
 if (P_DIM(P) != M_ROW(A))
  {
   MatrixError(SIZE_ERR); 
   return 0; 
  }
 if (X != A) 
  {
   if ((M_ROW(A) != M_ROW(X)) || (M_COL(A) != M_COL(X)))
    {
     MatrixError(SIZE_ERR);
     return 0;
    }
   memcpy(X[0],A[0],M_ROW(A)*M_COL(A)*sizeof(PRECISION));
  }
 bLU(P,X[0],M_ROW(A),M_COL(A));
 P_USE(P)=(M_ROW(A) < M_COL(A)) ? M_ROW(A) : M_COL(A);
 return 1;
}

/*
   Assumes
     x  : n vector
     y  : n vector
     LU : n x n matrix with non-zero diagonal
     p  : integer array of length n with i <= p[i] < n for 0 <= i < n.

   Results
     Solves P * L * U * x = y, where P is a permutation matrix, L is lower
     triangular with ones along the diagonal, and U is upper triangular.

   Notes
     The vectors x and y do not have to be distinct.  The integer array p
     must represent a permutation.

     The matrices P, L, and U are stored as follows.

        U is stored in the upper half of LU, including the diagonal.

        L is stored in the lower half of LU.  The diagonal of L is not stored.

        The matrix P is defined by P*e(i)=e(p[i]), where e(j) is the jth column
        of the n x n identity matrix.  See the notes in matrix.h for further
        details on permutation matrices.
*/
TVector LU_SolveCol(TVector x, TVector y, TMatrix LU, TPermutation P)
{
 int i, j, n=V_DIM(y);
 TVector z=x;

 if (!y || !LU || !P)
  {
   MatrixError(NULL_ERR); 
   return (TVector)NULL; 
  }
 if ((V_DIM(y) != M_ROW(LU)) || (V_DIM(y) != M_COL(LU))) 
  {
   MatrixError(SIZE_ERR); 
   return (TVector)NULL; 
  }
 if (!x)
   {
    if (!(x=CreateVector(V_DIM(y)))) return (TVector)NULL;
   }
  else
   if (V_DIM(x) != V_DIM(y)) 
    { 
     MatrixError(SIZE_ERR); 
     return (TVector)NULL; 
    }
 
 if (x != y) memcpy(x,y,V_DIM(y)*sizeof(int));
 if (!bPermutationMultiply(P,x,V_DIM(y),1,P_USE(P),1,0) && 
       !bSolveUnitTriangular(LU[0],x,V_DIM(x),1,0,0,0) &&
         !bSolveTriangular(LU[0],x,V_DIM(x),1,1,0,0)) return x;
 if (!z) FreeVector(x);
 return (TVector)x;
}

/*
   Assumes
     x       : n vector
     y       : n vector
     LU      : n x n matrix with non-zero diagonal
     permute : integer array of length n

   Results
     Solves x * P * L * U = y, where P is a permutation matrix, L is lower
     triangular with ones along the diagonal, and U is upper triangular.

   Notes
     The vectors x and y do not have to be distinct.  The integer array permute
     must be of length at least n.

     The matrices P, L, and U are stored as follows.

        U is stored in the upper half of LU, including the diagonal.

        L is stored in the lower half of LU.  The diagonal of L is not stored.

        The matrix P is defined by P*e(i)=e(p[i]), where e(j) is the jth column
        of the n x n identity matrix.  See the notes in matrix.h for further
        details on permutation matrices.
*/
TVector LU_SolveRow(TVector x, TVector y, TMatrix LU, int *p)
{
 int i, j, n=V_DIM(y);
 PRECISION *z, sum;

 if (!x)
   x=CreateVector(n);
#ifdef CHECK_SIZE
  else
   if (V_DIM(x) != n) { MatrixError(SIZE_ERR); return (TVector)NULL; }
 if ((n != M_ROW(LU)) || (n != M_COL(LU))) { MatrixError(SIZE_ERR); return (TVector)NULL; }
#endif

 if (!(z=(PRECISION*)malloc(sizeof(PRECISION)*n))) { MatrixError(SIZE_ERR); return (TVector)NULL; }

 for (j=0; j < n; j++)
  {
   sum=y[j];
   for (i=j-1; i >= 0; i--) sum-=z[i]*LU[i][j];
   if (LU[j][j] == 0.0) MatrixError(SING_ERR);
   z[j]=sum/LU[j][j];
  }

 for (j=n-2; j >= 0; j--)
  {
   sum=z[j];
	for (i=j+1; i < n; i++) sum-=z[i]*LU[i][j];
   z[j]=sum;
  }

 for (i=n-1; i >= 0; i--) x[p[i]]=z[i];
 free(z);
 return x;
}

/******************************************************************************/
/**************************** Permutation Matrices ****************************/
/******************************************************************************/
TPermutation CreatePermutation(int m)
{
 TPermutation X;
 if (m <= 0) 
  {
   MatrixError(SIZE_ERR); 
   return (TPermutation)NULL; 
  }
 if (X=(TPermutation)malloc((m+2)*sizeof(int)))
   {
    X[0]=m;
    X[1]=0;
   }
  else
   MatrixError(MEM_ERR);
 return (TPermutation)(X+2);
}

void FreePermutation(TPermutation X)
{
 if (!X)
   MatrixError(NULL_ERR);
  else
   free(X-2);
}

/*
   Assumes
     X : m-permutation or null pointer
     Y : m-permutation

   Results
     X = Y.  If X is null pointer, X is created.

   Returns
     Returns X upon success and null on failure.  Call GetMatrixError() to
     determine the cause of failure.
*/
TPermutation EquatePermutation(TPermutation X, TPermutation Y)
{
 if (!Y) 
  {
   MatrixError(NULL_ERR); 
   return (TPermutation)NULL; 
  }
 if (!X)
   { 
    if (!(X=CreatePermutation(P_DIM(Y)))) 
     return (TPermutation)NULL; 
   }
  else
   if (P_DIM(X) != P_DIM(Y))
    { 
     MatrixError(SIZE_ERR); 
     return (TPermutation)NULL; 
    }
 memcpy(X-1,Y-1,(P_USE(Y)+1)*sizeof(int));
 return X;
}

TMatrix PermutationMatrix(TMatrix X, TPermutation Y)
{
 if (!Y)
  {
   MatrixError(NULL_ERR);
   return (TMatrix)NULL;
  }
 if (!X)
   {
    if (!(X=CreateMatrix(P_DIM(Y),P_DIM(Y))))
     {
      MatrixError(MEM_ERR);
      return (TMatrix)NULL;
     }
   }
  else
   if ((M_ROW(X) != P_DIM(Y)) || (M_COL(X) != P_DIM(Y)))
    {
     MatrixError(SIZE_ERR);
     return (TMatrix)NULL;
    }
 bPermutation(X[0],Y,P_DIM(Y),P_USE(Y),0);
 return X;
}

TMatrix ProductPM(TMatrix X, TPermutation Y, TMatrix Z)
{
 if (!Y || !Z)
  {
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (P_DIM(Y) != M_ROW(Z))
  {
   MatrixError(SIZE_ERR);
   return (TMatrix)NULL; 
  }
 if (!X)
   {
    if (!(X=EquateMatrix((TMatrix)NULL,Z))) 
     return (TMatrix)NULL;   
   }
  else
   if (X != Z)
    if ((M_ROW(X) != M_ROW(Z)) || (M_COL(X) != M_COL(Z)))
      {
       MatrixError(SIZE_ERR); 
       return (TMatrix)NULL; 
      }
     else
      memcpy(X[0],Z[0],M_ROW(Z)*M_COL(Z)*sizeof(PRECISION));
 bPermutationMultiply(Y,X[0],M_ROW(X),M_COL(X),P_USE(Y),0,0);
 return X;
}

TMatrix ProductMP(TMatrix X, TMatrix Y, TPermutation Z)
{
 if (!Y || !Z)
  {
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (M_COL(Y) != P_DIM(Z))
  {
   MatrixError(SIZE_ERR);
   return (TMatrix)NULL; 
  }
 if (!X)
   {
    if (!(X=EquateMatrix((TMatrix)NULL,Y))) 
     return (TMatrix)NULL;   
   }
  else
   if (X != Y)
    if ((M_ROW(X) != M_ROW(Y)) || (M_COL(X) != M_COL(Y)))
      {
       MatrixError(SIZE_ERR); 
       return (TMatrix)NULL; 
      }
     else
      memcpy(X[0],Y[0],M_ROW(Y)*M_COL(Y)*sizeof(PRECISION));
 bPermutationMultiply(Z,X[0],M_COL(X),M_ROW(X),P_USE(Z),1,1);
 return X;
}

TVector ProductPV(TVector x, TPermutation Y, TVector z)
{
 if (!Y || !z)
  {
   MatrixError(NULL_ERR); 
   return (TVector)NULL; 
  }
 if (P_DIM(Y) != V_DIM(z))
  {
   MatrixError(SIZE_ERR);
   return (TVector)NULL; 
  }
 if (!x)
   {
    if (!(x=EquateVector((TVector)NULL,z))) 
     return (TVector)NULL;   
   }
  else
   if (V_DIM(x) != V_DIM(z))
     {
      MatrixError(SIZE_ERR); 
      return (TVector)NULL; 
     }
    else
     if (x != z)
      memcpy(x,z,V_DIM(z)*sizeof(PRECISION));
 bPermutationMultiply(Y,x,V_DIM(x),1,P_USE(Y),0,0);
 return x;
}

TVector ProductVP(TVector x, TVector y, TPermutation Z)
{
 if (!y || !Z)
  {
   MatrixError(NULL_ERR); 
   return (TVector)NULL; 
  }
 if (V_DIM(y) != P_DIM(Z))
  {
   MatrixError(SIZE_ERR);
   return (TVector)NULL; 
  }
 if (!x)
   {
    if (!(x=EquateVector((TVector)NULL,y))) 
     return (TVector)NULL;   
   }
  else
   if (V_DIM(x) != V_DIM(y)) 
     {
      MatrixError(SIZE_ERR); 
      return (TVector)NULL; 
     }
    else
     if (x != y)
      memcpy(x,y,V_DIM(y)*sizeof(PRECISION));
 bPermutationMultiply(Z,x,V_DIM(x),1,P_USE(Z),1,1);
 return x;
}

TMatrix TransposeProductPM(TMatrix X, TPermutation Y, TMatrix Z)
{
 if (!Y || !Z)
  {
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (P_DIM(Y) != M_ROW(Z))
  {
   MatrixError(SIZE_ERR);
   return (TMatrix)NULL; 
  }
 if (!X)
   {
    if (!(X=EquateMatrix((TMatrix)NULL,Z))) 
     return (TMatrix)NULL;   
   }
  else
   if (X != Z)
    if ((M_ROW(X) != M_ROW(Z)) || (M_COL(X) != M_COL(Z)))
      {
       MatrixError(SIZE_ERR); 
       return (TMatrix)NULL; 
      }
     else
      memcpy(X[0],Z[0],M_ROW(Z)*M_COL(Z)*sizeof(PRECISION));
 bPermutationMultiply(Y,X[0],M_ROW(X),M_COL(X),P_USE(Y),1,0);
 return X;
}

TMatrix ProductTransposeMP(TMatrix X, TMatrix Y, TPermutation Z)
{
 if (!Y || !Z)
  {
   MatrixError(NULL_ERR); 
   return (TMatrix)NULL; 
  }
 if (M_COL(Y) != P_DIM(Z))
  {
   MatrixError(SIZE_ERR);
   return (TMatrix)NULL; 
  }
 if (!X)
   {
    if (!(X=EquateMatrix((TMatrix)NULL,Y))) 
     return (TMatrix)NULL;   
   }
  else
   if (X != Y)
    if ((M_ROW(X) != M_ROW(Y)) || (M_COL(X) != M_COL(Y)))
      {
       MatrixError(SIZE_ERR); 
       return (TMatrix)NULL; 
      }
     else
      memcpy(X[0],Y[0],M_ROW(Y)*M_COL(Y)*sizeof(PRECISION));
 bPermutationMultiply(Z,X[0],M_COL(X),M_ROW(X),P_USE(Z),0,1);
 return X;
}


void PrintPermutation(FILE *f, TPermutation X)
{
 int i, j, k;
 for (i=0; i < P_DIM(X); i++)
  fprintf(f,"%3d ",i);
 fprintf(f,"\n");
 for (i=0; i < P_DIM(X); i++)
  {
   for (j=i-1, k=(i < P_DIM(X)-1) ? X[i] : i; j >= 0; j--)
    if (k == X[j]) k=j;
   fprintf(f,"%3d ",k); 
  }
 fprintf(f,"\n");
}
/**/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/


