/*******************************************************************************
 A TVector is an array of floating points, however, sizeof(int) extra bytes
 are allocated to contain the length of the vector.  The length of the vector
 can obtained with the V_DIM macro.  Vectors must be created with calls to
 CreateVector and freed with calls to FreeVector.

 A TMatrix is an array of an array of floating points, but 2*sizeof(int) extra
 bytes are allocated to contain the number of rows and columns in the matrix.
 These values can be obtained with the M_ROW and M_COL macros.  Matrices must
 be created with calls to CreateMatrix and freed with calls to FreeMatrix.

 The precision (float or double) is controled by the define PRECISION contained
 in the file prcsn.h.

 The binary layout of a vector
      _ ____ ____     ______
     | |    |    |   |      |
     |d|v[0]|v[1]|...|v[d-1]|
     |_|____|____|   |______|

    d    : sizeof(int) bytes containing the dimension of the vector
    v[i] : sizeof(PRECISION) bytes containing the elements of the vector
    v    : pointer to the position v[0]


 The binary layout of a matrix
      _ _ ____   ______ _______   _________ _______   _________   _________   ___________
     | | |    | |      |       | |         |       | |         | |         | |           |
     |r|c|X[0]|.|X[r-1]|X[0][0]|.|X[0][c-1]|X[1][0]|.|X[1][c-1]|.|X[r-1][0]|.|X[r-1][c-1]|
     |_|_|____| |______|_______| |_________|_______| |_________| |_________| |___________|

    r       : sizeof(int) bytes containing the number of rows
    c       : sizeof(int) bytes containing the number of columns
    X[i]    : sizeof(PRECISION*) bytes containing pointer to position X[i][0]
    X[i][j] : sizeof(PRECISION) bytes containing the elements of the matrix
    X       : pointer to position X[0]


 PERMUTATION MATRICES
 Let I be the m x m identity matrix and let p be a permutation of {0,...,m-1}.
 The permutation matrix associated with p, is an m x m matrix P defined by
 
                              P[p(i)][j] = I[i][j]
                                
 Note that if e(i) is the ith column of I, then
 
                               e(p(i))' * P = e(i)' 

 Thus left multiplication by P permutes the rows.  This means that if A is a
 m x n matrix, then column i of A is column p(i) of P*A.  Right multiplication
 by P permutes the columns by the inverse of p.  

 Permutation matrices are stored as integer arrays of length m, with an extra 
 sizeof(int) bytes allocated to contain the length of the array.  The length of 
 the array can be obtained with the P_DIM macro.  Permutations must be created
 with calls to CreatePermutation and freed with calls to FreePermutation.

 If p is the integer array representing a permutation, then it must be the case 
 that i <= p[i] < m for all 0 <= i < m.  The representation is as a product of
 transpositions. 
     
                            (1,p[1])...(m-1,p[m-1])

 where (r,s) denotes the transposition

                                    ---
                                    | s   if k=r
                           (r,s)k = | r   if k=s
                                    | k   otherwise
                                    ---

*******************************************************************************/
#ifndef __MATRIX__
#define __MATRIX__

/* define to check sizes of matrices */
#define CHECK_SIZE
#define CHECK_NULL

#include "prcsn.h"
#include "m_err.h"

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <math.h>

/* Error Routines */
void MatrixError(int err);
void ClearMatrixError(void);
int  GetMatrixError(void);
int  SetMatrixErrorVerbose(int err);
int  SetMatrixErrorTerminate(int err);

#define V_DIM(x) (((int*)(x))[-1])
#define M_ROW(x) (((int*)(x))[-2])
#define M_COL(x) (((int*)(x))[-1])
#define P_DIM(x) (((int*)(x))[-2])
#define P_USE(x) (((int*)(x))[-1])

typedef PRECISION*  TVector;
typedef PRECISION** TMatrix;
typedef int*        TPermutation;

/* Permutations */
TPermutation CreatePermutation(int m);
void FreePermutation(TPermutation X);
TPermutation EquatePermutation(TPermutation X, TPermutation Y);
TMatrix PermutationMatrix(TMatrix X, TPermutation Y);
TMatrix ProductPM(TMatrix X, TPermutation Y, TMatrix Z);
TMatrix ProductMP(TMatrix X, TMatrix Y, TPermutation Z);
TVector ProductPV(TVector x, TPermutation Y, TVector z);
TVector ProductVP(TVector x, TVector y, TPermutation Z);
TMatrix TransposeProductPM(TMatrix X, TPermutation Y, TMatrix Z);
#define TransposeProductPV(x,Y,z) ProductVP(x,z,Y)
TMatrix ProductTransposeMP(TMatrix X, TMatrix Y, TPermutation Z);
#define ProductTransposeVP(x,y,Z) ProductPV(x,Z,y)

/* Allocation/Deallocation Routines */
TVector CreateVector(int m);
TMatrix CreateMatrix(int m, int n);
void FreeVector(TVector x);
void FreeMatrix(TMatrix X);

/* Initialization Routines */
TVector InitializeVector(TVector x, PRECISION c);
TMatrix InitializeMatrix(TMatrix X, PRECISION c);

/* Assignment Routines */
TVector EquateVector(TVector x, TVector y);
TMatrix EquateMatrix(TMatrix X, TMatrix Y);
TMatrix Transpose(TMatrix X, TMatrix Y);
TMatrix IdentityMatrix(TMatrix X, int m);
TMatrix DiagonalMatrix(TMatrix X, TVector y);
TVector AbsV(TVector x, TVector y);
TMatrix AbsM(TMatrix X, TMatrix Y);
TVector MinusV(TVector x, TVector y);
TMatrix MinusM(TMatrix X, TMatrix Y);
TMatrix SubMatrix(TMatrix X, TMatrix Y, int brow, int bcol, int rows, int cols);
TVector ColumnVector(TVector x, TMatrix Y, int col);
TVector RowVector(TVector x, TMatrix Y, int row);
TMatrix ColumnMatrix(TMatrix X, TVector y);
TMatrix RowMatrix(TMatrix X, TVector y);

/* Addition Routines */
TVector AddVV(TVector x, TVector y, TVector z);
TMatrix AddMM(TMatrix X, TMatrix Y, TMatrix Z);
TVector SubtractVV(TVector x, TVector y, TVector z);
TMatrix SubtractMM(TMatrix X, TMatrix Y, TMatrix Z);

/* Multiplication Routines */
TVector ProductSV(TVector x, PRECISION s, TVector y);
#define ProductVS(x,y,s) ProductSV(x,s,y)
TMatrix ProductSM(TMatrix X, PRECISION s, TMatrix Y);
#define ProductMS(X,Y,s) ProductSM(X,s,Y)
TVector ProductVM(TVector x, TVector y, TMatrix Z);
TVector ProductMV(TVector x, TMatrix Y, TVector z);
TMatrix ProductMM(TMatrix X, TMatrix Y, TMatrix Z);
TVector ProductInverseVM(TVector x, TVector y, TMatrix Z);
TMatrix ProductInverseMM(TMatrix X, TMatrix Y, TMatrix Z);
TVector ProductInverseVU(TVector x, TVector y, TMatrix Z);
TMatrix ProductInverseMU(TMatrix X, TMatrix Y, TMatrix Z);
TVector ProductInverseVL(TVector x, TVector y, TMatrix Z);
TMatrix ProductInverseML(TMatrix X, TMatrix Y, TMatrix Z);
TVector InverseProductMV(TVector x, TMatrix Y, TVector z);
TMatrix InverseProductMM(TMatrix X, TMatrix Y, TMatrix Z);
TVector InverseProductUV(TVector x, TMatrix Y, TVector z);
TMatrix InverseProductUM(TMatrix X, TMatrix Y, TMatrix Z);
TVector InverseProductLV(TVector x, TMatrix Y, TVector z);
TMatrix InverseProductLM(TMatrix X, TMatrix Y, TMatrix Z);
TMatrix TransposeProductMM(TMatrix X, TMatrix Y, TMatrix Z);
#define TransposeProductMV(x,Y,z) ProductVM(x,z,Y)
TMatrix ProductTransposeMM(TMatrix X, TMatrix Y, TMatrix Z);
#define ProductTransposeVM(x,y,Z) ProductMV(x,Z,y)

/* Matrix Inverse Routines */
TMatrix Inverse_LU(TMatrix X, TMatrix Y);
TMatrix Inverse_SVD(TMatrix X, TMatrix Y);
TMatrix Inverse_Cholesky(TMatrix X, TMatrix Y);
TMatrix Inverse_UT(TMatrix X, TMatrix Y);
TMatrix Inverse_LT(TMatrix X, TMatrix Y);

/* Matrix Decompositions */
int SVD(TMatrix U, TVector d, TMatrix V, TMatrix A);
int QR(TMatrix Q, TMatrix R, TMatrix X);
int LU(TPermutation P, TMatrix X, TMatrix A);
TMatrix CholeskyUT(TMatrix T, TMatrix X);
TMatrix CholeskyLT(TMatrix T, TMatrix X);

TVector LU_SolveCol(TVector x, TVector y, TMatrix LU, int *permute);
TVector LU_SolveRow(TVector x, TVector y, TMatrix LU, int *permute);

/* Miscellaneous Routines */
PRECISION Norm(TVector x);
PRECISION DotProduct(TVector x, TVector y);
PRECISION InnerProduct(TVector x, TVector y, TMatrix S);
TMatrix OuterProduct(TMatrix X, TVector y, TVector z);
PRECISION Trace(TMatrix X);
PRECISION Determinant_LU(TMatrix X);
PRECISION Determinant_QR(TMatrix X);
int Rank_SVD(TMatrix X);
TVector CrossProduct_LU(TVector x, TMatrix Y);
TVector CrossProduct_QR(TVector x, TMatrix Y);

/* Kronecker Routines */
TVector Vec(TVector x, TMatrix Y);
TMatrix KroneckerProduct(TMatrix X, TMatrix Y, TMatrix Z);

/* Input - Output Routines */
void PrintVector(FILE *f, TVector x, char *format);
void PrintMatrix(FILE *f, TMatrix X, char *format);
int ReadMatrix(FILE *f, TMatrix X);
int ReadVector(FILE *f, TVector x);
int OutVectorFloat(FILE *f, TVector x);
int OutMatrixFloat(FILE *f, TMatrix X);
int OutVectorDouble(FILE *f, TVector x);
int OutMatrixDouble(FILE *f, TMatrix X);
TVector InVector(FILE *f, TVector x);
TMatrix InMatrix(FILE *f, TMatrix X);

void PrintPermutation(FILE *f, TPermutation);

/* Old Style Syntax *
#define Inverse(X,Y)                           Inverse_LU(X,Y)
#define TransposeProduct(X,Y,Z)                TransposeProductMM(X,Y,Z)
#define ProductTranspose(X,Y,Z)                ProductTransposeMM(X,Y,Z)
#define InverseProduct(X,Y,Z)                  InverseProductMM(X,Y,Z)
#define ProductInverse(X,Y,Z)                  ProductInverseMM(X,Y,Z)
#define VectorProductInverse(x,y,Z)            ProductInverseVM(x,y,Z)
#define ERROR(i)                               MatrixError(i)

// int Cholesky_U(TMatrix T, TMatrix X)                                           X = T'* T  (T upper triangular)
// int Cholesky_L(TMatrix T, TMatrix X)                                           X = T * T' (T lower triangular)
// int SingularValueDecomposition(TMatrix U, TVector d, TMatrix V, TMatrix A)     A = U * Diagonal(d) * V'
// int* QR_RPivot(TMatrix R, TMatrix X);
// int* QR_QRPivot(TMatrix Q, TMatrix R, TMatrix X);
/**/

#endif
