
#ifndef _VAR_SIMULATION_
#define _VAR_SIMULATION_

#include "matrix.h"
#include "m_rand.h"

/* Use fast algorithm for non-standard distribution */
#define USE_FAST_DRAW

/* Flags for normalization routine */
#define NORMALIZATION_NONE   0x0000
#define NORMALIZATION_DIAG   0x0001
#define NORMALIZATION_WZ     0x0002
#define NORMALIZATION_TRAD   0x0003
#define NORMALIZATION_CUSTOM 0x0004

/* Flags for maximization routine */
#define RANDOM_STARTING_VALUE  1
#define CURRENT_STARTING_VALUE 2

/* Error Constants */
#define SVAR_NO_ERR             0           // No error
#define SVAR_SIZE_ERR           1           // Vector, matrix or matrix array not the correct size 
#define SVAR_PARAMETER_ERR      2           // Parameter not in correct range
#define SVAR_MEMORY_ERR         3           // Out of memory
#define SVAR_SYMMETRIC_ERR      4           // Matrix not symmetric
#define SVAR_ORTHONORMAL_ERR    5           // Columns of matrix not orthonormal
#define SVAR_RESTRICTION_ERR    6           // Too many contemporaneous restrictions
#define SVAR_POSDEF_ERR         7           // Matrix not positive definite
#define SVAR_SIMULATE_ERR       8           // T_SVAR structure not ready for simulation
#define SVAR_DECOMPOSITION_ERR  9           // Error in decomposition routine

typedef void NormalizationFunctionType(TMatrix A, void *ptr);
typedef void NormalizationDestructType(void *ptr);

typedef struct
{
 int      n_var;           // number endogenous variables
 int      n_exg;           // number exogenous variables
 int      n_lag;           // number of lagged endogenous variables
 int      n_obs;           // number observations, not including initial conditions
 int      n_pre;           // number predetermined variables (n_lag*n_var + n_exg = n_pre)

 TMatrix *SimulateS;       // U*Sqrt(Inverse(S))                S = U'*{S_hat + O'*[Inverse(H_hat) - V*Inverse(H)*V']*O}*U
 TMatrix *SimulateH;       // V*Sqrt(Inverse(H))                H = V'*H_hat*V
 TMatrix *SimulateP;       // V*Inverse(H)*V'*O - W             O = H_hat*P_hat + H_hat*W

 TMatrix *U;
 TMatrix *V;
 TMatrix *W;
 
 TMatrix *UpdateS;         // S_hat + P_hat'*H_hat*P_hat = S_hat + (H_hat*P_hat)'*Inverese(H_hat)*(H_hat*P_hat)
 TMatrix *UpdateH;         // H_hat
 TMatrix *UpdateP;         // H_hat*P_hat 

} T_SVAR;

/* Construction and Destruction */
T_SVAR* Construct_SVAR(int n_var, int n_lag, int n_exg);
void Destruct_SVAR(T_SVAR *svar);

/* Restrictions */
int NoRestrictions(T_SVAR *svar);
int SetRestrictions(TMatrix *U, TMatrix *V, TMatrix *W, T_SVAR *svar);
int SetLinearRestrictions(TMatrix *Q, T_SVAR *svar);
int SetExclusionRestrictions(TMatrix EC, TMatrix EP, T_SVAR *svar);

/* Updating */
int Update(TMatrix *UpdateS, TMatrix *UpdateH, TMatrix *UpdateP, int n_obs, T_SVAR *svar);
int BaysianUpdate(TMatrix YY, TMatrix XX, TMatrix XY, int n_obs, T_SVAR *svar);
int PriorUpdate(TMatrix *S_bar, TMatrix *H_bar, TMatrix *P_bar, int n_obs, T_SVAR *svar);
int DataUpdate(TMatrix Y, TMatrix Z, int y1, int z1, int T, int n_obs, T_SVAR *svar);

/* Simulations */ 
int SetSimulations(T_SVAR *svar);

/* Check Structure */
int CheckRestrictions(TMatrix *U, TMatrix *V, TMatrix *W, int n_var, int n_pre);
int CheckUpdateMatrices(TMatrix UpdateS, TMatrix UpdateH, TMatrix UpdateP, int n_var, int n_pre);
int CheckUpdateArrays(TMatrix *UpdateS, TMatrix *UpdateH, TMatrix *UpdateP, int n_var, int n_pre);
int CheckSimulations(T_SVAR *svar);

void PrintSVAR(FILE *f, T_SVAR *svar);


/* Maximization Routine */
int Maximize(TMatrix A_hat, int max_iter, PRECISION tol_chg, PRECISION tol_det, int flag, T_SVAR *svar);

/* Maximum Likelihood Routines */
TMatrix GetA_hat(TMatrix A_hat, T_SVAR *svar);
TMatrix GetF_hat(TMatrix F_hat, T_SVAR *svar, TMatrix A_hat);
TMatrix GetB_hat(TMatrix F_hat, T_SVAR *svar, TMatrix A_hat);
TMatrix GetSigma_hat(TMatrix Sigma_hat, T_SVAR *svar, TMatrix A_hat);
PRECISION EvaluateLogLikelihood(TMatrix A, TMatrix F, T_SVAR *svar);

/* Simulation Routines */
void UniformSphericalInitialValue(TMatrix A, T_SVAR *svar);
void RandomInitialValue(TMatrix A, int n, T_SVAR *svar);
void MLEInitialValue(TMatrix A, int n, T_SVAR *svar);
void Simulate_A(TMatrix A, T_SVAR *svar);
void Simulate_F(TMatrix A, TMatrix F, T_SVAR *svar);
void Simulate_Full(TMatrix A, TMatrix F, T_SVAR *svar);
void Simulate_ReducedForm(TMatrix A, TMatrix B, T_SVAR *svar);
void Simulate_Sigma(TMatrix A, TMatrix Sigma, T_SVAR *svar);
void Simulate_FullReducedForm(TMatrix A, TMatrix Sigma, TMatrix B, T_SVAR *svar);

void Simulate_A_WZ(TMatrix A, TMatrix Ref, T_SVAR *svar);
void Simulate_Full_WZ(TMatrix A, TMatrix F, TMatrix Ref, T_SVAR *svar);
void SimulateA_A_Traditional(TMatrix A, TMatrix Ref, T_SVAR *svar);
void Simulate_Full_Traditional(TMatrix A, TMatrix F, TMatrix Ref, T_SVAR *svar);
void Simulate_A_Diagonal(TMatrix A, T_SVAR *svar);
void Simulate_Full_Diagonal(TMatrix A, TMatrix F, T_SVAR *svar);

/* Normalization Routines */
void Normalize_WZ(TMatrix A, TMatrix Ref);
void Normalize_Traditional(TMatrix A, TMatrix Ref);
void Normalize_Diagonal(TMatrix A);

/* Utilities */
//TMatrix SymmetricForm(TMatrix X, TMatrix Y);
//int Symmetric(TMatrix X);
//int OrthonormalColumns(TMatrix X);

#endif
